#ifndef _LINUX_FS_H
#define _LINUX_FS_H

#include <linux/types.h>

#define FMODE_READ 1
#define FMODE_WRITE 2

struct poll_table_struct;


/***************************************/
/*I had to add this because the fasync */
/*stuff needs it (pid,uid,signum)      */
/***************************************/
struct fown_struct { 
  int pid;                /* pid or -pgrp where SIGIO should be sent */ 
  uid_t uid, euid;        /* uid/euid of process setting the owner */ 
  int signum;             /* posix.1b rt signal to be delivered on IO */ 
}; 



struct file {
  void *  private_data;
  struct file_operations  *f_op;
  unsigned int            f_flags;
  loff_t                  f_pos;
  int                     f_mode;  /* should be mode_t */
  struct fown_struct      f_owner; /* for fasync stuff */
	
/*        struct list_head        f_list;
        struct dentry           *f_dentry;
        struct vfsmount         *f_vfsmnt;
        struct file_operations  *f_op;
        atomic_t                f_count;
        unsigned long           f_reada, f_ramax, f_raend, f_ralen, f_rawin;
        struct fown_struct      f_owner;
        unsigned int            f_uid, f_gid;
        int                     f_error;

        unsigned long           f_version;

*/        /* needed for tty driver, and maybe others */
/*        void                    *private_data;

*/        /* preallocated helper kiobuf to speedup O_DIRECT */
/*        struct kiobuf           *f_iobuf;
        long                    f_iobuf_lock;
*/
};

struct inode {
	int	i_rdev;
	
/*	
        struct list_head        i_hash;
        struct list_head        i_list;
        struct list_head        i_dentry;

        struct list_head        i_dirty_buffers;
        struct list_head        i_dirty_data_buffers;

        unsigned long           i_ino;
        atomic_t                i_count;
        kdev_t                  i_dev;
        umode_t                 i_mode;
        nlink_t                 i_nlink;
        uid_t                   i_uid;
        gid_t                   i_gid;
        kdev_t                  i_rdev;
        loff_t                  i_size;
        time_t                  i_atime;
        time_t                  i_mtime;
        time_t                  i_ctime;
        unsigned int            i_blkbits;
        unsigned long           i_blksize;
        unsigned long           i_blocks;
        unsigned long           i_version;
        struct semaphore        i_sem;
        struct semaphore        i_zombie;
        struct inode_operations *i_op;
        struct file_operations  *i_fop; *//* former ->i_op->default_file_ops */
/*        struct super_block      *i_sb;
        wait_queue_head_t       i_wait;
        struct file_lock        *i_flock;
        struct address_space    *i_mapping;
        struct address_space    i_data;
        struct dquot            *i_dquot[MAXQUOTAS];
*/        /* These three should probably be a union */
/*        struct list_head        i_devices;
        struct pipe_inode_info  *i_pipe;
        struct block_device     *i_bdev;
        struct char_device      *i_cdev;

        unsigned long           i_dnotify_mask; *//* Directory notify events */
/*        struct dnotify_struct   *i_dnotify; *//* for directory notifications */
/*
        unsigned long           i_state;

        unsigned int            i_flags;
        unsigned char           i_sock;

        atomic_t                i_writecount;
        unsigned int            i_attr_flags;
        __u32                   i_generation;
        union {
                struct minix_inode_info         minix_i;
                struct ext2_inode_info          ext2_i;
                struct ext3_inode_info          ext3_i;
                struct hpfs_inode_info          hpfs_i;
                struct ntfs_inode_info          ntfs_i;
                struct msdos_inode_info         msdos_i;
                struct umsdos_inode_info        umsdos_i;
                struct iso_inode_info           isofs_i;
                struct nfs_inode_info           nfs_i;
                struct sysv_inode_info          sysv_i;
                struct affs_inode_info          affs_i;
                struct ufs_inode_info           ufs_i;
                struct efs_inode_info           efs_i;
                struct romfs_inode_info         romfs_i;
                struct shmem_inode_info         shmem_i;
                struct coda_inode_info          coda_i;
                struct smb_inode_info           smbfs_i;
                struct hfs_inode_info           hfs_i;
                struct adfs_inode_info          adfs_i;
                struct qnx4_inode_info          qnx4_i;
                struct reiserfs_inode_info      reiserfs_i;
                struct bfs_inode_info           bfs_i;
                struct udf_inode_info           udf_i;
                struct ncp_inode_info           ncpfs_i;
                struct proc_inode_info          proc_i;
                struct socket                   socket_i;
                struct usbdev_inode_info        usbdev_i;
                struct jffs2_inode_info         jffs2_i;
                void                            *generic_ip;
        } u;
*/
};

struct file_operations {
        loff_t (*llseek) (struct file *, loff_t, int);
        ssize_t (*read) (struct file *, char *, size_t, loff_t *);
        ssize_t (*write) (struct file *, const char *, size_t, loff_t *);
        int (*ioctl) (struct inode *, struct file *, unsigned int, unsigned long);
        int (*open) (struct inode *, struct file *);
        int (*release) (struct inode *, struct file *);

        unsigned int (*poll) (struct file *, struct poll_table_struct *);
        int (*fasync) (int, struct file *, int);
/*
        struct module *owner;
        int (*readdir) (struct file *, void *, filldir_t);
        int (*mmap) (struct file *, struct vm_area_struct *);
        int (*flush) (struct file *);
        int (*fsync) (struct file *, struct dentry *, int datasync);
        int (*lock) (struct file *, int, struct file_lock *);
        ssize_t (*readv) (struct file *, const struct iovec *, unsigned long, lo
ff_t *);
        ssize_t (*writev) (struct file *, const struct iovec *, unsigned long, l
off_t *);
        ssize_t (*sendpage) (struct file *, struct page *, int, size_t, loff_t *
, int);
        unsigned long (*get_unmapped_area)(struct file *, unsigned long, unsigne
d long, unsigned long, unsigned long);
*/
};

struct fasync_struct {
  int     magic;
  int     fa_fd;
  struct  fasync_struct   *fa_next; /* singly linked list */
  struct  file            *fa_file;
  struct fown_struct          fa_owner;
  void *proc_ref;
  int bitmask;
};


extern int unregister_chrdev(unsigned int, const char *);
extern int register_chrdev(unsigned int, const char *, struct file_operations *);
extern void kill_fasync(struct fasync_struct **, int, int);
extern int fasync_helper(int, struct file *, int, struct fasync_struct **);


#endif
