
#ifndef _ASM_IO_H
#define _ASM_IO_H

#include <linux/config.h>
#include <linux/ioport.h>
#include <linux/vmalloc.h>
/*
 * This file contains the definitions for the x86 IO instructions
 * inb/inw/inl/outb/outw/outl and the "string versions" of the same
 * (insb/insw/insl/outsb/outsw/outsl). You can also use "pausing"
 * versions of the single-IO instructions (inb_p/inw_p/..).
 *
 * This file is not meant to be obfuscating: it's just complicated
 * to (a) handle it all in a way that makes gcc able to optimize it
 * as well as possible and (b) trying to avoid writing the same thing
 * over and over again with slight variations and possibly making a
 * mistake somewhere.
 */
#define IO_SPACE_LIMIT 0xffff


extern void * __ioremap(unsigned long offset, 
			unsigned long size, 
			unsigned long flags);

static inline void * ioremap (unsigned long offset, unsigned long size)
{
  return __ioremap(offset,size,0);
}

/*
 * This one maps high address device memory and turns off caching for that area.
 * it's useful if some control registers are in such an area and write combining
 * or read caching is not desirable:
 */
static inline void * ioremap_nocache (unsigned long offset, unsigned long size)
{
        return __ioremap(offset, size, _PAGE_PCD);
}

static inline void iounmap(void *virt_addr)
{
  ;
}

/*
 * Change virtual addresses to physical addresses and vv.
 * These are pretty trivial
 */
static inline unsigned long virt_to_phys(volatile void * address)
{
  return *(unsigned long*)address;
  //return __pa(address);
}

static inline void * phys_to_virt(unsigned long address)
{
  return (void*)address;
  //return __va(address);
}


#define page_to_phys(page)	((page - mem_map) << PAGE_SHIFT)

/*
 * IO bus memory addresses are also 1:1 with the physical address
 */
#define virt_to_bus virt_to_phys
#define bus_to_virt phys_to_virt
#define page_to_bus page_to_phys

#define flush_write_buffers()


/*the macros were heavily abused here, must be an optimization 
 *choice over readablity 
 *as stated above... :(, i'll make them inline functions instead, partly because
 *we have to make the appropiate sun call mapping... 
 */
#define readb(addr) linux_readb((unsigned long)addr);
#define writeb(b,addr) linux_writeb(b,(unsigned long)addr);

static inline unsigned char linux_readb(unsigned long addr)
{
  printk("inside linux_readb, not implemented yet!");
  return 0;
}
static inline void linux_writeb(unsigned char byte,unsigned long addr)
{
  printk("inside linux_writeb, not implemented yet!");
  ;
}



static inline void insb(unsigned short port, void *addr,unsigned long count)
{
  struct resource *res = NULL;

  int offset = 0;

  res = get_port_by_addr(port);

  if (res == NULL)
    {
      printk("Unallocated port\n");
      ;
    }
  else
    {
      offset = port - res -> start;
      ddi_rep_get8(res -> sun_handles8[offset],
		   (uint8_t*) addr,
		   res -> mapped_addrs8[offset],
		   count,
		   DDI_DEV_NO_AUTOINCR);
    }

}
static inline void outsb(unsigned short port, void *addr,unsigned long count)
{
  struct resource *res = NULL;

  int offset = 0;
  
  res = get_port_by_addr(port);

  if (res == NULL)
    {
      printk("Unallocated port\n");
    }
  else
    {
      offset = port - res -> start;
      ddi_rep_put8(res -> sun_handles8[offset],
		   (uint8_t*)addr,
		   res -> mapped_addrs8[offset],
		   count,
		   DDI_DEV_NO_AUTOINCR);
    }
}


static inline void insl(unsigned short port,void *addr,unsigned long count)
{
  struct resource *res = NULL;
  
  int offset = 0;
  
  res = get_port_by_addr(port);

  if (res == NULL)
    {
      printk("Unallocated port\n");
    }
  else
    {
      offset = port - res -> start;
      ddi_rep_get16(res -> sun_handles16[offset],
		    (uint16_t*)addr,
		    res -> mapped_addrs16[offset],
		    count,
		    DDI_DEV_NO_AUTOINCR);
    }

}

/***********************************/
/*outsl (): write 32-bit values to a single 32-bit port.*/
/***********************************/
static inline void outsl(unsigned short port, void *addr,unsigned long count)
{
  struct resource *res = NULL;
  
  int offset = 0;
  
  res = get_port_by_addr(port);
  
  if (res == NULL)
    {
      printk("unallocated port\n");
      
    }
  else
    {
      /**********************************
       *FIX ME: We need to find out how
       *to allocate 8,16, and 32 bit ports.
       *presently, only 8-bit ports are supported.
       **********************************/
      offset = port - res -> start;
      ddi_rep_put32(res -> sun_handles32[offset],
		    (uint32_t*) addr,
		    (uint32_t*)res -> mapped_addrs32[offset],
		    count,
		    DDI_DEV_NO_AUTOINCR);
    }
}


#define outb(byte,port) linux_outb(byte,port)
#define inb(port) linux_inb(port)


static inline unsigned char linux_inb(unsigned port)
{
  struct resource *res = NULL;
  unsigned char in_byte = 0;
  int offset = 0;
  //printk("linux_inb");
  res = get_port_by_addr(port);

  
  if (res == NULL)
    {
      printk("inb:Unallocated port %08ux\n",port);
      return 0;
    }
  else
    {
      
      offset = port - res -> start;
      //printk("inputting from port %081x",port); 
      in_byte = ddi_get8(res -> sun_handles8[offset],
			 res -> mapped_addrs8[offset]);
      //printk("in byte is:%d -> %c\n",(int)in_byte,(char)in_byte);
      return in_byte;
    }

}

static inline void linux_outb(unsigned char byte,unsigned port)
{
  struct resource *res = NULL;
  int offset = 0;

  //printk("linux_outb");

  //printk("outputting %d to port %08lx",byte,port);
  res = get_port_by_addr(port);
  
  if (res == NULL)
    {
      printk("outb:Unallocated port %081x\n",port);
    }
  else
    {
      offset = port - res -> start;
      //printk("offset is %d output is %d -> %c\n",offset,(int)byte,(char)byte);
      ddi_put8(res -> sun_handles8[offset],
	       res -> mapped_addrs8[offset],byte);
    }
}



/*slow down functions, need to put a delay after an I/O operation as stated in
 *the LINUX device drivers manual */

static inline unsigned inb_p(unsigned port)
{
  unsigned data;
  data = inb(port);
  //delay 10 microseconds...
  delay(10 * drv_usectohz(1));
  return(data);
}

static inline void outb_p(unsigned char byte,unsigned port)
{
  outb(byte,port);
  //delay 10 microseconds...
  delay(10 * drv_usectohz(1));
}


#endif

