#ifndef __i386_UACCESS_H
#define __i386_UACCESS_H

#include <sys/types.h>
#include <sys/ddi.h>
#include <sys/sunddi.h>

#ifndef FKIOCTL
#define FKIOCTL 0x80000000
#endif

#define VERIFY_READ 0
#define VERIFY_WRITE 1


/*
 * The fs value determines whether argument validity checking should be
 * performed or not.  If get_fs() == USER_DS, checking is performed, with
 * get_fs() == KERNEL_DS, checking is bypassed.
 *
 * For historical reasons, these macros are grossly misnamed.
 */

#define MAKE_MM_SEG(s)	((mm_segment_t) { (s) })

#define KERNEL_DS	MAKE_MM_SEG(0xFFFFFFFF)

//#define USER_DS		MAKE_MM_SEG(PAGE_OFFSET)

#define get_ds()	(KERNEL_DS)

//#define get_fs()	(current->addr_limit)
#define get_fs() KERNEL_DS

//#define set_fs(x)	(current->addr_limit = (x))
#define set_fs(x) ;



static inline int verify_area(int type, const void * addr, unsigned long size)
{
  return 0;
}

//#define copy_to_user(to,from,n)    copyout(from,to,n)
#define copy_to_user(dest,src,n)                            \
({                                                          \
   int error;                                               \
   error = ddi_copyout((void*)src,(void*)dest,n,0);         \
   if (error != 0)                                          \
   {                                                        \
      printk("error copying,trying to do FKIOCTL\n");       \
      error = ddi_copyout((void*)src,(void*)dest,n,FKIOCTL);\
   }                                                        \
   error;                                                   \
})



#define copy_from_user(to,from,n)  copyin(from,to,n)
#define __copy_from_user(to,from,n) copyin(from,to,n)

/*set this to one (TRUE), since ddi_copyin and ddi_copyout handle this.*/ 
#define access_ok(type,addr,size)  1

#define __get_user(local,ptr)                                     \
({                                                                \
   int __ret;                                                     \
   printk("calling ddi_copyin on size %d\n",sizeof(*(ptr)));        \
   __ret = ddi_copyin((void*)ptr,(void*)&local,sizeof(*(ptr)),0); \
   printk("return %d\n",__ret);                                     \
   __ret;                                                         \
})

/*make the corresponding calls to ddi_copyout*/


#define __put_user(local,ptr) \
({          \
  int __ret = 0;                \
  __typeof__(*ptr) temp = local; \
  __ret = copy_to_user((void*)ptr,(void*)&temp,sizeof(*ptr)); \
  __ret; \
})

/*
#define __put_user(local,ptr)                                      \
({                                                                 \
   int __ret = 0;                                                  \
   void *driverbuf = (void*)&(local);                                \
   printk("calling ddi_copyout on size %d\n",sizeof(*ptr));        \
  __ret = copy_to_user((void*)ptr,driverbuf,sizeof(*ptr));             \
   __ret;                                                          \
})
*/


//just the same thing, since the checking is done by ddi.
#define put_user(local,ptr)     __put_user(local,ptr)
#define get_user(local,ptr)     __get_user(local,ptr)


#endif
