#ifndef _LINUX_FS_H
#define _LINUX_FS_H

#include <asm/byteorder.h>

#include <linux/types.h>
#include <linux/dcache.h>
#include <linux/shmem_fs.h>
#include <linux/kdev_t.h>
#include <linux/stat.h>

#include <linux/fcntl.h>


#define BLOCK_SIZE_BITS 10

#define FMODE_READ 1
#define FMODE_WRITE 2

#define READ 0
#define WRITE 1
#define READA 2		/* read-ahead  - don't block if no resources */
#define SPECIAL 4	/* For non-blockdevice requests in request queue */

struct poll_table_struct;

/* the read-only stuff doesn't really belong here, but any other place is
   probably as bad and I don't want to create yet another include file. */

#define BLKROSET   _IO(0x12,93)	/* set device read-only (0 = read-write) */
#define BLKROGET   _IO(0x12,94)	/* get read-only status (0 = read_write) */
#define BLKRRPART  _IO(0x12,95)	/* re-read partition table */
#define BLKGETSIZE _IO(0x12,96)	/* return device size /512 (long *arg) */
#define BLKFLSBUF  _IO(0x12,97)	/* flush buffer cache */
#define BLKRASET   _IO(0x12,98)	/* Set read ahead for block device */
#define BLKRAGET   _IO(0x12,99)	/* get current read ahead setting */
#define BLKFRASET  _IO(0x12,100)/* set filesystem (mm/filemap.c) read-ahead */
#define BLKFRAGET  _IO(0x12,101)/* get filesystem (mm/filemap.c) read-ahead */
#define BLKSECTSET _IO(0x12,102)/* set max sectors per request (ll_rw_blk.c) */
#define BLKSECTGET _IO(0x12,103)/* get max sectors per request (ll_rw_blk.c) */
#define BLKSSZGET  _IO(0x12,104)/* get block device sector size */


//#ifndef BLKPG
#define BLKPG      _IO(0x12,105)/* See blkpg.h */
//#endif


#define BLKELVGET 0x50
//#ifndef BLKELVGET
//#define BLKELVGET  _IOR(0x12,106,sizeof(blkelv_ioctl_arg_t))/* elevator get */
//#endif

#define BLKELVSET 0x51

//#ifndef BLKELVSET
//#define BLKELVSET  _IOW(0x12,107,sizeof(blkelv_ioctl_arg_t))/* elevator set */
//#endf

/* This was here just to show that the number is taken -
   probably all these _IO(0x12,*) ioctls should be moved to blkpg.h. */


/* A jump here: 108-111 have been used for various private purposes. */
#define BLKBSZGET  _IOR(0x12,112,sizeof(int))
#define BLKBSZSET  _IOW(0x12,113,sizeof(int))
#define BLKGETSIZE64 _IOR(0x12,114,sizeof(u64))	/* return device size in bytes (u64 *arg) */

#define BMAP_IOCTL 1		/* obsolete - kept for compatibility */
#define FIBMAP	   _IO(0x00,1)	/* bmap access */
#define FIGETBSZ   _IO(0x00,2)	/* get the block size used for bmap */

/***************************************/
/*I had to add this because the fasync */
/*stuff needs it (pid,uid,signum)      */
/***************************************/
struct fown_struct { 
  int pid;                /* pid or -pgrp where SIGIO should be sent */ 
  uid_t uid, euid;        /* uid/euid of process setting the owner */ 
  int signum;             /* posix.1b rt signal to be delivered on IO */ 
}; 



struct file {
  //sun stuff.
  int sun_opentype;
  //linux stuff.
  void *  private_data;
  struct file_operations  *f_op;
  unsigned int            f_flags;
  loff_t                  f_pos;
  int usage_count;
  struct dentry           *f_dentry;
  
  struct file *next_file;
  
  //  loff_t                  f_pos;
  int                     f_mode;  /* should be mode_t */
  struct fown_struct      f_owner; /* for fasync stuff */
	
  /*        struct list_head        f_list; 
        struct dentry           *f_dentry;
        struct vfsmount         *f_vfsmnt;
        struct file_operations  *f_op;
        atomic_t                f_count;
        unsigned long           f_reada, f_ramax, f_raend, f_ralen, f_rawin;
        struct fown_struct      f_owner;
        unsigned int            f_uid, f_gid;
        int                     f_error;

        unsigned long           f_version;

*/        /* needed for tty driver, and maybe others */
/*        void                    *private_data;

*/        /* preallocated helper kiobuf to speedup O_DIRECT */
/*        struct kiobuf           *f_iobuf;
        long                    f_iobuf_lock;
*/
};

struct inode {
  //int	i_rdev;
  //solaris dev_t for i_rdev;
  dev_t i_rdev;

  struct inode *next_inode;
  int usage_count;
/*	
        struct list_head        i_hash;
        struct list_head        i_list;
        struct list_head        i_dentry;

        struct list_head        i_dirty_buffers;
        struct list_head        i_dirty_data_buffers;

        unsigned long           i_ino;
        atomic_t                i_count;
        kdev_t                  i_dev;
        umode_t                 i_mode;
        nlink_t                 i_nlink;
        uid_t                   i_uid;
        gid_t                   i_gid;
        kdev_t                  i_rdev;
        loff_t                  i_size;
        time_t                  i_atime;
        time_t                  i_mtime;
        time_t                  i_ctime;
        unsigned int            i_blkbits;
        unsigned long           i_blksize;
        unsigned long           i_blocks;
        unsigned long           i_version;
        struct semaphore        i_sem;
        struct semaphore        i_zombie;
        struct inode_operations *i_op;
        struct file_operations  *i_fop; *//* former ->i_op->default_file_ops */
/*        struct super_block      *i_sb;
        wait_queue_head_t       i_wait;
        struct file_lock        *i_flock;
        struct address_space    *i_mapping;
        struct address_space    i_data;
        struct dquot            *i_dquot[MAXQUOTAS];
*/        /* These three should probably be a union */
/*        struct list_head        i_devices;
        struct pipe_inode_info  *i_pipe;
        struct block_device     *i_bdev;
        struct char_device      *i_cdev;

        unsigned long           i_dnotify_mask; *//* Directory notify events */
/*        struct dnotify_struct   *i_dnotify; *//* for directory notifications */
/*
        unsigned long           i_state;

        unsigned int            i_flags;
        unsigned char           i_sock;

        atomic_t                i_writecount;
        unsigned int            i_attr_flags;
        __u32                   i_generation;
        union {
                struct minix_inode_info         minix_i;
                struct ext2_inode_info          ext2_i;
                struct ext3_inode_info          ext3_i;
                struct hpfs_inode_info          hpfs_i;
                struct ntfs_inode_info          ntfs_i;
                struct msdos_inode_info         msdos_i;
                struct umsdos_inode_info        umsdos_i;
                struct iso_inode_info           isofs_i;
                struct nfs_inode_info           nfs_i;
                struct sysv_inode_info          sysv_i;
                struct affs_inode_info          affs_i;
                struct ufs_inode_info           ufs_i;
                struct efs_inode_info           efs_i;
                struct romfs_inode_info         romfs_i;
                struct shmem_inode_info         shmem_i;
                struct coda_inode_info          coda_i;
                struct smb_inode_info           smbfs_i;
                struct hfs_inode_info           hfs_i;
                struct adfs_inode_info          adfs_i;
                struct qnx4_inode_info          qnx4_i;
                struct reiserfs_inode_info      reiserfs_i;
                struct bfs_inode_info           bfs_i;
                struct udf_inode_info           udf_i;
                struct ncp_inode_info           ncpfs_i;
                struct proc_inode_info          proc_i;
                struct socket                   socket_i;
                struct usbdev_inode_info        usbdev_i;
                struct jffs2_inode_info         jffs2_i;
                void                            *generic_ip;
        } u;
*/
};

struct block_device_operations {
	int (*open) (struct inode *, struct file *);
	int (*release) (struct inode *, struct file *);
	int (*ioctl) (struct inode *, struct file *, unsigned, unsigned long);
	int (*check_media_change) (kdev_t);
	int (*revalidate) (kdev_t);
	struct module *owner;
};



struct file_operations {
  struct module *owner;
  loff_t (*llseek) (struct file *, loff_t, int);
        ssize_t (*read) (struct file *, char *, size_t, loff_t *);
        ssize_t (*write) (struct file *, const char *, size_t, loff_t *);
        int (*ioctl) (struct inode *, struct file *, unsigned int, unsigned long);
        int (*open) (struct inode *, struct file *);
        int (*release) (struct inode *, struct file *);

        unsigned int (*poll) (struct file *, struct poll_table_struct *);
        int (*fasync) (int, struct file *, int);
/*
        struct module *owner;
        int (*readdir) (struct file *, void *, filldir_t);
        int (*mmap) (struct file *, struct vm_area_struct *);
        int (*flush) (struct file *);
        int (*fsync) (struct file *, struct dentry *, int datasync);
        int (*lock) (struct file *, int, struct file_lock *);
        ssize_t (*readv) (struct file *, const struct iovec *, unsigned long, lo
ff_t *);
        ssize_t (*writev) (struct file *, const struct iovec *, unsigned long, l
off_t *);
        ssize_t (*sendpage) (struct file *, struct page *, int, size_t, loff_t *
, int);
        unsigned long (*get_unmapped_area)(struct file *, unsigned long, unsigne
d long, unsigned long, unsigned long);
*/
};

struct fasync_struct {
  int     magic;
  int     fa_fd;
  struct  fasync_struct   *fa_next; /* singly linked list */
  struct  file            *fa_file;
  struct fown_struct          fa_owner;
  void *proc_ref;
  int bitmask;
};

/* bh state bits */
enum bh_state_bits {
  BH_Uptodate,    /* 1 if the buffer contains valid data */
  BH_Dirty,       /* 1 if the buffer is dirty */
  BH_Lock,        /* 1 if the buffer is locked */
  BH_Req,         /* 0 if the buffer has been invalidated */
  BH_Mapped,      /* 1 if the buffer has a disk mapping */
  BH_New,         /* 1 if the buffer is new and not yet written out */
  BH_Async,       /* 1 if the buffer is under end_buffer_io_async I/O \
		   */
  BH_Wait_IO,     /* 1 if we should write out this buffer */
  BH_launder,     /* 1 if we should throttle on this buffer */
  BH_JBD,         /* 1 if it has an attached journal_head */

  BH_PrivateStart,/* not a state bit, but the first bit available
		   * for private allocation by other entities
		   */
};



/*
 * Try to keep the most commonly used fields in single cache lines (16
 * bytes) to improve performance.  This ordering should be
 * particularly beneficial on 32-bit processors.
 * 
 * We use the first 16 bytes for the data which is used in searches
 * over the block hash lists (ie. getblk() and friends).
 * 
 * The second 16 bytes we use for lru buffer scans, as used by
 * sync_buffers() and refill_freelist().  -- sct
 */
struct buffer_head {
	/* First cache line: */
	struct buffer_head *b_next;	/* Hash queue list */
	unsigned long b_blocknr;	/* block number */
	unsigned short b_size;		/* block size */
	unsigned short b_list;		/* List that this buffer appears */
	kdev_t b_dev;			/* device (B_FREE = free) */

	atomic_t b_count;		/* users using this block */
	kdev_t b_rdev;			/* Real device */
	unsigned long b_state;		/* buffer state bitmap (see above) */
	unsigned long b_flushtime;	/* Time when (dirty) buffer should be written */

	struct buffer_head *b_next_free;/* lru/free list linkage */
	struct buffer_head *b_prev_free;/* doubly linked list of buffers */
	struct buffer_head *b_this_page;/* circular list of buffers in one page */
	struct buffer_head *b_reqnext;	/* request queue */

	struct buffer_head **b_pprev;	/* doubly linked list of hash-queue */
	char * b_data;			/* pointer to data block */
	struct page *b_page;		/* the page this bh is mapped to */
	void (*b_end_io)(struct buffer_head *bh, int uptodate); /* I/O completion */
 	void *b_private;		/* reserved for b_end_io */

	unsigned long b_rsector;	/* Real buffer location on disk */
	wait_queue_head_t b_wait;

	struct inode *	     b_inode;
	struct list_head     b_inode_buffers;	/* doubly linked list of inode dirty buffers */

  //sun stuff.
  struct buf *sun_buffer;
  unsigned long lbyte_offset;
  unsigned long nr_sectors;
  struct buffer_head *p_next;
};

#define buffer_locked(bh) 1


/* fs/devices.c */

extern int invalidate_device(kdev_t, int);

extern int unregister_chrdev(unsigned int, const char *);
extern int register_chrdev(unsigned int, const char *, struct file_operations *);

extern int register_blkdev(unsigned int, const char *, struct block_device_operations *);
extern int unregister_blkdev(unsigned int, const char *);

extern const char *kdevname(kdev_t);
extern void kill_fasync(struct fasync_struct **, int, int);
extern int fasync_helper(int, struct file *, int, struct fasync_struct **);

extern loff_t no_llseek(struct file *file, loff_t offset, int origin);

extern void refile_buffer(struct buffer_head * buf);

//extern struct buffer_head * bread(kdev_t, int, int);
extern struct buffer_head * linux_bread(kdev_t, int, int);

extern void __brelse(struct buffer_head *);
static inline void linux_brelse(struct buffer_head *buf)
{
  /*
	if (buf)
		__brelse(buf);
  */
}


/*
 * This is called by bh->b_end_io() handlers when I/O has completed.
 */
static inline void mark_buffer_uptodate(struct buffer_head * bh, int on)
{
  /*
	if (on)
	set_bit(BH_Uptodate, &bh->b_state);
	else
		clear_bit(BH_Uptodate, &bh->b_state);
  */
}


extern int check_disk_change(kdev_t);



//#define invalidate_buffers(dev)	__invalidate_buffers((dev), 0)
#define invalidate_buffers(dev) 


extern int fsync_dev(kdev_t);
extern ssize_t block_read(struct file *, char *, size_t, loff_t *);

//#define read_ahead get_read_ahead_ptr()
//int *get_read_ahead_ptr();
extern int read_ahead[];

extern ssize_t block_write(struct file *, const char *, size_t, loff_t *);

/* Alas, no aliases. Too much hassle with bringing module.h everywhere */


#define fops_get(fops) (fops)

/*
#define fops_get(fops) \
	(((fops) && (fops)->owner)	\
		? ( try_inc_mod_count((fops)->owner) ? (fops) : NULL ) \
		: (fops))

#define fops_put(fops) \
do {	\
	if ((fops) && (fops)->owner) \
		__MOD_DEC_USE_COUNT((fops)->owner);	\
} while(0)
*/
#define fops_put(fops) do {;} while(0)

#endif
