#ifndef _LINUX_SCHED_H
#define _LINUX_SCHED_H

#include <asm/param.h> /* for HZ */
#include <linux/kernel.h>
#include <linux/wait.h>
#include <linux/binfmts.h>
#include <linux/time.h>
//#include <asm/semaphore.h>
#include <linux/timer.h>
#include <linux/signal.h>

#include <sys/proc.h>


/*
 * cloning flags:
 */
#define CSIGNAL         0x000000ff      /* signal mask to be sent at exit */
#define CLONE_VM        0x00000100      /* set if VM shared between processes */
#define CLONE_FS        0x00000200      /* set if fs info shared between processes */
#define CLONE_FILES     0x00000400      /* set if open files shared between processes */
#define CLONE_SIGHAND   0x00000800      /* set if signal handlers and blocked signals shared */
#define CLONE_PID       0x00001000      /* set if pid shared */
#define CLONE_PTRACE    0x00002000      /* set if we want to let tracing continue on the child too */
#define CLONE_VFORK     0x00004000      /* set if the parent wants the child to wake it up on mm_release */
#define CLONE_PARENT    0x00008000      /* set if we want to have the same parent as the cloner */
#define CLONE_THREAD    0x00010000      /* Same thread group? */

#define CLONE_SIGNAL    (CLONE_SIGHAND | CLONE_THREAD)



#define TASK_RUNNING		0
#define TASK_INTERRUPTIBLE	1
#define TASK_UNINTERRUPTIBLE	2
#define TASK_ZOMBIE		4
#define TASK_STOPPED		8

/*#define wait_event_interruptible(wq, condition) 0*/   
/*#define wake_up_interruptible(x)       0 */
#define set_current_state(state_value) __set_current_state(state_value);

#define __set_current_state(state_value)                        
//                 do { current->state = (state_value); } while (0) 



void schedule(void);


#include <asm/current.h>


#define jiffies (unsigned long)ddi_get_lbolt()
//extern unsigned long volatile jiffies;

extern signed long FASTCALL(schedule_timeout(signed long timeout));


#define wait_event(wq,condition)                                 \
({                                                               \
   wait_queue_head_t *t = &wq;                                   \
   int __ret = 1;                                                \
   if (t -> initialized == FALSE) {init_waitqueue_head(t);}      \
   mutex_enter(&(t -> queue_mutex));                             \
   printk("entering condition");                                 \
   while(!condition && __ret != 0)                               \
   {                                                             \
      t -> numwaiting++;                                         \
      __ret = cv_wait(&(t -> queue_condvar),&(t -> queue_mutex));  \
      t -> numwaiting--;printk("__ret %d",__ret);                \
   }                                                             \
   printk("exiting condition");                                  \
   mutex_exit(&(t -> queue_mutex));                              \
   if (__ret == 0) { __ret = 1;} else {__ret=0;}                 \
   __ret;                                                        \
})


/**********************************************/
#define wait_event_interruptible(wq,condition)                   \
({                                                               \
   wait_queue_head_t *t = &wq;                                   \
   int __ret = 1;                                                \
   if (t -> initialized == FALSE) {init_waitqueue_head(t);}      \
   mutex_enter(&(t -> queue_mutex));                             \
   printk("entering condition");                                 \
   while(!condition && __ret != 0)                               \
   {                                                             \
      t -> numwaiting++;                                         \
      __ret = cv_wait_sig(&(t -> queue_condvar),&(t -> queue_mutex)); \
      t -> numwaiting--;printk("__ret %d",__ret);                \
   }                                                             \
   printk("exiting condition");                                  \
   mutex_exit(&(t -> queue_mutex));                                 \
   if (__ret == 0) { __ret = 1;} else {__ret=0;}                 \
   __ret;                                                        \
})
/**********************************************/
/**********************************************/
#define wake_up_interruptible(wq)                                \
({                                                               \
  wait_queue_head_t *t = (wq);                                   \
  int __ret = 0;                                                 \
  if (t -> initialized == FALSE) {init_waitqueue_head(t);}       \
  mutex_enter(&((t) -> queue_mutex));                               \
 cv_broadcast(&((t) -> queue_condvar));                             \
 mutex_exit(&((t) -> queue_mutex));                                 \
 __ret;                                                          \
})
/***********************************************/


#undef NOT_DEFINED
struct task_struct {
  struct list_head list_entry;
  int sigpending;
  char comm[16];
  uid_t uid,euid,suid,fsuid;
  struct tty_struct *tty; /* NULL if no tty */
  volatile long state;
  sigset_t blocked;

  //sun stuff.
  int is_thread;
  kthread_t *threadp;
  kmutex_t mutex;
  kcondvar_t *cv_wait;
  int pits_refcnt;
#ifdef NOT_DEFINED
        /*
         * offsets of these are hardcoded elsewhere - touch with care
         */
        volatile long state;    /* -1 unrunnable, 0 runnable, >0 stopped */
        unsigned long flags;    /* per process flags, defined below */
        int sigpending;
        mm_segment_t addr_limit;        /* thread address space:
                                              0-0xBFFFFFFF for user-thead
                                              0-0xFFFFFFFF for kernel-thread
                                         */
        struct exec_domain *exec_domain;
#endif
        volatile long need_resched;
#ifdef NOT_DEFINED
        unsigned long ptrace;

        int lock_depth;         /* Lock depth */

/*
 * offset 32 begins here on 32-bit platforms. We keep
 * all fields in a single cacheline that are needed for
 * the goodness() loop in schedule().
 */
        long counter;
        long nice;
        unsigned long policy;
        struct mm_struct *mm;
        int processor;
        /*
         * cpus_runnable is ~0 if the process is not running on any
         * CPU. It's (1 << cpu) if it's running on a CPU. This mask
         * is updated under the runqueue lock.
         *
         * To determine whether a process might run on a CPU, this
         * mask is AND-ed with cpus_allowed.
         */
        unsigned long cpus_runnable, cpus_allowed;
        /*
         * (only the 'next' pointer fits into the cacheline, but
         * that's just fine.)
         */
        struct list_head run_list;
        unsigned long sleep_time;

        struct task_struct *next_task, *prev_task;
        struct mm_struct *active_mm;
        struct list_head local_pages;
        unsigned int allocation_order, nr_local_pages;

/* task state */
        struct linux_binfmt *binfmt;
        int exit_code, exit_signal;
        int pdeath_signal;  /*  The signal sent when the parent dies  */
        /* ??? */
        unsigned long personality;
        int did_exec:1;
        
#endif
  pid_t pid;
  pid_t pgrp;
#ifdef NOT_DEFINED
        pid_t tty_old_pgrp;
#endif

        pid_t session;
#ifdef NOT_DEFINED
        pid_t tgid;
        /* boolean value for session group leader */
        int leader;
        /*
         * pointers to (original) parent process, youngest child, younger sibling,
         * older sibling, respectively.  (p->father can be replaced with
         * p->p_pptr->pid)
         */
        struct task_struct *p_opptr, *p_pptr, *p_cptr, *p_ysptr, *p_osptr;
        struct list_head thread_group;

        /* PID hash table linkage. */
        struct task_struct *pidhash_next;
        struct task_struct **pidhash_pprev;

        wait_queue_head_t wait_chldexit;        /* for wait4() */
        struct completion *vfork_done;          /* for vfork() */
        unsigned long rt_priority;
        unsigned long it_real_value, it_prof_value, it_virt_value;
        unsigned long it_real_incr, it_prof_incr, it_virt_incr;
        struct timer_list real_timer;
        struct tms times;
        unsigned long start_time;
        long per_cpu_utime[NR_CPUS], per_cpu_stime[NR_CPUS];
/* mm fault and swap info: this can arguably be seen as either mm-specific or th
 * read-specific */
        unsigned long min_flt, maj_flt, nswap, cmin_flt, cmaj_flt, cnswap;
        int swappable:1;
/* process credentials */
        uid_t uid,euid,suid,fsuid;
        gid_t gid,egid,sgid,fsgid;
        int ngroups;
        gid_t   groups[NGROUPS];
        kernel_cap_t   cap_effective, cap_inheritable, cap_permitted;
        int keep_capabilities:1;
        struct user_struct *user;
/* limits */
        struct rlimit rlim[RLIM_NLIMITS];
        unsigned short used_math;
        char comm[16];
/* file system info */
        int link_count, total_link_count;
        struct tty_struct *tty; /* NULL if no tty */
        unsigned int locks; /* How many file locks are being held */
/* ipc stuff */
        struct sem_undo *semundo;
        struct sem_queue *semsleeping;
/* CPU-specific state of this task */
        struct thread_struct thread;
/* filesystem information */
        struct fs_struct *fs;
/* open file information */
        struct files_struct *files;
/* signal handlers */
        spinlock_t sigmask_lock;        /* Protects signal and blocked */
        struct signal_struct *sig;

        sigset_t blocked;
        struct sigpending pending;

        unsigned long sas_ss_sp;
        size_t sas_ss_size;
        int (*notifier)(void *priv);
        void *notifier_data;
        sigset_t *notifier_mask;

/* Thread group tracking */
        u32 parent_exec_id;
        u32 self_exec_id;
/* Protection of (de-)allocation: mm, files, fs, tty */
        spinlock_t alloc_lock;

/* journalling filesystem info */
        void *journal_info;
#endif /*NOT_DEFINED*/
};


extern void FASTCALL(__wake_up(wait_queue_head_t *q, unsigned int mode, int nr));

#define wake_up(x)			__wake_up((x),TASK_UNINTERRUPTIBLE | TASK_INTERRUPTIBLE, 1)

extern void FASTCALL(add_wait_queue(wait_queue_head_t *q, wait_queue_t *wait));
extern void FASTCALL(add_wait_queue_exclusive(wait_queue_head_t *q, wait_queue_t *wait));
extern void FASTCALL(remove_wait_queue(wait_queue_head_t *q, wait_queue_t *wait));
extern void daemonize(void);

extern void FASTCALL(interruptible_sleep_on(wait_queue_head_t *q));
//extern void interruptible_sleep_on(wait_queue_head_t *q);

extern long FASTCALL(interruptible_sleep_on_timeout(wait_queue_head_t *q,
						    signed long timeout));



extern int kill_proc(pid_t,int,int);
extern int send_sig(int, struct task_struct *, int);


static inline int signal_pending(struct task_struct *p)
{
	        return (p->sigpending != 0);
}

/*find out where these are defined in order to overwrite them with the proper sun calls.*/

extern int request_irq(unsigned int,
		       void (*handler)(int, void *, struct pt_regs *),
		       unsigned long, const char *, void *);
extern void free_irq(unsigned int, void *);

#endif

