
#ifndef _PITS_DMA
#define _PITS_DMA

//Linux includes.
#include <linux/spinlock.h>


//Solaris includes.
#include <sys/ddi.h>
#include <sys/sunddi.h>
#include <sys/types.h>

//PITS includes.
#include <misc/config.h>


#define PITS_DMA_STREAMING_BUF_SIZE (size_t)0xffff

#define DEFAULT_PCI_DMA_ATTR \
{ dma_attr_version    : DMA_ATTR_V0,                           \
  dma_attr_addr_lo    : (uint64_t)0,                           \
  dma_attr_addr_hi    : (uint64_t)0xffffffff,                  \
  dma_attr_count_max  : (uint64_t)0x00ffffff,                  \
  dma_attr_align      : (uint64_t)4,                           \
  dma_attr_burstsizes : DEFAULT_BURSTSIZE | BURST32 | BURST64, \
  dma_attr_minxfer    : 1,                                     \
  dma_attr_maxxfer    : (uint64_t)0x00ffffff,                  \
  dma_attr_seg        : (1<<24)-1,                             \
  dma_attr_sgllen     : 1,                                     \
  dma_attr_granular   : 1,                                     \
  dma_attr_flags      : 0                                      \
}




struct pits_dma_memory;

struct pits_dma_handle
{
  dev_info_t *dip;
  ddi_dma_handle_t handle;
  ddi_dma_attr_t dma_attr;
  struct pits_dma_memory *memory;
  struct pits_dma_handle *next;
};

struct pits_dma_memory
{
  struct pits_dma_handle *phandle;
  ddi_dma_handle_t mem_handle;
  caddr_t addr;
  int size;
  int flags;
  int bound;
  ddi_dma_cookie_t cookie;
  int cookie_count;
  struct pits_dma_memory *next;
};

struct pits_streaming_dma
{
  struct pits_dma_memory *dma_memory;
  void *kernel_buffer;
  int buffer_size;
  spinlock_t dma_lock;
  int mapped;
};


int init_pci_dma(dev_info_t *dip);
int fini_pci_dma(dev_info_t *dip);

struct pits_dma_handle *createDMAHandle(dev_info_t *dip,
					ddi_dma_attr_t *attr);

struct pits_dma_memory *createDMAMemory(struct pits_dma_handle *handle,
					int type,
					int size);
int bindDMAMemory(struct pits_dma_memory *memory,
		  int direction);
int unbindDMAMemory(struct pits_dma_memory *memory);
int deleteDMAMemory(struct pits_dma_memory **memory);
int deleteDMAHandle(struct pits_dma_handle **handle);
/*
void pits_dma_alloc_handle_error(int status);
void pits_dma_alloc_mem_error(int status);
void pits_dma_bind_handle_error(int status);
*/
#endif
