#ifndef _PITS_IRQ_H_
#define _PITS_IRQ_H_

/*Linux Includes*/
#include <linux/interrupt.h> /*for tasklet_struct*/

#define PITS_NIRQ 20

typedef struct linux_irq_struct
{
  char devname[80];
  dev_info_t *dip;
  void (*handler)(int, void *, struct pt_regs *);
  void *dev_id;
  struct pt_regs ipt_regs;
  int flags;
  ddi_idevice_cookie_t idevice_cookie;

  struct linux_irq_struct *next;
}linux_irq_t;

typedef struct pits_irq_struct
{
  uint8_t inumber;
  uint8_t initialized;
  uint8_t interrupting;

  uint64_t total_intr;
  uint64_t claimed_intr;
  uint64_t unclaimed_intr;
  uint64_t claimed_sintr;
  uint64_t unclaimed_sintr;
  timeout_id_t timer_id;

  dev_info_t *dip;
  int8_t enable_count;
  uint32_t run_count;
  uint32_t total_count;
  ddi_softintr_t soft_id;
  ddi_iblock_cookie_t sblock_cookie;
  ddi_iblock_cookie_t hblock_cookie;
  kmutex_t smutex; // softirq mutex.
  kmutex_t hmutex; // hardirq mutex.
  linux_irq_t *irq_list;
  uint8_t running;
  uint8_t scheduled;
  struct tasklet_struct irq_tasklet;
  kcondvar_t tasklet_wait;
}pits_irq_t;

typedef struct pits_irq_vector_struct
{
  int initialized;
  kmutex_t mutex; /*Global mutex*/
  pits_irq_t irqs[PITS_NIRQ];
  
}pits_irq_vec_t;

//the pits tasklet.
void pits_irq_tasklet_func(unsigned long inumber);

//pits soft interrupt.
uint_t pits_irq_soft_intr(caddr_t arg);

//the pits service routine.
uint_t pits_isr(caddr_t arguments);

/*these are going to be in the module pits_kernel.*/
int init_pits_irq();
int fini_pits_irq();
pits_irq_t *getPitsIrqStruct(int inumber);
linux_irq_t *getLinuxIrqStruct(int inumber);

int addLinuxIrq(dev_info_t *dip,
		unsigned int irq, 
		void (*handler)(int, void *, struct pt_regs *),
		unsigned long irqflags, 
		const char * devname,
		void *dev_id);

int removeLinuxIrq(unsigned int irq,void *dev_id);

void pits_enable_irq(unsigned int irq);
void pits_disable_irq(unsigned int irq);
void pits_disable_irq_nosync(unsigned int irq);

#endif
