#ifndef _PITS_SLAB_H

#define _PITS_SLAB_H
#include <sys/types.h>
#include <sys/kmem.h>
#include <sys/conf.h>

typedef struct pits_kernel_dma_memory
{
  dev_info_t *dip;
  ddi_dma_handle_t *handle;
  ddi_dma_handle_t mem_handle;
  caddr_t addr;
  uint32_t real_length;
  uint32_t req_size;
  int type;
  int bind_flags;
  int bound;
  kmem_cache_t *dma_handle_cache;


}pits_kernel_dma_memory_t;

typedef struct pits_kernel_memory_node
{
  void *allocated;
  size_t size;
  int flags;
  pits_kernel_dma_memory_t *dma;
  struct pits_kernel_memory_node *next;
}kmemptr_t;


//structure pits_kmem_list will keep track of the mem sizes and statistics
//as well.
typedef struct pits_klist_struct
{
  kmutex_t mutex;
  dev_info_t *dip;
  timeout_id_t timer_id;
  int timer_running;
  ddi_dma_attr_t dma_attr;
  ddi_device_acc_attr_t dev_acc_attr;
  uint64_t total_allocations;
  uint64_t total_allocated;
  uint64_t total_dma_allocated;
  uint64_t dma_handles;
  kmem_cache_t *dma_handle_cache;

  kmemptr_t *head;
}pits_klist_t;




//This function will report the total memory that has been allocated by the
//subsequent calls to kmalloc. (see linux/slab.h).
uint64_t getTotalMemAllocated();
uint64_t getTotalAllocations();

//functions that will init/fini the static variable klist structure.
int init_pits_slab(dev_info_t *dip);
int fini_pits_slab(dev_info_t *dip);

//functions that will modify the memory list.
int klist_isempty();
int klist_empty();
int klist_remove(void *ptr);
void *klist_add(size_t size,int flag);

kmemptr_t *kmalloc_normal(size_t size,int flag);

int init_dma_handle(ddi_dma_handle_t *ptr);

ddi_dma_handle_t *alloc_dma_handle(int kmflags);
void release_dma_handle(ddi_dma_handle_t *handle);


kmemptr_t *kmalloc_dma(size_t size,int flag);
void kfree_dma(kmemptr_t *ptr);
void kfree_normal(kmemptr_t *ptr);

//error routines.
void pits_dma_alloc_handle_error(int status);
void pits_dma_alloc_mem_error(int status);
void pits_dma_bind_handle_error(int status);
#endif
