/**********************************************/
/*pits_lib/globals.c                          */
/*Functions to init/fini global structures    */
/**********************************************/
//PITS Includes
#include <misc/globals.h>
//Linux Includes
#include <linux/slab.h>


//Static Global variable.
static struct pits_global_struct *pits_globals;

/***********************************************/
/*getGlobal() : returns the global structure   */
/* pointer.                                    */
/***********************************************/
struct pits_global_struct *getGlobal()
{
  return (pits_globals);
}

/***********************************************/
/*init_pits_globals() : initializes the global */
/***********************************************/
int init_pits_globals()
{
  int ret;
  ddi_dma_attr_t *pci_dma_attr = NULL;
  pits_globals = kmem_alloc(sizeof(struct pits_global_struct),KM_NOSLEEP);
  
  if (pits_globals == NULL)
    {
      printk("init_pits_global() : error allocating pits_globals\n");
      return DDI_FAILURE;
    }

  pci_dma_attr = &(pits_globals -> pci_dma_attr);
  pits_globals -> sun_dev_acc_attr.devacc_attr_version = DDI_DEVICE_ATTR_V0;
  pits_globals -> sun_dev_acc_attr.devacc_attr_endian_flags = DDI_NEVERSWAP_ACC;
  pits_globals -> sun_dev_acc_attr.devacc_attr_dataorder = DDI_STRICTORDER_ACC;

  pits_globals -> panic = FALSE;
  pits_globals -> current_dip = NULL;
  pits_globals -> attach_count = 0;
  pits_globals -> spin_lock_list = NULL;
  
  //initialize dma_attr
  pci_dma_attr -> dma_attr_version = DMA_ATTR_V0;
  pci_dma_attr -> dma_attr_addr_lo = (uint64_t)0;
  pci_dma_attr -> dma_attr_addr_hi = (uint64_t)0xffffffff;
  pci_dma_attr -> dma_attr_count_max = (uint64_t)0x00ffffff;
  pci_dma_attr -> dma_attr_align = (uint64_t)4;
  pci_dma_attr -> dma_attr_burstsizes = DEFAULT_BURSTSIZE | BURST32 | BURST64;
  pci_dma_attr -> dma_attr_minxfer = 1;
  pci_dma_attr -> dma_attr_maxxfer = (uint64_t)0x00ffffff;
  pci_dma_attr -> dma_attr_seg =(1<<24)-1;
  pci_dma_attr -> dma_attr_sgllen = 1;
  pci_dma_attr -> dma_attr_granular = 1;
  pci_dma_attr -> dma_attr_flags = 0;

  pits_globals -> pci_dma_initialized = FALSE;

  //init_pdi_table() : initializes the pits_dev_i table.
  ret = init_pdi_table();
  if (ret != DDI_SUCCESS)
    {
      printk("init_pits_globals() : error initializing pdi_table!\n");
      kmem_free(pits_globals,sizeof(struct pits_global_struct)); 
    }
  return (ret);
}

/******************************************************/
/*fini_pits_globals() : uninitializes the global      */
/*structure for the driver.                           */
/******************************************************/
int fini_pits_globals()
{
  if (pits_globals)
    {
      fini_pdi_table();
      kfree(pits_globals);
      pits_globals = NULL;
    }
  return DDI_SUCCESS;
}
