#include <misc/instminor.h>
#include <misc/sunlinux.h>


static inst_minor *im_mapping = NULL;

int addInstMinor(int instance,int minor,char *name)
{
  inst_minor *newnode = NULL;


  if (existsInstMinor(instance,minor))
    {
      return (FAILURE);
    }

  newnode = kmalloc(sizeof(inst_minor),0);
  
  if (newnode == NULL)
    {
      return (FAILURE);
    }

  newnode -> instance_number = instance;
  newnode -> minor_number = minor;
  newnode -> minor_name = name;
  newnode -> next = NULL;
  newnode -> prev = NULL;

  if (im_mapping == NULL)
    {
      /*first entry, add to head*/
      im_mapping = newnode;
    }
  else
    {
      /*add to front of the list*/
      newnode -> next = im_mapping;
      im_mapping -> prev = newnode;
      im_mapping = newnode;
    }
  return (SUCCESS);

}

int existsInstMinor(int instance_number,int minor_number)
{
  if ((getInstMinorByInstance(instance_number) == NULL) && 
      (getInstMinorByMinor(minor_number) == NULL))
    {
      return (FALSE);
    }
  else
    {
      return (TRUE);
    }

}
/*****************************************************
 *remInstMinorByInstance() : 
 *****************************************************/
int remInstMinorByInstance(int instance)
{
  inst_minor *node = NULL;

  //DEBUGDMESG("remInstMinorByInstance");

  node = getInstMinorByInstance(instance);

  if (node == NULL)
    {
      DEBUGDMESG("null node");
      return (FAILURE);
    }
  

  //DEBUGDMESG("remInstMinorByInstance, removing node");
  if (node -> prev == NULL)  
    {
      /*head just set the pointer to the next element.*/
      //DEBUGDMESG("Remove head");
      im_mapping = im_mapping -> next;
      if (im_mapping != NULL)
	{
	  im_mapping -> prev = NULL;
	}
    }
  else
    {
      if (node -> next == NULL)
      {
	//DEBUGDMESG("Remove end");
	/*at the end of the list.*/
	node -> prev -> next = NULL;
      }
      else
	{
	  /* at the middle */
	  //DEBUGDMESG("remove middle");
	  node -> prev -> next = node -> next;
	  node -> next -> prev = node -> prev;
	} 
    }
  //DEBUGDMESG("kfree");
  kfree(node);

  //DEBUGDMESG("remInstMinorByInstance, done");

  return(SUCCESS);

}

int remInstMinorByMinor(int minor)
{
  inst_minor *node = getInstMinorByMinor(minor);
  
  if (node == NULL)
    {
      return (FAILURE);
    }

  //DEBUGDMESG("remInstMinorByMinor, remove node");

  if (node -> prev == NULL) 
     
    {
      //DEBUGDMESG("remove head");
      /*head just set the pointer to the next element.*/ 
      im_mapping = im_mapping -> next;
      if (im_mapping != NULL)
	{
	  im_mapping -> prev = NULL;
	} 
    } 
  else 
    { 
      if (node -> next == NULL) 
	{ 
	  //DEBUGDMESG("remove end");
	  /*at the end of the list.*/ 
	  node -> prev -> next = NULL; 
	} 
    else 
      { 
        /* at the middle */ 
	//DEBUGDMESG("remove middle");
        node -> prev -> next = node -> next; 
        node -> next -> prev = node -> prev; 
      }
    }
  //DEBUGDMESG("kfree");
  kfree(node);

  //DEBUGDMESG("done remInstMinorByMinor");

  return(SUCCESS);
}

void remAllInstMinor()
{
  
  while (im_mapping != NULL)
    {
      remInstMinorByMinor(im_mapping -> minor_number);
    }
  
}

inst_minor *getInstMinorByInstance(int instance)
{

  inst_minor *node = im_mapping;

  //DEBUGDMESG("getInstMinorByInstance");

  while (node != NULL && node -> instance_number != instance)
    {
      node = node -> next;
    }
  /*if node is not found, null will be returned*/

  //DEBUGDMESG("done getInstMinorByInstance");
  return (node);
}

inst_minor *getInstMinorByMinor(int minor)
{
  inst_minor *node = im_mapping;


  //DEBUGDMESG("getInstMinorByMinor");
  while (node != NULL && node -> minor_number != minor)
    {
      node = node -> next;
    }

  //DEBUGDMESG("done getInstMinorByMinor");
  return (node);
}

int getInstance(int minor)
{
  /*
  inst_minor *node = getInstMinorByMinor(minor);

  if (node == NULL)
    {
      return FAILURE;
    }
  return (node -> instance_number); 
  */
  return 0;
}
int getMinor(int instance)
{
  inst_minor *node = getInstMinorByInstance(instance);

  if (node == NULL)
    {
      return FAILURE;
    }
  return (node -> minor_number);
}
