/*
 *	linux/arch/i386/kernel/irq.c
 *
 *	Copyright (C) 1992, 1998 Linus Torvalds, Ingo Molnar
 *
 * This file contains the code used by various IRQ handling routines:
 * asking for different IRQ's should be done through these routines
 * instead of just grabbing them. Thus setups with different IRQ numbers
 * shouldn't result in any weird surprises, and installing new handlers
 * should be easier.
 */

/*
 * (mostly architecture independent, will move to kernel/irq.c in 2.5.)
 *
 * IRQs are in fact implemented a bit like signal handlers for the kernel.
 * Naturally it's not a 1:1 relation, but there are similarities.
 */

/*PITS includes*/
#include <misc/globals.h>
#include <misc/pits_state.h>
#include <misc/pits_irq.h>

/*Linux includes*/
#include <linux/config.h>
#include <linux/ptrace.h>
#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/ioport.h>
#include <linux/interrupt.h>
//#include <linux/timex.h>
#include <linux/slab.h>
//#include <linux/random.h>
#include <linux/smp_lock.h>
//#include <linux/init.h>
//#include <linux/kernel_stat.h>
#include <linux/irq.h>
#include <linux/proc_fs.h>

#include <asm/atomic.h>
#include <asm/io.h>
//#include <asm/smp.h>
#include <asm/system.h>
#include <asm/bitops.h>
#include <asm/uaccess.h>
//#include <asm/pgalloc.h>
#include <asm/delay.h>
//#include <asm/desc.h>
#include <asm/irq.h>


/*
 * Controller mappings for all interrupt sources:
 */

//static void register_irq_proc (unsigned int irq);

/*
 * Special irq handlers.
 */



void no_action(int cpl, void *dev_id, struct pt_regs *regs) { }
	
#define MAXCOUNT 100000000
#define EFLAGS_IF_SHIFT 9



/*
 * This should really return information about whether
 * we should do bottom half handling etc. Right now we
 * end up _always_ checking the bottom half, which is a
 * waste of time and is not what some drivers would
 * prefer.
 */
int handle_IRQ_event(unsigned int irq, struct pt_regs * regs, struct irqaction * action)
{
  printk("inside handle_IRQ_event\n");
  printk("done handle_IRQ_event\n");
  return 0;
}


/*
 * Generic enable/disable code: this just calls
 * down into the PIC-specific version for the actual
 * hardware disable after having gotten the irq
 * controller lock. 
 */
 
/**
 *	disable_irq_nosync - disable an irq without waiting
 *	@irq: Interrupt to disable
 *
 *	Disable the selected interrupt line.  Disables and Enables are
 *	nested.
 *	Unlike disable_irq(), this function does not ensure existing
 *	instances of the IRQ handler have completed before returning.
 *
 *	This function may be called from IRQ context.
 */
 
inline void disable_irq_nosync(unsigned int irq)
{
  printk("disable_irq_nosync: %d\n",irq);
  pits_disable_irq_nosync(irq);
  printk("done disable_irq_nosync\n");
}

/**
 *	disable_irq - disable an irq and wait for completion
 *	@irq: Interrupt to disable
 *
 *	Disable the selected interrupt line.  Enables and Disables are
 *	nested.
 *	This function waits for any pending IRQ handlers for this interrupt
 *	to complete before returning. If you use this function while
 *	holding a resource the IRQ handler may need you will deadlock.
 *
 *	This function may be called - with care - from IRQ context.
 */
 
void disable_irq(unsigned int irq)
{
  printk("disable_irq: %d\n",irq);
  pits_disable_irq(irq);
  printk("done disable_irq\n");
}

/**
 *	enable_irq - enable handling of an irq
 *	@irq: Interrupt to enable
 *
 *	Undoes the effect of one call to disable_irq().  If this
 *	matches the last disable, processing of interrupts on this
 *	IRQ line is re-enabled.
 *
 *	This function may be called from IRQ context.
 */
 
void enable_irq(unsigned int irq)
{
  printk("enable_irq:%d\n",irq);
  pits_enable_irq(irq);
  printk("done enable_irq\n");
}

/*
 * do_IRQ handles all normal device IRQ's (the special
 * SMP cross-CPU interrupts have their own specific
 * handlers).
 */
//asmlinkage unsigned int do_IRQ(struct pt_regs regs)

unsigned int do_IRQ(struct pt_regs regs)
{	
  printk("do_IRQ\n");
  return 0;
}




/**
 *	request_irq - allocate an interrupt line
 *	@irq: Interrupt line to allocate
 *	@handler: Function to be called when the IRQ occurs
 *	@irqflags: Interrupt type flags
 *	@devname: An ascii name for the claiming device
 *	@dev_id: A cookie passed back to the handler function
 *
 *	This call allocates interrupt resources and enables the
 *	interrupt line and IRQ handling. From the point this
 *	call is made your handler function may be invoked. Since
 *	your handler function must clear any interrupt the board 
 *	raises, you must take care both to initialise your hardware
 *	and to set up the interrupt handler in the right order.
 *
 *	Dev_id must be globally unique. Normally the address of the
 *	device data structure is used as the cookie. Since the handler
 *	receives this value it makes sense to use it.
 *
 *	If your interrupt is shared you must pass a non NULL dev_id
 *	as this is required when freeing the interrupt.
 *
 *	Flags:
 *
 *	SA_SHIRQ		Interrupt is shared
 *
 *	SA_INTERRUPT		Disable local interrupts while processing
 *
 *	SA_SAMPLE_RANDOM	The interrupt can be used for entropy
 *
 */
int request_irq(unsigned int irq, 
		void (*handler)(int, void *, struct pt_regs *),
		unsigned long irqflags, 
		const char * devname,
		void *dev_id)
{
  int ret = 0;
  //struct linux_irq_t *linux_irq;
  
  
  dev_info_t *dip = getGlobal() -> current_dip;
  //sun_linux_state *instance_state  = NULL;
  //int instance = ddi_get_instance(dip);
  DEBUGCMD(printk("Inside request irq irq - %d",irq));

  ret = addLinuxIrq(dip,irq,handler,irqflags,devname,dev_id);

  if (ret != DDI_SUCCESS)
    {
      printk("request_irq() : error adding irq handler!\n");
      
    }
  return ret;
}



/**
 *	free_irq - free an interrupt
 *	@irq: Interrupt line to free
 *	@dev_id: Device identity to free
 *
 *	Remove an interrupt handler. The handler is removed and if the
 *	interrupt line is no longer in use by any driver it is disabled.
 *	On a shared IRQ the caller must ensure the interrupt is disabled
 *	on the card it drives before calling this function. The function
 *	does not return until any executing interrupts for this IRQ
 *	have completed.
 *
 *	This function may be called from interrupt context. 
 *
 *	Bugs: Attempting to free an irq in a handler for the same irq hangs
 *	      the machine.
 */
 
void free_irq(unsigned int irq, void *dev_id)
{

  int ret;
  ret = removeLinuxIrq(irq,dev_id);
  if (ret != DDI_SUCCESS)
    {
      printk("free_irq() : error freeing irq! %d\n",ret);
    }
}

/*
 * IRQ autodetection code..
 *
 * This depends on the fact that any interrupt that
 * comes in on to an unassigned handler will get stuck
 * with "IRQ_WAITING" cleared and the interrupt
 * disabled.
 */

//static DECLARE_MUTEX(probe_sem);

/**
 *	probe_irq_on	- begin an interrupt autodetect
 *
 *	Commence probing for an interrupt. The interrupts are scanned
 *	and a mask of potential interrupt lines is returned.
 *
 */
 
unsigned long probe_irq_on(void)
{

  unsigned long val = 0;

  return val;
}

/*
 * Return a mask of triggered interrupts (this
 * can handle only legacy ISA interrupts).
 */
 
/**
 *	probe_irq_mask - scan a bitmap of interrupt lines
 *	@val:	mask of interrupts to consider
 *
 *	Scan the ISA bus interrupt lines and return a bitmap of
 *	active interrupts. The interrupt probe logic state is then
 *	returned to its previous value.
 *
 *	Note: we need to scan all the irq's even though we will
 *	only return ISA irq numbers - just so that we reset them
 *	all to a known state.
 */
unsigned int probe_irq_mask(unsigned long val)
{
  return val;
}

/*
 * Return the one interrupt that triggered (this can
 * handle any interrupt source).
 */

/**
 *	probe_irq_off	- end an interrupt autodetect
 *	@val: mask of potential interrupts (unused)
 *
 *	Scans the unused interrupt lines and returns the line which
 *	appears to have triggered the interrupt. If no interrupt was
 *	found then zero is returned. If more than one interrupt is
 *	found then minus the first candidate is returned to indicate
 *	their is doubt.
 *
 *	The interrupt probe logic state is returned to its previous
 *	value.
 *
 *	BUGS: When used in a module (which arguably shouldnt happen)
 *	nothing prevents two IRQ probe callers from overlapping. The
 *	results of this are non-optimal.
 */
 
int probe_irq_off(unsigned long val)
{
  return 0;
}

/* this was setup_x86_irq but it seems pretty generic */
int setup_irq(unsigned int irq, struct irqaction * new)
{
  return 0;
}

//static struct proc_dir_entry * root_irq_dir;
//static struct proc_dir_entry * irq_dir [NR_IRQS];

//#define HEX_DIGITS 8

//static unsigned int parse_hex_value (const char *buffer,
//		unsigned long count, unsigned long *ret)
//{
//	unsigned char hexnum [HEX_DIGITS];
//	unsigned long value;
//	int i;
//
//	if (!count)
//		return -EINVAL;
//	if (count > HEX_DIGITS)
//		count = HEX_DIGITS;
//	if (copy_from_user(hexnum, buffer, count))
//		return -EFAULT;
//
//	/*
//	 * Parse the first 8 characters as a hex string, any non-hex char
//	 * is end-of-string. '00e1', 'e1', '00E1', 'E1' are all the same.
//	 */
//	value = 0;
//
//	for (i = 0; i < count; i++) {
//		unsigned int c = hexnum[i];
//
//		switch (c) {
//			case '0' ... '9': c -= '0'; break;
//			case 'a' ... 'f': c -= 'a'-10; break;
//			case 'A' ... 'F': c -= 'A'-10; break;
//		default:
//			goto out;
//		}
//		value = (value << 4) | c;
//	}
//out:
//	*ret = value;
//	return 0;
//}


