#include <misc/pits_init.h>
#include <misc/pits_state.h>
#include <misc/globals.h>
#include <linux/tqueue.h>
#include <linux/blkdev.h>
#include <linux/blk.h>
#include <linux/completion.h>
//int getHardSect_Size(dev_t dev);
/*******************************************/
/*block_size(ldev) returns the block size  */
/*for a given linux device                 */
/*******************************************/
unsigned int block_size(kdev_t ldev)
{
	int retval = BLOCK_SIZE;
	int major = MAJOR(ldev);
	
	if (blksize_size[major]) 
	  {
	    int minor = MINOR(ldev);
	    if (blksize_size[major][minor])
	      {
		retval = blksize_size[major][minor];
	      }
	  }
	return retval;
}

/*******************************************/
/*get_blk_default_queue()                  */
/*returns the default request queue for    */
/*a given major                            */
/*******************************************/
request_queue_t *get_blk_default_queue(major_t lmajor)
{
  return (&(blk_dev[lmajor].request_queue));
}

/*******************************************/
/*blk_get_queue()                          */
/*returns the assigned request queue for   */
/*the given device.                        */
/*If if doesn't exist, return the default  */
/*queue for the major number               */
/*******************************************/
request_queue_t *blk_get_queue(kdev_t ldev)
{
  major_t major;
  request_queue_t *ans = NULL;
  //printk("inside blk_get_queue\n");
  //printk("getting major\n");
  major = getmajor(ldev);

  //printk("is blk_dev[%d] null?\n",major);
  if (blk_dev[major].queue == NULL)
    {
      //printk("yes, return default queue\n");
      ans = get_blk_default_queue(major);
    }
  else
    {
      //printk("no, return queue(ldev)\n");
      ans = blk_dev[major].queue(ldev);
    }
  return(ans);
}


/***********************************************/
/*blk_queue_headactive() : sets the head active*/
/*field in the queue                           */
/***********************************************/
void blk_queue_headactive(request_queue_t *q, int val)
{
  q -> head_active = val;
}


/************************************************/
/*pits_make_request() :                         */
/*make request function for the request queues  */
/************************************************/
int pits_make_request(request_queue_t *queue, int rw, struct buffer_head *bh)
{
  struct request *req = NULL;
  struct list_head *new_element = NULL;
  struct list_head *head = NULL;
  DEBUGCMD(struct request *test = NULL);

  sun_linux_state *instance_state = NULL;


  instance_state = ddi_get_soft_state(getSunStateP(),0);


  //printk("pits_make_request()\n");
  //printk("gettin io_request_lock\n");
  //spin_lock(&io_request_lock);
  //printk("pits_make_request() : entering queue mutex.\n");

  if (queue -> pits_init == FALSE)
    {
      printk("error queue not initialized!\n");
      return DDI_FAILURE;
    }

  //mutex_enter(&(queue -> mutex));

  //build request structure.
  DEBUGCMD(printk("getting free request\n"));

  if (instance_state -> block_req == NULL)
    {
      printk("block request null, allocate new..\n");
      instance_state -> block_req = kmalloc(sizeof(struct request),0);
      bzero(instance_state -> block_req,sizeof(struct request));
    }
  
  req = instance_state -> block_req;

  if (req == NULL)
    {
      printk("pits_make_request() : error allocating request structure!\n");
      mutex_exit(&(queue -> mutex));
      spin_unlock(&io_request_lock);
      return DDI_FAILURE;
    }

  //bzero(req,sizeof(struct request));
  new_element = &req -> queue;

  /*fill in the request structure */
  req -> rq_dev                = bh -> b_rdev;
  req -> cmd                   = rw;
  req -> errors                = 0;
  req -> hard_sector           = bh -> b_rsector;
  req -> sector                = bh -> b_rsector;
  req -> hard_nr_sectors       = bh -> nr_sectors;
  req -> nr_sectors            = bh -> nr_sectors;
  req -> current_nr_sectors    = bh -> nr_sectors;
  req -> nr_segments           = 1; /*Always 1 for a new request */
  req -> nr_hw_segments        = 1;
  req -> buffer                = bh -> b_data;
  req -> bh                    = bh;
  req -> bhtail                = bh;
  req -> q                     = queue;

  //get the head of the queue.
  head = &queue -> queue_head;
  //printk("adding to list\n");
  //add the request.
  INIT_LIST_HEAD(new_element);
  if (list_empty(head))
    {
      //printk("list empty, add request to list\n");
      list_add(new_element,head);
      //printk("added, try list entry\n");     
    
      DEBUGCMD(
      test = list_entry(head->next,struct request,queue);
      if (test == NULL)
	{
	  printk("pits_make_request() : NULL entry\n");
	  spin_unlock(&(io_request_lock));
	  //mutex_exit(&(queue -> mutex));
	  return DDI_FAILURE;
	}
      );
    }
  else
    {
      DEBUGCMD(printk("request queue not empty, add element\n"));
      list_add_tail(new_element,head);
      DEBUGCMD(
      test = list_entry(head -> prev,struct request,queue);
      if (test == NULL)
	{
	  printk("pits_make_request() : NULL entry 2\n");
	  //mutex_exit(&(queue -> mutex));
	  return DDI_FAILURE;
	}
      );
    }
  
  DEBUGCMD(
	   printk("req dev %lu cmd %s s %lu nr %lu\n",
		  req -> rq_dev,
		  req -> cmd ? "W":"R",
		  req -> sector,
		  req -> nr_sectors)
	     	   );
  
  if (queue -> head_active)
    {
      printk("pits_make_request() : head is still active, return\n");
      printk("  this shouldn't happen right now..\n");
      //mutex_exit(&(queue -> mutex));
      return DDI_SUCCESS;
    }
  
  
  queue -> head_active = TRUE;
  //printk("pits_make_request() : calling request_fn\n");
  queue -> request_fn(queue);
  //printk("pits_make_request() : after request_fn\n");

  queue -> head_active = FALSE;

  //mutex_exit(&(queue -> mutex));
  //printk("pits_make_request() : done.\n");
  return DDI_SUCCESS;
  //check to see if rq is active, if not, fire interrupt.
}


/************************************************/
/*end_that_request_first() :                    */
/************************************************/
int end_that_request_first(struct request *req, int uptodate, char *name)
{
  struct buffer_head *bh;
  //int nsect;

  req -> errors = 0;

  //printk("end_that_request_first\n");

  if (!uptodate)
    {
      printk("end request: I/O error\n");
    }
  
  bh = req -> bh;

  if (bh != NULL)
    {
      req -> bh = bh -> b_reqnext;
      bh -> b_reqnext = NULL;
      bh -> b_end_io(bh,uptodate);

      bh = req -> bh;
      if (bh != NULL)
	{
	  printk("This shouldn't happen yet.\n");
	}
      
    }
  else
    {
      printk("end_that_request_first() : null bh!\n");
    }
  return 0;

}


/******************************************************/
/*blkdev_release_request()                            */
/******************************************************/
void blkdev_release_request(struct request *req)
{
  request_queue_t *q = NULL;


  int rw;
  /*
  sun_linux_state *instance_state = NULL;


  instance_state = ddi_get_soft_state(getSunStateP(),0);
  */
  //printk("inside blkdev_release_request()\n");
  
  if (req == NULL)
    {
      printk("blkdev_release_request() :NULL request\n");
      return;
    }

  /*
  q = req -> q;
  rw = req -> cmd;

  req -> rq_status = RQ_INACTIVE;
  */
  //we are done with this request, free it.
  //printk("freeing request\n");
  //kmem_cache_free(instance_state ->req_cache,req);
  //kmem_free(req,sizeof(struct request));
  //kfree(req);
  //cv_broadcast(instance_state -> blist_cv);
  //printk("done blkdev_release_request()\n");

}

/******************************************************/
/*end_that_request_last() :                           */
/******************************************************/
void end_that_request_last(struct request *req)
{
  //printk("end_that_request_last\n");
  
  if (req -> waiting != NULL)
    {
      complete(req -> waiting);
    }
  blkdev_release_request(req);
}


static void generic_plug_device(request_queue_t *q, kdev_t dev)
{
  //printk("generic_plug_device\n");
}

 /** 
  * blk_init_queue  - prepare a request queue for use with a block device 
  * @q:    The &request_queue_t to be initialised 
  * @rfn:  The function to be called to process requests that have been 
  *        placed on the queue. 
  * 
  * Description: 
  *    If a block device wishes to use the standard request handling procedures, 
  *    which sorts requests and coalesces adjacent requests, then it must 
  *    call blk_init_queue().  The function @rfn will be called when there 
  *    are requests on the queue that need to be processed.  If the device 
  *    supports plugging, then @rfn may not be called immediately when requests 
  *    are available on the queue, but may be called at some time later instead. 
  *    Plugged queues are generally unplugged when a buffer belonging to one 
  *    of the requests on the queue is needed, or due to memory pressure. 
  * 
  *    @rfn is not required, or even expected, to remove all requests off the 
  *    queue, but only as many as it can handle at a time.  If it does leave 
  *    requests on the queue, it is responsible for arranging that the requests 
  *    get dealt with eventually. 
  * 
  *    A global spin lock $io_request_lock must be held while manipulating the 
  *    requests on the request queue. 
  * 
  *    The request on the head of the queue is by default assumed to be 
  *    potentially active, and it is not considered for re-ordering or merging 
  *    whenever the given queue is unplugged. This behaviour can be changed with 
  *    blk_queue_headactive(). 
  * Note: 
  *    blk_init_queue() must be paired with a blk_cleanup_queue() call 
  *    when the block device is deactivated (such as at module unload). 
  **/
 
void blk_init_queue(request_queue_t * q, request_fn_proc * rfn) 
{ 

  //dev_info_t *dip = getGlobal() -> current_dip;
  //int status;
  printk("inside blk_init_queue\n");

  
  INIT_LIST_HEAD(&q->queue_head); 
  //  elevator_init(&q->elevator, ELEVATOR_LINUS); 
  //blk_init_free_list(q); 
  q->request_fn           = rfn; 
  //q->back_merge_fn        = ll_back_merge_fn; 
  //q->front_merge_fn       = ll_front_merge_fn; 
  //q->merge_requests_fn    = ll_merge_requests_fn; 
  q->make_request_fn      = pits_make_request; 
  q->plug_tq.sync         = 0; 
  q->plug_tq.routine      = &generic_unplug_device; 
  q->plug_tq.data         = q; 
  q->plugged              = 0; 
  /* 
   * These booleans describe the queue properties.  We set the 
   * default (and most common) values here.  Other drivers can 
   * use the appropriate functions to alter the queue properties. 
   * as appropriate. 
   */
 
  q->plug_device_fn       = generic_plug_device; 
  q->head_active          = FALSE; 

 
  //pits initialization stuff.
  //
  q -> queue_lock.locked = FALSE;
  pits_add_spin_lock(&(q -> queue_lock));
  printk("adding queue mutex\n");

  bzero(&(q -> mutex),sizeof(kmutex_t));
  mutex_init(&(q -> mutex),
	     NULL,
	     MUTEX_DRIVER,
	     NULL);
  q -> pits_init = TRUE;

  printk("done blk_init_queue\n");
} 
 



void blk_cleanup_queue(request_queue_t *request_queue)
{
  printk("inside blk_cleanup_queue\n");
  if (request_queue -> pits_init)
    {

      /*
      ddi_remove_softintr(request_queue -> pits_interrupt_id);
      */
      
      //mutex_destroy(&(request_queue -> queue_lock));
      //pits_
    }
  printk("done blk_cleanup_queue\n");
  ;
}



void blk_queue_make_request(request_queue_t *q, make_request_fn *make_request_function)
{
  printk("inside blk_queue_make_request()\n");
  q->make_request_fn = make_request_function; 
  printk("done blk_queue_make_request()\n");
  ;
}


/*****************************************************
 *register_disk: look in linux/fs/partitions/check.c *
 *****************************************************/
void register_disk(struct gendisk *dev, kdev_t first, unsigned minors, struct block_device_operations *ops, long size)
{
  printk("inside register disk\n");
  if (dev == NULL)
    {
      return;
    }
  grok_partitions(dev, MINOR(first)>>dev -> minor_shift,minors,size);
  printk("done register disk\n");

}

#ifndef LINUX_REQUEST
void generic_unplug_device(void *a)
{
  //printk("inside generic_unplug_device\n");
  ;
  //printk("done generic_unplug_device\n");
}
#endif

int init_block_arrays(dev_info_t *dip)
{
  int i,status;
  //  sun_linux_state *instance_state = NULL;
  //struct blk_dev_struct *test = blk_dev;
  
  status = 0;

  gendisk_head = NULL;
  
  //blk_size = kmalloc(sizeof(int*)*MAX_BLKDEV,0);
  //blksize_size = kmalloc(sizeof(int*)*MAX_BLKDEV,0);
  //hardsect_size = kmalloc(sizeof(int*)*MAX_BLKDEV,0);
  //max_readahead = kmalloc(sizeof(int*)*MAX_BLKDEV,0);
  //getGlobal() -> _max_sectors   = kmalloc(sizeof(int*)*MAX_BLKDEV,0);
  //getGlobal() -> _max_segments  = kmalloc(sizeof(int*)*MAX_BLKDEV,0);
  //blk_sec = kmalloc(sizeof(struct sec_size*)*MAX_BLKDEV,0);

  getGlobal() -> _gendisk_array = kmalloc(sizeof(struct gendisk) * MAX_BLKDEV,0);
  getGlobal() -> _gendisk_head = NULL;
  
  //blk_dev = kmalloc(sizeof(struct blk_dev_struct) * MAX_BLKDEV,0);
  //bzero(blk_dev,sizeof(struct blk_dev_struct) * MAX_BLKDEV);

  for (i=0;i<MAX_BLKDEV;i++)
    {
      //blk_size[i] = NULL;
      //blksize_size[i] = NULL;
      //hardsect_size[i] = NULL;
      //max_readahead[i] = NULL;
      //getGlobal() -> _max_sectors[i] = NULL;
      
      // getGlobal() -> _max_segments[i] = 0;
      //getGlobal() -> _gendisk_array[i] = NULL;

      //read_ahead[i] = 0;
      
      //memset(&(blk_dev[i]),'\0',sizeof(struct blk_dev_struct));

    }


  set_initialized(dip,BLOCK_INIT);

  

  return (DDI_SUCCESS);
}

int fini_block_arrays(dev_info_t *dip)
{

  sun_linux_state *instance_state = NULL;

  //kfree(getGlobal() -> _blk_size);
  //kfree(getGlobal() -> _blksize_size);
  //kfree(getGlobal() -> _hardsect_size);
  //kfree(getGlobal() -> _max_readahead);
  //kfree(getGlobal() -> _max_sectors);
  //kfree(getGlobal() -> _max_segments);
  kfree(getGlobal() -> _gendisk_array);
  //kfree(blk_sec);
  
  //kfree(blk_dev);
  
  getInstanceFReturn(dip,instance_state);
  


  



  kfree(instance_state -> blk_default_queue);
  

  if (get_initialized(dip) & BLOCK_INIT)
    {
      unset_initialized(dip,BLOCK_INIT);
    }
  return DDI_SUCCESS;

}
