#include <misc/pits_block.h>
#include <misc/pits_dev_inst.h>
#include <misc/pits_block_ioctl.h>



/*sunlinuxdrv include*/
//#include <misc/sunlinuxdrv.h>
#include <misc/instminor.h>
#include <misc/tasklets.h>
#include <misc/pits_init.h>
#include <misc/pits_param.h>
#include <misc/pits_block.h>
#include <misc/pits_char.h>

#include <linux/hdreg.h> /*HDIO_GETGEO*/
#include <linux/blkdev.h>
#include <linux/genhd.h>


/*sun include*/
#include <sys/types.h>
/*#include <sys/file.h>*/
#include <sys/errno.h>
#include <sys/open.h>
#include <sys/cred.h>


//for dkio ioctl calls. (man dkio(7)).
#include <sys/dkio.h>
#include <sys/vtoc.h>


/*
int ioctl_dkiocinfo(dev_t dev,int cmd, intptr_t arg, 
		    int mode, cred_t *credp, int *rvalp);

int ioctl_dkiocggeom(dev_t dev,int cmd, intptr_t arg, 
		     int mode, cred_t *credp, int *rvalp);

int ioctl_dkiocgvtoc(dev_t dev,int cmd, intptr_t arg, 
		     int mode, cred_t *credp, int *rvalp);
*/
int getBlkSize_Size(dev_t dev);
int getBlkSize(dev_t dev);
int getHardSect_Size(dev_t dev);


int pits_block_ioctl(dev_t sdev,int cmd, intptr_t arg, 
			   int mode, cred_t *credp, int *rvalp)
{
  sun_linux_state *instance_state = NULL;
  int minor;
  int instance = 0;
  int cmd_claimed;
  //int status;
  major_t lmajor,lminor;
  //struct hd_geometry linux_geo;
  //struct dk_geom     pits_geo;
  //struct vtoc        vtoc;
  //  struct dk_cinfo    dkc;
  linux_file *open_file = NULL;
  linux_inode *open_inode = NULL;


  printk("inside pits_block_ioctl()\n");

  instance_state = ddi_get_soft_state(getSunStateP(),instance);

  if (instance_state == NULL)
    {
      printk("pits_block_ioctl():null instance\n");
      return ENXIO;
    }
  
  minor = getminor(sdev);
  lminor = getlminor(sdev);
  lmajor = getlmajor(sdev);
  open_file = getOpenFile(lmajor,lminor);
  open_inode = getOpenInode(lmajor,lminor);

  if (open_file == NULL || open_inode == NULL)
    {
      printk("pits_block_ioctl():null file or inode\n");
      return ENXIO;
    }

  cmd_claimed = TRUE;

  switch (cmd)
    {
      //disk block commands required by solaris.
    case DKIOCINFO:
      return(ioctl_dkiocinfo(sdev,cmd,arg,mode,credp,rvalp));
    case DKIOCGAPART:
      printk("DKIOCGAPART\n");
      //return a disk's partition map.
      break;
    case DKIOCGGEOM:
      //called first by newfs.
      //return a disk's geometry.
      printk("DKIOCGGEOM\n");
      return(ioctl_dkiocggeom(sdev,cmd,arg,mode,credp,rvalp));
    case DKIOCSGEOM:
      printk("DKIOCSGEOM\n");
      //set a disk's geometry.
      break;
    case DKIOCGVTOC:      
      //return volume table of contents.
      return(ioctl_dkiocgvtoc(sdev,cmd,arg,mode,credp,rvalp));
    case DKIOCSVTOC:
      printk("DKIOCSVTOC\n");
      //set's a disk's Volume Table of Contents.
      break;
    default:
      cmd_claimed = FALSE;
    }
  return (DDI_SUCCESS);

}


int ioctl_dkiocinfo(dev_t sdev,int cmd, intptr_t arg, 
			     int mode, cred_t *credp, int *rvalp)

{
  struct dk_cinfo dkc;
  
  printk("inside ioctl_dkiocinfo()\n");
  //return information describing disk controller.
  
  bzero(&dkc,sizeof(struct dk_cinfo));
  
  bcopy("PITS",dkc.dki_cname,5);
  dkc.dki_ctype = DKC_UNKNOWN;
  dkc.dki_flags = DKI_BAD144 | DKI_FMTVOL;
  dkc.dki_partition = 0;
  dkc.dki_maxtransfer = 1;
  
  if (copyout((caddr_t)&dkc,(caddr_t)arg,sizeof(struct dk_cinfo)) != 0)
    {
      printk("error copying dk_cinfo()\n");
      return (EFAULT);
    }
  printk("done ioctl_dkiocinfo()\n");
  return 0;

}

int getLinuxGeo(dev_t sdev,struct hd_geometry *linux_geo)
{
  sun_linux_state *instance_state = NULL;
  int instance = 0;
  minor_t lminor;
  major_t lmajor;
  int status;

  linux_file *open_file = NULL;
  linux_inode *open_inode = NULL;

  lmajor = getlmajor(sdev);
  lminor = getlminor(sdev);

  instance_state = ddi_get_soft_state(getSunStateP(),instance);
  if (instance_state == NULL)
    {
      printk("getLinuxGeo: null instance\n");
      status = -EINVAL;
      goto end;
    }
  
  open_file = getOpenFile(lmajor,lminor);
  open_inode = getOpenInode(lmajor,lminor);

  if (open_file == NULL || open_inode == NULL)
    {
      printk("getLinuxGeo:null open file or inode\n");
      status = -EINVAL;
      goto end;
    }

  status = instance_state -> bdops -> ioctl(open_inode,
					    open_file,
					    HDIO_GETGEO,
					    (unsigned long)linux_geo);
 end:
  if (status != 0)
    {
      printk("getLinuxGeo:error calling linux ioctl(HDIO_GETGEO): %d\n",status);
    }
  return status;
}

int ioctl_dkiocggeom(dev_t sdev,int cmd, intptr_t arg, 
			     int mode, cred_t *credp, int *rvalp)
{
  struct dk_geom pits_geo;
  struct hd_geometry linux_geo;
  int status=0;

  printk("DKIOCGGEOM\n");
  
  status = getLinuxGeo(sdev,&linux_geo);

  if (status == 0)
    {
      printk("got linux geo\n");
      printk("cylinders %d heads %d sectors %d start %d\n",
	     linux_geo.cylinders,
	     linux_geo.heads,
	     linux_geo.sectors,
	     linux_geo.start);
      bzero(&pits_geo,sizeof(struct dk_geom));
      pits_geo.dkg_nhead=linux_geo.heads;
      pits_geo.dkg_nsect=linux_geo.sectors;
      pits_geo.dkg_ncyl =linux_geo.cylinders;
      pits_geo.dkg_acyl = 0;
      pits_geo.dkg_pcyl = pits_geo.dkg_ncyl + pits_geo.dkg_acyl;
      pits_geo.dkg_bcyl = 0;
      pits_geo.dkg_intrlv = 0;
      pits_geo.dkg_rpm = 0;
      pits_geo.dkg_write_reinstruct = 0;
      pits_geo.dkg_read_reinstruct = 0;
      
      status = copyout((caddr_t)&pits_geo,(caddr_t)arg,sizeof(struct dk_geom));

      if (status != 0)
	printk("error copying geometry\n");

      
    }
  else
    {
      printk("error getting geo %d\n",status);
     
    }
  return status;
  //return disk's geometry.
}

struct gendisk *getLinuxGendisk(dev_t ldev)
{
  struct gendisk *ptr = gendisk_head;
  
  printk("inside getLinuxGendisk major %d minor %d\n",
	 getsmajor(ldev),getsminor(ldev));
  printk("lmajor :%d\n",ldev);
  while (ptr != NULL)
    {
      printk("ptr -> major:%d getmajor(dev):%d\n",ptr -> major,getmajor(ldev));
      if (ptr -> major == ldev)
	{
	  printk("found\n");
	  printk("done getLinuxGendisk()\n");
	  return ptr;
	}
      ptr = ptr -> next;
    }
  printk("not found\n");
  printk("done getLinuxGendisk()\n");
  return NULL;

}

int ioctl_dkiocgvtoc(dev_t sdev,int cmd, intptr_t arg, 
			     int mode, cred_t *credp, int *rvalp)
{

  //sun_linux_state *instance_state = NULL;
  int status;
  //int i;

  struct gendisk *linux_gendisk_head = NULL;
  struct hd_geometry  linux_geo;
  struct vtoc         vtoc;
  
  dev_t ldev = getldev(sdev);

  printk("inside ioctl_dkiocgvtoc()\n");
  bzero(&vtoc,sizeof(struct vtoc));  
  //returns a disk's Volume Table of Contents.
  printk("getting sector size\n");
  vtoc.v_sectorsz = getHardSect_Size(ldev); //sector size in bytes.
  vtoc.v_sanity = VTOC_SANE;
  vtoc.v_version = V_VERSION;
  bcopy("PITS",vtoc.v_volume,5);
  
  //get the genhd for this device.
  printk("calling getLinuxGendisk(dev)\n");
  linux_gendisk_head = getLinuxGendisk(ldev);
  
  
  //if (linux_gendisk_head == NULL)
    {
      //assume 1 partition.
      printk("assuming 1 partition\n");
      status = getLinuxGeo(sdev,&linux_geo);
      if (status == 0)
	{
	  vtoc.v_nparts = 1;
	  vtoc.v_part[0].p_tag = V_UNASSIGNED;
	  vtoc.v_part[0].p_flag = V_UNMNT;
	  vtoc.v_part[0].p_start = linux_geo.start; 
	  vtoc.v_part[0].p_size  = getBlkSize(ldev)*1024/DEV_BSIZE; 
	  //have to do some calculation.  p_size is in # of blocks.


	  status = ddi_copyout((caddr_t)&vtoc,(caddr_t)arg,sizeof(struct vtoc),mode);
	  if (status != 0)
	    {
	      printk("error during copyout:%d\n",status);
	    }
	}
      else
	{
	  printk("error during getGeo:%d\n",status);
	}
    }
    return status;

    /*
  else
    {
      vtoc.v_nparts = linux_gendisk_head -> max_p;
      for (i=0;i<vtoc.v_nparts;i++)
	{
	  vtoc.v_part[i].p_tag  = V_UNASSIGNED;
	  vtoc.v_part[i].p_flag = V_UNMNT;
	  vtoc.v_part[i].p_start= linux_gendisk_head -> part[i].start_sect;
	  vtoc.v_part[i].p_size = linux_gendisk_head -> part[i].nr_sects * getHardSect_Size(ldev)/DEV_BSIZE;
	}
      ;
    }
    */
}
