/******************************************************/
/*pits_block.c : contains the pits_block entry points */
/*for block drivers                                   */
/******************************************************/

/*PITS includes*/
#include <misc/pits_block.h>
#include <misc/instminor.h>
#include <misc/tasklets.h>
#include <misc/pits_init.h>
#include <misc/pits_param.h>
#include <misc/pits_block.h>
#include <misc/pits_char.h>
#include <misc/pits_dev_inst.h>

/*Linux Include*/
#include <linux/hdreg.h> /*HDIO_GETGEO*/
#include <linux/blkdev.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/fcntl.h>
#include <linux/poll.h>

/*SUN Include*/
#include <sys/types.h>
#include <sys/errno.h>
#include <sys/open.h>
#include <sys/cred.h>
/*for dkio ioctl calls. (man dkio(7)).*/
#include <sys/dkio.h>
#include <sys/vtoc.h>
#include <sys/ddi.h>
#include <sys/sunddi.h>
#include <stdarg.h>  /*for va_list */ 
#include <sys/signal.h>
#include <unistd.h>
#include <sys/stropts.h>

#ifndef FEXCL
#define FEXCL 0x0400
#endif

//#define MAX_MAJOR 256

void pits_biodone(struct buf *sun_bp);


//Global Static variables.

pits_bh_queue_t pits_bh_queue;

typedef struct bp_queue_
{
  int size;
  struct buf **array;
  int hptr;
  int tptr;
  kcondvar_t cv;
  kmutex_t mutex;
}bp_queue_t;


static bp_queue_t *bp_queue;

bp_queue_t *create_queue_array(int size)
{
  bp_queue_t *ba = kmem_alloc(sizeof(bp_queue_t),KM_SLEEP);

  ba -> size = size;
  ba -> array = kmem_alloc(sizeof(struct buf*)*size,KM_SLEEP);
  ba -> hptr = 0;

  ba -> tptr = 0;
  mutex_init(&(ba -> mutex),NULL,MUTEX_DRIVER,NULL);
  cv_init(&(ba -> cv),NULL,CV_DRIVER,NULL);
  
  return ba;
}

void enter_queue_mutex(bp_queue_t *bq)
{
  mutex_enter(&(bq -> mutex));
}
void exit_queue_mutex(bp_queue_t *bq)
{
  mutex_exit(&(bq -> mutex));
}

void destroy_queue_array(bp_queue_t *ba)
{
  cv_destroy(&(ba -> cv));
  mutex_destroy(&(ba -> mutex));
  kmem_free(ba -> array,sizeof(struct buf*)*ba -> size);
  
  kmem_free(ba,sizeof(bp_queue_t));
}

int bq_full(bp_queue_t *bq)
{
  int tptr = bq -> tptr;
  int hptr = bq -> hptr;
  int size = bq -> size;

  if (tptr == 0 && hptr == size - 1)
    {
      return TRUE;
    }
  if (tptr-hptr == 1)
    {
      return TRUE;
    }
  return FALSE;

}

int bq_empty(bp_queue_t *bq)
{
  return (bq -> tptr == bq -> hptr);
}


int enqueue_bp(bp_queue_t *bq,struct buf *bp)
{
  if (bq_full(bq) == TRUE)
    {
      return DDI_FAILURE;
    }
  bq -> array[bq -> hptr] = bp;
  bq -> hptr ++;
  if (bq -> hptr >= bq -> size)
    {
      bq -> hptr = 0;
    }
  return DDI_SUCCESS;
}

struct buf *dequeue_bp(bp_queue_t *bq)
{
  struct buf *ans;
  if (bq_empty(bq)) 
    {
      return NULL;
    }
  
  ans = bq -> array[bq -> tptr];
  bq -> tptr++;
  if (bq -> tptr >= bq -> size)
    {
      bq -> tptr = 0;
    }

  cv_signal(&(bq -> cv));
  return ans;


}




/**********************************/
/*getBlkSize() : returns the block*/
/*size for a block device.        */
/**********************************/
int getBlkSize(dev_t ldev)
{
  major_t major = getmajor(ldev);
  minor_t minor = getminor(ldev);

  if (blk_size[major] == NULL) 
    {
      return (BLOCK_SIZE);
    }
  return(blk_size[major][minor]);
}
/**********************************/
/*getBlkSize_Size()               */
/**********************************/
int getBlkSize_Size(dev_t ldev)
{
  major_t major = getmajor(ldev);
  minor_t minor = getminor(ldev);
  //should be BLOCK_SIZE
  if (blksize_size[major] == NULL) 
    {
      return(BLOCK_SIZE);
    }
  return(blksize_size[major][minor]);
}
/**********************************/
/*getHardSect_Size()              */
/**********************************/
int getHardSect_Size(dev_t ldev)
{
  major_t major = getmajor(ldev);
  minor_t minor = getminor(ldev);

  if (hardsect_size[major] == NULL) 
    {
      printk("hardsect_size[%d] is null, return %d\n",major,BLOCK_SIZE);
      return BLOCK_SIZE;
    }
  return(hardsect_size[major][minor]);
}

int pits_bh_constructor(void *buf,void *un,int kmflags)
{
  printk("inside pits_bh_constructor\n");
  return 0;
}
void pits_bh_destructor(void *buf,void *un)
{
  //return 0;
}

int pits_req_constructor(void *buf,void *un,int kmflags)
{
  struct request *req = (struct request *)buf;
  printk("inside pits_req_constructor\n");
  

  return (int)req;
}
void pits_req_destructor(void *buf,void *un)
{
  printk("inside pits_req_destructor\n");
  ;
}

/***********************************/
/*register_blkdev():               */
/*Registers a block device with    */
/*the system.                      */
/***********************************/
int register_blkdev(unsigned int major,const char *name,struct block_device_operations *bdops)
{
  sun_linux_state *instance_state; 
  int dmajor;
  
  printk("inside register_blkdev() major = %d name = %s\n",major,name);

  
  getInstanceFReturn(getGlobal() -> current_dip,instance_state);
  printk("after getInstanceFReturn\n");

  printk("assigning default bdops\n");
  initDefaultBdops(instance_state,bdops);

  instance_state -> bname = (char*)name;
  
  dmajor = ddi_name_to_major((char *)name);

  printk("dmajor: %d pits_major %d\n",dmajor,instance_state -> pits_major);
  /*
  if (dmajor < 0) 
    {
      
      if (instance_state -> pits_major <= 0)
	{
	  return (-1);
	}      
    }
  else
    {
      instance_state -> pits_major = dmajor;
    }
  */
  //add entry to the pdi_table.
  instance_state -> linux_major = register_lmajor(major);
  if (instance_state -> linux_major <= 0)
    {
      printk("register_blkdev() : error registering major!\n");
      finiDefaultBdops(instance_state);
      return -1;
    }
  instance_state -> blist = getrbuf(KM_NOSLEEP);
  instance_state -> block_start_running = FALSE;
  instance_state -> block_start_scheduled = FALSE;

  bp_queue = create_queue_array(100);
 
  #ifdef __PITS_BLOCK_DMA_BUF
  printk("allocating buffer..\n");
  instance_state -> block_dma_buf = kmalloc(65535,GFP_DMA);
  
  if (instance_state -> block_dma_buf == NULL)
    {
      printk("error allocating buffer!\n");
      return -1;
    }

  printk("done allocating buffer...\n");
  #endif


  if (instance_state -> blist == NULL)
    {
      printk("register_blkdev() : error allocating blist head!\n");
      finiDefaultBdops(instance_state);
      return -1;
    }
  
  instance_state -> block_linux_bh = kmem_zalloc(sizeof(struct buffer_head),KM_NOSLEEP);

  init_waitqueue_head(&(instance_state -> block_linux_bh -> b_wait));

  instance_state -> block_linux_bh -> b_data = instance_state -> block_dma_buf;

  instance_state -> block_req = kmem_zalloc(sizeof(struct request),KM_NOSLEEP);

  if (instance_state -> block_linux_bh == NULL ||
      instance_state -> block_req == NULL)
    {
      printk("error allocating bh or req struct!\n");
      return -1;
    }


  mutex_init(&(instance_state -> blist_mutex),NULL,MUTEX_DRIVER,NULL);
  cv_init(&(instance_state -> blist_cv),NULL,CV_DRIVER,NULL);
 
  instance_state -> blist -> av_forw=NULL;
  init_buffer_head_queue(&pits_bh_queue);
  if (major == 0) 
    {
      return instance_state -> linux_major;
    }
  else 
    {
      return 0;
    }
}
/**********************************************/
/*unregister_blkdev() :                       */
/*unregisters the device                      */
/**********************************************/
int unregister_blkdev(unsigned int major,
		      const char *name)
{
  int instance = 0;
  sun_linux_state *instance_state;

  instance_state = ddi_get_soft_state(getSunStateP(),instance); 

  destroy_queue_array(bp_queue);

  if (instance_state != NULL)
    {
      //DEBUGDMESG("deleteing default fops");
      finiDefaultBdops(instance_state);    
      if (instance_state -> blist) {freerbuf(instance_state -> blist);}
      mutex_destroy(&(instance_state -> blist_mutex));
      
      #ifdef __PITS_BLOCK_DMA_BUF
	kfree(instance_state -> block_dma_buf);
      #endif
     }
  else
    {
      printk("warning null instance state!\n");
    }
  return 0;
}

/*******************************/
/*cb_ops operations            */
/*for block devices            */
/*******************************/
int pits_block_open(dev_t *sdevp,int flags,int otyp,cred_t *credp)
{
  int instance = 0;
  linux_file *open_file = NULL; 
  linux_inode *open_inode = NULL; 
  minor_t sminor = getminor(*sdevp);
  major_t smajor = getmajor(*sdevp);
  sun_linux_state *instance_state;
  int error = 0;

  instance_state = ddi_get_soft_state(getSunStateP(),instance);
  if (instance_state == NULL)
    {
      return (ENXIO);
    }
  /*
   *only honor FEXCL.  If a regular open or a layered open is still
   *outstanding on the device, the exclusive open must fail.
   */

  if ((flags & FEXCL) && (instance_state -> open || instance_state -> nlayered))
    {
      return (EAGAIN);
    }
  
  switch (otyp)
    {
    case OTYP_LYR:
      instance_state -> nlayered++;
      break;
    case OTYP_BLK:
      instance_state -> open = 1;
      break;
    default:
      return (EINVAL);
    }

  if (getBdops(*sdevp) == NULL || getBdops(*sdevp) -> open == NULL)
    {
      printk("pits_block_open: bdops is not assigned\n");
      return(ENXIO);
    }
  
  open_inode = getOpenInodeS(smajor, 
			     sminor); 
  open_file  = getOpenFileS(smajor, 
			    sminor); 
 
  if (open_inode == NULL || open_file == NULL) 
    { 
  
      printk("creating inode and file major %d minor %d\n",smajor,sminor);
      open_inode = createOpenInodeS(smajor,sminor); 
      if (open_inode == NULL)
	{
	  printk("Error allocating open_inode!\n");
	  return ENXIO;
	}
      open_file  = createOpenFile(instance_state -> fops, 
                                  open_inode, 
                                  flags, 
                                  otyp); 
      if (open_file == NULL)
	{
	  printk("Error allocating open_file!\n");
	  destroyOpenInode(open_inode);
	  return ENXIO;
	  
	}
    } 
  
  if (!checkLinuxFile(open_file)) 
    { 
      printk("sun_open:open_file is not correct.\n"); 
      
      destroyOpenFile(open_file);
      destroyOpenInode(open_inode);
      return DDI_FAILURE; 
    } 
  else 
    { 
      addOpenFile(open_file);
      addOpenInode(open_inode);
      instance_state -> num_open++; 
      open_file -> usage_count++;
      open_inode -> usage_count++;
    } 
  error = getBdops(*sdevp) -> open(open_inode,open_file);
  return error;
}

int pits_block_close(dev_t sdev,int flag,int otyp,cred_t *credp)
{
  minor_t sminor; 
  major_t smajor;
  int instance; 
  sun_linux_state *instance_state;  
  linux_file *open_file;
  linux_inode *open_inode;
   
  sminor = getminor(sdev);
  smajor = getmajor(sdev);
  instance = 0;

  /* test for a valid instance */  
  if (instance < 0 || instance > 26)  
    {  
      DEBUGDMESG("Invalid instance, return EINVAL");  
      return (EINVAL);  
    }  
  /* Is the instance attached? */  
  instance_state = ddi_get_soft_state(getSunStateP(),instance);  
  if (instance_state == NULL)  
    {  
      DEBUGDMESG("device not attached, return ENXIO");  
      return (ENXIO);  
    }  
    
  /*verify otyp is correct for the device  */  
  //if (otyp == global_state.open_type)  
  //  {  
      //DEBUGDMESG("invalid open type for device, return EINVAL");  
      //DEBUGDMESG("not really, I don't know why this isnt working but disabled error return");
      /*return (EINVAL);*/  
  //  }  
 
  open_file = getOpenFileS(smajor,sminor);
  open_inode = getOpenInodeS(smajor,sminor);
  if (open_inode == NULL || open_file == NULL) 
    { 
      DEBUGDMESG("inode or file is null!"); 
      return (EINVAL); 
    } 
  if (open_inode -> usage_count == 0 ||
      open_file  -> usage_count == 0)
    {
      printk("already closed?");
      getBdops(sdev) -> release(open_inode,open_file);
      removeOpenFile(open_file);
      removeOpenInode(open_inode);
      destroyOpenInode(open_inode);
      destroyOpenFile(open_file);
      return DDI_SUCCESS;
    }
   
  /*call close, only if the usage count is 1?*/ 
  //DEBUGDMESG("Calling linux close..."); 
   
  open_inode -> usage_count--;
  open_file  -> usage_count--;

  if (open_inode -> usage_count == 0 || open_file -> usage_count == 0)
    {
      
      getBdops(sdev) -> release(open_inode,open_file); 
      removeOpenFile(open_file);
      removeOpenInode(open_inode);
      kfree(open_inode);
      destroyOpenFile(open_file);
    }
  return (DDI_SUCCESS); 

}

static void pits_block_minphys(struct buf *bp)
{
  //need a check to make sure the min phys.
  
  int curphys = 8192;

  if (bp -> b_bcount > curphys)
    {
      bp -> b_bcount = curphys;
    } 
  
  minphys(bp);
}

int pits_block_read(dev_t sdev,struct uio *uio_p,cred_t *credp)
{
  int instance =0;
  int ret = 0;
  sun_linux_state *instance_state = NULL;

  dev_t ldev = getldev(sdev);
  
  int size = getBlkSize(ldev) * getBlkSize_Size(ldev);

  //DEBUGCMD(printk("inside pits_block_read\n"));
  instance_state = ddi_get_soft_state(getSunStateP(),instance);
  if (instance_state == NULL)
    {
      printk("NULL state?\n");
      return (ENXIO);
    }
  if (size > 0) 
    {
      if (uio_p -> uio_offset >= size) 
	{
	  printk("bad value\n");
	  return (EINVAL);
	}
    }

  //return 0;
  ret = physio(pits_block_strategy,(struct buf *)NULL,sdev,B_READ,pits_block_minphys,uio_p);
  if (ret != DDI_SUCCESS)
    {
      printk("Error physio %d\n",ret);
    }
  return ret;
}

int pits_block_write(dev_t sdev,struct uio *uio_p,cred_t *credp)
{
  int instance = 0;
  sun_linux_state *instance_state = NULL;
  int ret;
  dev_t ldev = getldev(sdev);

  int size = getBlkSize(ldev) * getBlkSize_Size(ldev);

  //DEBUGCMD(printk("pits_block_write\n"));

  instance_state = ddi_get_soft_state(getSunStateP(),instance);
  if (instance_state == NULL)
    {
      printk("NULL STATE?\n");
      return (ENXIO);
    }

  if (size > 0) 
    {
      if (uio_p -> uio_offset >= size) 
	{
	  printk("bad size...\n");
	  return (EINVAL);
	}
    }
  
  
  ret = physio(pits_block_strategy, (struct buf *)NULL,sdev,B_WRITE,pits_block_minphys,uio_p);

  //DEBUGCMD(printk("done pits_block_write %d\n",ret));

  return ret;

}
int pits_block_chpool(dev_t sdev,short events,int anyyet,
			    short *reventsp,struct pollhead **phpp)
{
  return nodev();
}

static void pits_block_start(void *);

//static spinlock_t buffer_lock = SPIN_LOCK_UNLOCKED;

static int pits_submit_bh(struct buf*);

int pits_block_strategy(struct buf *bp)
{
  sun_linux_state *instance_state = NULL;

  int instance = 0;
  int call_start = FALSE;
  DEBUGCMD(printk("pits_block_strategy\n"));
  instance_state = ddi_get_soft_state(getSunStateP(),instance);
  if (instance_state == NULL) 
    {
      bioerror(bp,EINVAL);
      biodone(bp);
      return (0);
    }
  mutex_enter(&(instance_state -> blist_mutex));
  /*
  while (instance_state -> block_start_running == TRUE)
    {
      cv_wait(&(instance_state -> blist_cv),
	      &(instance_state -> blist_mutex));
    }
  instance_state -> block_start_running = TRUE;

  mutex_exit(&(instance_state -> blist_mutex));

  bp_mapin(bp);
  pits_submit_bh(bp);

  pits_biodone(bp);
  */

  

  while (bq_full(bp_queue) == TRUE)
    {
      DEBUGCMD(printk("queue full, wait...\n"));
      cv_wait(&(bp_queue -> cv),&(instance_state -> blist_mutex));
      
    }
  enqueue_bp(bp_queue,bp);

  if (instance_state -> block_start_running == FALSE &&
      instance_state -> block_start_scheduled == FALSE)
    {
      call_start = TRUE;
      instance_state -> block_start_scheduled = TRUE;
    }
  mutex_exit(&(instance_state -> blist_mutex));
  if (call_start)
    {
      pits_block_start(instance_state);
    }
  
  DEBUGCMD(printk("pits_strategy() : done pits_strategy()\n"));
  return 0;
}
  



void pits_bioerror(struct buf *sun_bp,int error)
{
  bioerror(sun_bp,error);
  //pits_biodone(sun_bp);
}

void pits_biodone(struct buf *sun_bp)
{
  sun_linux_state *instance_state = NULL;
  instance_state = ddi_get_soft_state(getSunStateP(),0);

  biodone(sun_bp);
  /*
  mutex_enter(&(instance_state -> blist_mutex));
  instance_state -> block_start_running = FALSE;
  cv_broadcast(&(instance_state -> blist_cv));
  mutex_exit(&(instance_state -> blist_mutex));
  */
}

void pits_o_end(struct buffer_head *linux_bh,int uptodate)
{
  struct buf * sun_bp = linux_bh -> sun_buffer;
  sun_linux_state *instance_state = NULL;


  //printk("inside pits_o_end\n");
  instance_state = ddi_get_soft_state(getSunStateP(),0);

  
  //free the bh, we should save these to improve performance here...    
  //if (sun_bp ->  b_un.b_addr != linux_bh -> b_data)
    {
#ifdef __PITS_BLOCK_DMA_BUF
      //linux_bh -> b_data = 0;
#else
      kfree(linux_bh -> d_data);
#endif
      //linux_bh -> b_data = 0;
      //kfree(linux_bh -> b_data);
    }
  
    //linux_bh -> sun_buffer = NULL;
    //linux_bh -> b_end_io = NULL;  
  //kmem_cache_free(instance_state -> bh_cache,linux_bh);
  
  //kmem_free(linux_bh,sizeof(struct buffer_head));
  //kfree(linux_bh);
  if (uptodate == 0)
    {
      printk("not uptodate...\n");
      //pits_bioerror(sun_bp,EIO);
    }
  else
    {
      //printk("calling pits_biodone..\n");
      //pits_biodone(sun_bp);
      //printk("done calling pits_biodone...\n");
    }
  wake_up(&(linux_bh -> b_wait));
}

void pits_io_end(struct buffer_head *linux_bh, int uptodate)
{
  struct buf *sun_bp = linux_bh -> sun_buffer;
  //  sun_linux_state *instance_state = NULL;
  request_queue_t *linux_queue = NULL;
  void* sun_addr;
  void* linux_addr;

  //printk("inside pits_io_end()\n");
   if (uptodate == 0)
    {
      pits_o_end(linux_bh,uptodate);
    }
  else
    {
      pits_o_end(linux_bh,uptodate);
    }
   
   /*
	{
	  //okay, we read in the offseting sector, now write whole thing back out after
	  //we change the required part of the buffer.
	  printk("B_WRITE() : overwriting and resubmitting....\n");
	  bcopy(sun_addr,linux_addr,sun_bp -> b_bcount);

	  linux_bh -> b_end_io = pits_o_end;
	  //printk("calling make_request_fn(,WRITE,)\n");
	  linux_queue = blk_get_queue(linux_bh -> b_rdev);
	  linux_queue -> make_request_fn(linux_queue,WRITE,linux_bh);
	  //enqueue_buffer_head(linux_bh,WRITE);
	}
   */
    
}


static int calcSectors(struct buffer_head *lbuf,
		       struct buf *sbuf,
		       unsigned long *start_sector,
		       unsigned long *n_sectors,
		       unsigned long *lbyte_offset,
		       int *need_read)
{
  
  dev_t ldev;
  int multiplier;
  int sector;
  int nr_sectors;
  
  *lbyte_offset = 0;
  
  if (lbuf == NULL || sbuf == NULL)
    {
      return DDI_FAILURE;
    }


  *need_read = FALSE;

  ldev = getldev(sbuf -> b_edev);
  
  multiplier = getHardSect_Size(ldev)/DEV_BSIZE;

  nr_sectors = sbuf -> b_bcount / DEV_BSIZE;

    
  sector = sbuf -> b_blkno;
  nr_sectors = sbuf -> b_bcount / DEV_BSIZE;
  if (sbuf -> b_bcount % DEV_BSIZE != 0)
    {
      *need_read = TRUE;
      nr_sectors++;
    }
  if (sector % multiplier != 0)
    {
      sector = sector - sector % multiplier;
      *lbyte_offset = DEV_BSIZE * (sector) % multiplier;
      nr_sectors = nr_sectors + (sector) % multiplier;
      *need_read = TRUE;
    }
  if (nr_sectors % multiplier != 0)
    {
      nr_sectors = nr_sectors + multiplier - (nr_sectors % multiplier);
      *need_read = TRUE;
    }

  *start_sector = sector;
  *n_sectors = nr_sectors;
  //*bsize = sbuf -> b_bcount + lbyte_offset;
  

  return DDI_SUCCESS;
  
}

static int already_read(struct buffer_head *lbuf,
			     struct buf *sbuf)
{
  
  dev_t ldev;
  unsigned long nr_sector;
  unsigned long l_start,l_end,s_start,s_end;
  int need_read;
  unsigned long lbyte_offset;

  if (lbuf == NULL)
    {
      return FALSE;
    }
  if (sbuf == NULL)
    {
      return FALSE;
    }

  ldev = getldev(sbuf -> b_edev);

  if (lbuf -> b_rdev != ldev)
    {
      return FALSE;
    }

  l_start = lbuf -> b_rsector;
  l_end   = lbuf -> b_rsector + lbuf -> nr_sectors;

  calcSectors(lbuf,
	      sbuf,
	      &s_start,
	      &nr_sector,
	      &lbyte_offset,
	      &need_read);
  
  s_end = s_start + nr_sector;


  
   
  if (s_start >= l_start &&
      s_end   <= l_end)
    {
      return TRUE;
    }
  return FALSE;

}


static int pits_cmpcpy(void *src,void *dest,int n)
{
  int count=0;
  int i;
  char *csrc = (char*)src;
  char *cdest = (char*)dest;


  for (i=0;i<n;i++)
    {
      if (cdest[i] != csrc[i])
	{
	  count++;
	  cdest[i] = csrc[i];
	}
    }
  return count;

}

static int already_written(struct buffer_head *lbuf,
			      struct buf *sbuf)

{

  void* sun_addr;
  void* linux_addr;

  if (already_read(lbuf,sbuf) == FALSE)
    {
      bcopy(sbuf -> b_un.b_addr,
	    lbuf -> b_data,
	    sbuf -> b_bcount);
      return FALSE;
    }

  
  sun_addr = (void*)sbuf -> b_un.b_addr; 
  linux_addr = (void *)(lbuf -> b_data + lbuf -> lbyte_offset); 
  //printk("lbuf -> lbyte_offset:%d\n",lbuf -> lbyte_offset);

  if (pits_cmpcpy(sun_addr,linux_addr,sbuf ->  b_bcount) == 0)
    {
      return TRUE;
    }
  else
    {
      return FALSE;
    }

}

struct buffer_head *getLinuxBuf()
{
  sun_linux_state *instance_state = NULL;

  instance_state = ddi_get_soft_state(getSunStateP(),0);
  if (instance_state == NULL)
    {
      return NULL;
    }

  return (instance_state -> block_linux_bh);
}

static int pits_submit_bh_read(struct buf *sbuf)
{

  void* sun_addr;
  void* linux_addr;
  int need_read;
  dev_t ldev;
  struct buffer_head *lbuf;
  request_queue_t *linux_queue;

  lbuf = getLinuxBuf();

  if (sbuf == NULL)
    {
      printk("null solaris buffer??\n");
      return DDI_FAILURE;
    }

  if (lbuf == NULL)
    {
      printk("null buffer!\n");
      pits_bioerror(sbuf,EINVAL);
      return DDI_FAILURE;
    }
  


  /*
  
  if (already_read(lbuf,sbuf) == TRUE)
    {
      printk("already read, copy buffer\n");
      sun_addr = (void*)sbuf -> b_un.b_addr;
      linux_addr = (void*)(lbuf -> b_data + lbuf -> lbyte_offset);
      bcopy(linux_addr,sun_addr,sbuf -> b_bcount);
      //pits_biodone(sbuf);
      return DDI_SUCCESS;
    }
  else
  */
  
  DEBUGCMD(printk("read: ls %d le %d ss %d se %d lbo %d\n",
	 l_start,
	 l_end,
	 s_start,
	 s_end,
         lbyte_offset));

    {
      ldev = getldev(sbuf -> b_edev);
      linux_queue = blk_get_queue(ldev);
      if (linux_queue == NULL || linux_queue -> make_request_fn == NULL)
	{
	  printk("NULL request queue for make_request function!\n");
	  //pits_bioerror(sbuf,EINVAL);
	  return DDI_FAILURE;
	}

      //perform read.
      lbuf -> b_rdev = ldev;
      lbuf -> b_end_io = pits_io_end;
      lbuf -> sun_buffer = sbuf;
      
      calcSectors(lbuf,
		  sbuf,
		  &(lbuf -> b_rsector),/*&sector,*/
		  &(lbuf -> nr_sectors),/*&nr_sectors*/
		  &(lbuf -> lbyte_offset),/*&lbyte_offset*/
		  &need_read);

      
  DEBUGCMD(printk("read: ls %d le %d ss %d se %d lbo %d\n",
	 l_start,
	 l_end,
	 s_start,
	 s_end,
         lbyte_offset));


      lbuf -> b_size = sbuf -> b_bcount + lbuf -> lbyte_offset;

      if (linux_queue -> make_request_fn(linux_queue,
					 READ,
					 lbuf) == DDI_SUCCESS)
	{

	  interruptible_sleep_on(&(lbuf -> b_wait));
	  sun_addr = (void*)sbuf -> b_un.b_addr;
	  linux_addr = (void*)(lbuf -> b_data + lbuf -> lbyte_offset);

	  bcopy(linux_addr,sun_addr,sbuf -> b_bcount);
	  return DDI_SUCCESS;
	}
      
      else
	{
	  pits_bioerror(sbuf,EIO);
	  return DDI_FAILURE;
	}
    }
}
static int pits_submit_bh_write(struct buf *sbuf)
{
  //void *sun_addr;
  //void *linux_addr;
  int need_read;
  int direction;
  dev_t ldev;
  void *sun_addr;
  void *linux_addr;
  struct buffer_head *lbuf;
  request_queue_t *linux_queue;

  int status;

  lbuf = getLinuxBuf();

  if (lbuf == NULL)
    {
      printk("null buffer!\n");
      pits_bioerror(sbuf,EINVAL);
      return DDI_FAILURE;
    }
  
  if ((status = already_written(lbuf,sbuf)) == TRUE)
    {
      printk("already written\n");
      //pits_biodone(sbuf);
      return DDI_SUCCESS;
    }
  else
    {
      ldev = getldev(sbuf -> b_edev);
      linux_queue = blk_get_queue(ldev);
      if (linux_queue == NULL || linux_queue -> make_request_fn == NULL)
	{
	  printk("NULL request queue or make_request_function!\n");
	  pits_bioerror(sbuf, EINVAL);
	  return DDI_FAILURE;
	}
      lbuf -> b_rdev = ldev;
      lbuf -> b_end_io = pits_io_end;
      lbuf -> sun_buffer = sbuf;
      calcSectors(lbuf,
		  sbuf,
		  &(lbuf -> b_rsector),/*&sector,*/
		  &(lbuf -> nr_sectors),/*&nr_sectors*/
		  &(lbuf -> lbyte_offset),/*&lbyte_offset*/
		  &need_read);

      lbuf -> b_size = sbuf -> b_bcount + lbuf -> lbyte_offset;
      
      if (need_read == FALSE)
	{
	  //printk("we can go ahead and write it...\n");
	  lbuf -> b_end_io = pits_o_end;
	  direction = WRITE;
	}
      else
	{
	  printk("need to read first...\n");
	  lbuf -> b_end_io = pits_io_end;
	  direction = READ;

	  if (linux_queue -> make_request_fn(linux_queue,
					     READ,
					     lbuf) != DDI_SUCCESS)
	    {
	      pits_bioerror(sbuf,EIO);
	      return DDI_FAILURE;
	    }
	  interruptible_sleep_on(&(lbuf -> b_wait));
	  sun_addr = (void*)sbuf -> b_un.b_addr;
	  linux_addr = (void*)(lbuf -> b_data + lbuf -> lbyte_offset);
	  bcopy(sun_addr,linux_addr,sbuf -> b_bcount);
	}

      if (linux_queue -> make_request_fn(linux_queue,
					 direction,
					 lbuf) == DDI_SUCCESS)
	{
	  interruptible_sleep_on(&(lbuf -> b_wait));
	  return DDI_SUCCESS;
	}
      else
	{
	  return DDI_FAILURE;
	}
    }
}



static int pits_submit_bh(struct buf *sun_bp)
{
  
  if (sun_bp -> b_flags & B_READ)
    {
      return(pits_submit_bh_read(sun_bp));
    }
  else
    {
      return(pits_submit_bh_write(sun_bp));
    }
}



int init_buffer_head_queue(pits_bh_queue_t *pqueue)
{

  printk("inside init_buffer_head_queue()\n");

  
  bzero(&(pqueue -> rqueue.mutex),sizeof(kmutex_t));
  mutex_init(&(pqueue -> rqueue.mutex),NULL,MUTEX_DRIVER,NULL);
  bzero(&(pqueue -> wqueue.mutex),sizeof(kmutex_t));
  mutex_init(&(pqueue -> wqueue.mutex),NULL,MUTEX_DRIVER,NULL);
  
  pqueue -> rqueue.direction = READ;
  pqueue -> rqueue.scheduled = FALSE;
  pqueue -> rqueue.running = FALSE;
  pqueue -> rqueue.head = NULL;

  pqueue -> wqueue.direction = WRITE;
  pqueue -> wqueue.scheduled = FALSE;
  pqueue -> wqueue.running = FALSE;
  pqueue -> wqueue.head = NULL;
  tasklet_init(&(pqueue -> rqueue.qtasklet),
	       pits_block_tasklet,
	       (unsigned long)READ);

  tasklet_init(&(pqueue -> wqueue.qtasklet),
	       pits_block_tasklet,
	       (unsigned long)WRITE);
  return DDI_SUCCESS;
}

static void __enqueue_buffer_head(buffer_head_queue_t *queue,
				  struct buffer_head *bh)
{
  struct buffer_head *ptr = NULL;
  int need_int = FALSE;

  if (queue == NULL)
    {
      printk("warning null queue!\n");
      return;
    }
  if (bh == NULL)
    {
      printk("warning null bh!\n");
      return;
    }
  printk("__enqueue_buffer_head: grabbing mutex..");
  mutex_enter(&(queue -> mutex));
  printk("__enqueue_buffer_head: got mutex..\n");
  if (queue -> head == NULL)
    {
      queue -> head = bh;
    }
  else
    {
      ptr = queue -> head;
      while (ptr -> p_next != NULL)
	{
	  ptr = ptr -> p_next;
	}
      ptr -> p_next = bh;
    }
  if (queue -> scheduled == FALSE && queue -> running == FALSE)
    {
      queue -> scheduled = TRUE;
      need_int = TRUE;
    }
  mutex_exit(&(queue -> mutex));
  printk("released mutex..\n");
  if (need_int)
    {
      printk("need a schedule....\n");
      tasklet_schedule(&(queue -> qtasklet));
    }


}
static void enqueue_buffer_head(struct buffer_head *linux_bh,
				int direction)
{
  printk("inside enqueue_buffer_head\n");
  switch(direction)
    {
    case READ:
      __enqueue_buffer_head(&(pits_bh_queue.rqueue),linux_bh);
      break;
    case WRITE:
      __enqueue_buffer_head(&(pits_bh_queue.wqueue),linux_bh);
      break;
    default:
      printk("this shouldn't happen!\n");
    }
  printk("done enqueue_buffer_head\n");
}

static void pits_block_tasklet(unsigned long args)
{
  struct buffer_head *ptr = NULL;
  request_queue_t *linux_queue = NULL;
  buffer_head_queue_t *queue = NULL;
  int need_rerun = FALSE;
  printk("pits_block_tasklet(): start");
  switch(args)
    {
    case READ:
      queue = &(pits_bh_queue.rqueue);
      break;
    case WRITE:
      queue = &(pits_bh_queue.wqueue);
      break;
    default:
      printk("warning bad queue\n");
    }


  if (queue == NULL) {printk("error null queue!\n");return;}

  mutex_enter(&(queue -> mutex));
  if (queue -> scheduled == TRUE)
    {
      if (queue -> running == TRUE)
	{
	  queue -> scheduled = FALSE;
	}
      else
	{
	  queue -> running = TRUE;
	  queue -> scheduled = FALSE;
	}
    }
  
  ptr = queue -> head;
  if (ptr != NULL)
    {
      queue -> head = ptr -> p_next;
    }
  else
    {
      queue -> running = FALSE;
      printk("pits_block_tasklet() : queue is empty!\n");
      mutex_exit(&(queue -> mutex));
      return;
    }
  mutex_exit(&(queue -> mutex));
  printk("b_rdev: major :%d minor %d\n",major(ptr -> b_rdev),minor(ptr -> b_rdev));
  linux_queue = blk_get_queue(ptr -> b_rdev);
  if (linux_queue == NULL)
    {
      printk("error Null queue!\n");
      return;
    }
  printk("calling make_request...\n");
  linux_queue -> make_request_fn(linux_queue,
				 queue -> direction,
				 ptr);
  printk("done calling make_request...\n");
  mutex_enter(&(queue -> mutex));
  if (queue -> head == NULL)
    {
      need_rerun = FALSE;
      queue -> running = FALSE;
    }
  else
    {
      need_rerun = TRUE;
    }
  mutex_exit(&(queue -> mutex));

  if (need_rerun)
    {
      printk("need_rerun....\n");
      tasklet_schedule(&(queue -> qtasklet));
    }
  
}



static void pits_block_start(void *arg)
{
  struct buf *current_bh;
  
  sun_linux_state *instance_state = (sun_linux_state*)arg;

  DEBUGCMD(printk("inside pits_block_start()\n"));

  //lock the blist and get the first bh.
  //enter mutex.

  mutex_enter(&(instance_state -> blist_mutex));
  //check to make sure we are not already running.
  if (instance_state -> block_start_running == TRUE)
    {
      printk("block start already running, return...\n");
      mutex_exit(&(instance_state -> blist_mutex));
      //spin_unlock(&buffer_lock);
      return;
    }
  
  instance_state -> block_start_scheduled = FALSE;
  instance_state -> block_start_running = TRUE;

  while (bq_empty(bp_queue) == FALSE)
    {
      current_bh = dequeue_bp(bp_queue);
      mutex_exit(&(instance_state -> blist_mutex));
      if (current_bh != NULL)
	{
	  bp_mapin(current_bh);
	  pits_submit_bh(current_bh);
	  pits_biodone(current_bh);
	  
	}
      else
	{
	  printk("????");
	}
      mutex_enter(&(instance_state -> blist_mutex));
    }
  instance_state -> block_start_running = FALSE;
  //exit mutex.
  //spin_unlock(&buffer_lock);
  mutex_exit(&(instance_state -> blist_mutex));
  DEBUGCMD(printk("done pits_block_start()\n"));
}




void initDefaultBdops(sun_linux_state *instance_state,
		      struct block_device_operations *bdops)
{
  instance_state -> bdops = kmem_alloc(sizeof (struct block_device_operations),KM_SLEEP);
  memcpy(instance_state -> bdops,bdops,sizeof(struct block_device_operations));
}
void finiDefaultBdops(sun_linux_state *instance_state)
{
  kmem_free(instance_state -> bdops,sizeof(struct block_device_operations));
  instance_state -> bdops = NULL;

}

struct block_device_operations *getBdops(kdev_t sdev)
{
  int major = getmajor(getldev(sdev));
  int instance = 0;
  sun_linux_state *instance_state;

  instance_state = ddi_get_soft_state(getSunStateP(),instance);

  if (instance_state == NULL) {return NULL;}

  if (instance_state -> linux_major != major) {return NULL;}

  return instance_state -> bdops;

  
}
