#include <misc/pits_char.h>
 
/*sunlinuxdrv include*/
//#include <misc/sunlinuxdrv.h>
#include <misc/instminor.h>
#include <misc/tasklets.h>
#include <misc/pits_init.h>
#include <misc/pits_param.h>
#include <misc/pits_block.h>
#include <misc/pits_char.h>

/*sun include*/
#include <sys/types.h>
/*#include <sys/file.h>*/
#include <sys/errno.h>
#include <sys/open.h>
#include <sys/cred.h>
#include <sys/ddi.h>
#include <sys/sunddi.h>
#include <stdarg.h>  /*for va_list */ 
#include <sys/signal.h>
#include <unistd.h>

#include <sys/stropts.h>

/*linux include*/
#include <linux/fs.h>
//#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/fcntl.h>
#include <linux/poll.h>


static u_int sun_pollwakeup(sun_linux_state *instance,short event);

//also want to define register_chrdev in the pits_char.c library file.
/*********************************************/
/*linux calls                                */
/*********************************************/
/*********************************************/
/*unregister_chrdev()                        */
/*********************************************/
int unregister_chrdev(unsigned int major, const char *name)
{
  //DEBUGDMESG("unregister_chrdev");
  //DEBUGDMESG3("major = %d name = %s\n",major,name); 
  int instance = 0;
  sun_linux_state *instance_state;

  instance_state = ddi_get_soft_state(getSunStateP(),instance); 

  if (instance_state != NULL)
    {
      //DEBUGDMESG("deleteing default fops");
      finiDefaultFops(instance_state);
    }
  //DEBUGDMESG("end unregister_chrdev");
  return 0;
}

/*********************************************/
/*register_chrdev()                          */
/*********************************************/
int register_chrdev(unsigned int major, const char *name, struct file_operations *fops)
{
  
  sun_linux_state *instance_state; 

  int dmajor;
  DEBUGDMESG("register_chrdev");
  DEBUGDMESG3("major = %d name = %s\n",major,name);

  getInstanceFReturn(getGlobal() -> current_dip,instance_state);

  DEBUGDMESG("assigning default fops");
  initDefaultFops(instance_state,fops);
  
  instance_state -> cname = (char*)name;
  
  dmajor = ddi_name_to_major((char *)name);

  printk("dmajor: %d pits_major %d\n",dmajor,instance_state -> pits_major);
  /*
  if (dmajor < 0)
    {
      if (instance_state -> pits_major <= 0)
	{
	  return (-1);
	}
    }
  else
    {
      instance_state -> pits_major = dmajor;
    }
  */
  //add entry to the pdi_table.
  instance_state -> linux_major = register_lmajor(major);
  if (instance_state -> linux_major <= 0)
    {
      printk("register_chrdev() : error registering major!\n");
      finiDefaultFops(instance_state);
      return -1;
    }
  if (major == 0) 
    { 
      /* want to assign a major number right? */
      return instance_state -> linux_major;
    } 
  else 
    { 
      return 0;
    } 
}



/*******************************/
/*cb_ops operations            */
/*******************************/
int pits_char_open(dev_t *devp,int flag,int otyp,cred_t *credp)
{
  
  int instance; 
  int error; 
  major_t smajor;
  minor_t sminor; 
  sun_linux_state *instance_state; 
  linux_file *open_file;
  linux_inode *open_inode;
  
  DEBUGDMESG("inside pits_char_open");
  smajor = getmajor(*devp);
  sminor = getminor(*devp); 
  instance = 0;    
 
  /* test for a valid instance */ 
  if (instance < 0 || instance > 16) 
    { 
      DEBUGDMESG("Invalid instance, return EINVAL"); 
      return (EINVAL); 
    } 
  /* Is the instance attached? */ 
  instance_state = ddi_get_soft_state(getSunStateP(),instance); 
  if (instance_state == NULL) 
    { 
      DEBUGDMESG("device not attached, return ENXIO"); 
      return (ENXIO); 
    }

  /*check for priveledge*/ 
  /* this is for root privledge only, commetted out*/ 
  /* 
  if ((flag & FWRITE) && drv_priv(credp) == EPERM) 
    { 
    printk("Open is for root only...\n");
    return (EPERM); 
    } 
  */ 
  /* call linux open */ 
  /*  int (*open) (struct inode *, struct file *); */ 
  /* we need to build the inode and file           */ 
 
  //DEBUGDMESG("creating inode and file structures"); 
  
  open_inode = getOpenInodeS(smajor,
			     sminor);
  open_file  = getOpenFileS(smajor,
			    sminor);

  if (open_inode == NULL)
    {
      open_inode = createOpenInodeS(smajor,sminor);
      open_file  = createOpenFileS(instance_state -> fops,
				  open_inode,
				  flag,
				  otyp);
      addOpenInode(open_inode);
      addOpenFile(open_file);
    }
  
  if (!checkLinuxFile(open_file))
    {
      printk("sun_open:open_file is not correct.");
      return DDI_FAILURE;
    }
  
  else
    {
      instance_state -> num_open++;
    }
  //DEBUGDMESG("Calling linux open code"); 
  error = open_file -> f_op -> open(open_inode,open_file); 
   
  /* fops may have been updated so, update the pointer to the correct fops. */   
  if (error == 0) 
    { 
      open_inode -> usage_count ++;
      open_file -> usage_count ++;

      return (DDI_SUCCESS); 
    } 
  else 
    { 
      /*error free the open_file and open_inode functions */ 
      DEBUGDMESG2("error calling linux open:%d",error); 
 
      removeOpenFile(open_file);
      removeOpenInode(open_inode);
      
      kfree(open_file); 
      kfree(open_inode); 
  
      //DEBUGDMESG("end sun_open()"); 
      return (DDI_FAILURE); 
    } 
}

int pits_char_close(dev_t sdev,int flag,int otyp,cred_t *credp)
{
  minor_t sminor;
  major_t smajor;
  int instance; 
  sun_linux_state *instance_state;  
  linux_file *open_file;
  linux_inode *open_inode;

  DEBUGDMESG("inside pits_char_close()");
  sminor = getminor(sdev);
  smajor = getmajor(sdev);
  instance = 0;

  /* test for a valid instance */  
  if (instance < 0 || instance > 26)  
    {  
      DEBUGDMESG("Invalid instance, return EINVAL");  
      return (EINVAL);  
    }  

  /* Is the instance attached? */  
  instance_state = ddi_get_soft_state(getSunStateP(),instance);  
  if (instance_state == NULL)  
    {  
      DEBUGDMESG("device not attached, return ENXIO");  
      return (ENXIO);  
    }  
    
  /*verify otyp is correct for the device  */  
  //if (otyp == global_state.open_type)  
  //  {  
      //DEBUGDMESG("invalid open type for device, return EINVAL");  
      //DEBUGDMESG("not really, I don't know why this isnt working but disabled error return");
      /*return (EINVAL);*/  
  //  }  
 
  open_file = getOpenFileS(smajor,sminor);
  open_inode = getOpenInodeS(smajor,sminor);
  

  DEBUGDMESG("testing for open state"); 
  if (open_inode == NULL || open_file == NULL) 
    { 
      printk("inode or file is null!"); 
      return (DDI_SUCCESS); 
    } 
  if (open_inode -> usage_count == 0 ||
      open_file  -> usage_count == 0)
    {
      printk("already closed?");
      open_file -> f_op -> release(open_inode,open_file);
      removeOpenFile(open_file);
      removeOpenInode(open_inode);
      kfree(open_inode);
      destroyOpenFile(open_file);
      return DDI_SUCCESS;
    }
   
  /*call close, only if the usage count is 1?*/ 
  open_inode -> usage_count--;
  open_file  -> usage_count--;

  if (open_inode -> usage_count == 0 || open_file -> usage_count == 0)
    {
      printk("usage count has dropped below 0, call linux release..\n");
      open_file -> f_op -> release(open_inode,open_file); 
      removeOpenFile(open_file);
      removeOpenInode(open_inode);
      kfree(open_inode);
      destroyOpenFile(open_file);
    }  
  return (DDI_SUCCESS); 
}

int pits_char_read(dev_t dev,struct uio *uio_p,cred_t *credp)
{
  //  offset_t off; 
  sun_linux_state *instance_state;

  linux_file *open_file;
  linux_inode *open_inode;

  int i;
  //char *buffer;
  size_t buffer_size;
  //  size_t count;

  int bytes_read = 0;
  int error;
  loff_t *current_position;
  off_t desired_position;
  caddr_t user_buffer;

  
  int minornum;
  int instance;

  DEBUGDMESG("inside pits_char_read()");

  minornum = getminor(dev);
  //instance = getInstance(minornum);
  instance = 0;
  if (instance <0 || instance > 16)
    {
      return (EINVAL);
    }
  
  
  instance_state = ddi_get_soft_state(getSunStateP(),instance);
  if (instance_state == NULL)
    {
      return(ENXIO);
    }
  
  open_file = getOpenFile(instance_state -> linux_major,minornum);
  open_inode = getOpenInode(instance_state -> linux_major,minornum);

  

  if (open_file == NULL || open_file -> usage_count == 0)
    {
      return(ENXIO);
    }
  

 //i don't know what to do here if a read read has not been defined...
  if (open_file -> f_op == NULL || open_file -> f_op -> read == NULL)
    {
      return(ENXIO);
    }

  /*set the linux file flags as directed.*/
  setLinuxFileFlags(open_file,uio_p -> uio_fmode);

  current_position = &(open_file -> f_pos); 
  desired_position = uio_p -> uio_offset;

  //DEBUGDMESG3("current:%d desired:%d",open_file->f_pos,desired_position);


  if (*current_position != desired_position)
    {
      /*call linux seek code*/
      if (open_file -> f_op -> llseek != NULL)
	{
	  //DEBUGDMESG("calling linux llseek call");
	  open_file -> f_op -> llseek(open_file,desired_position,SEEK_SET);
	  *current_position = desired_position;
	}
      else
	{
	  *current_position = desired_position;
	}
    }

  DEBUGDMESG2("current position is now:%d",*current_position);
  /* do the transfer */
  /* loop through each vector, copying the data out to the user program */
  
  //DEBUGDMESG2("uio:%d",uio_p -> uio_resid);

  for (i=0;i< uio_p -> uio_iovcnt;i++)
    {
      /*get size and start address of the user buffer */
      buffer_size = uio_p -> uio_iov[i].iov_len;
      user_buffer = uio_p -> uio_iov[i].iov_base;

      /*call linux read into the user buffer */
      error = open_file -> f_op -> read(open_file,
					user_buffer,
					buffer_size,
					current_position);
      if (error < 0)
	{
	  DEBUGDMESG2("error occurred %d",error);
	  goto error;
	}
      else
	{
	  uio_p -> uio_resid = uio_p -> uio_resid - error;
	  bytes_read += error;
	  //DEBUGDMESG("Calling pollwakeup -> POLLOUT");
	  sun_pollwakeup(instance_state,POLLOUT);
	}

    }
  
  //DEBUGDMESG2("new position is:%d",open_file -> f_pos);
  //DEBUGDMESG2("bytes read:%d",bytes_read);
  //DEBUGDMESG2("uio:%d",uio_p -> uio_resid);
  DEBUGDMESG("end sun_read()");

  

  return (0);
 error:
  return (-error);

}
int pits_char_write(dev_t dev,struct uio *uio_p,cred_t *credp)
{
  //offset_t off;  
  sun_linux_state *instance_state; 
 
  int i; 
  //  char *buffer; 
  size_t buffer_size; 
  //  size_t count; 
  int bytes_written = 0;
  int error;
  loff_t *current_position; 
 
  off_t desired_position; 
   
  caddr_t user_address; 
 
  linux_file *open_file; 
  int instance;
  minor_t sminor;
  major_t smajor;

  sminor = getminor(dev);
  smajor = getmajor(dev);

  DEBUGDMESG("inside pits_char_write()"); 

  //instance = getInstance(minornum);
  instance = 0;

  instance_state = ddi_get_soft_state(getSunStateP(),instance); 
  if (instance_state == NULL) 
    { 
      return(ENXIO); 
    } 

  open_file = getOpenFileS(smajor,sminor);

  if (open_file == NULL || open_file -> usage_count == 0) 
    { 
      return(ENXIO); 
    } 
 

  //DEBUGDMESG("setting file flags");
  setLinuxFileFlags(open_file,uio_p -> uio_fmode);

  //DEBUGDMESG("getting current position");
  current_position = &(open_file -> f_pos);  
  desired_position = uio_p -> uio_offset; 

  DEBUGDMESG3("current:%d desired:%d",open_file->f_pos,desired_position); 
 
  
  if (*current_position != desired_position) 
    { 
      /*call linux seek code*/ 
      if (open_file -> f_op -> llseek != NULL) 
        { 
          //DEBUGDMESG("calling linux llseek call"); 
          open_file -> f_op -> llseek(open_file,desired_position,SEEK_SET); 
          *current_position = desired_position; 
        } 
      else 
        { 
          *current_position = desired_position; 
        } 
    } 
 
   
  /* do the transfer */ 
  /* loop through each vector, copying the data out to the user program */ 
  
  //DEBUGDMESG2("number of vectors:%d",uio_p -> uio_iovcnt);
 
  for (i=0;i< uio_p -> uio_iovcnt;i++) 
    { 
      /*get size and start address of the user buffer */ 
      buffer_size = uio_p -> uio_iov[i].iov_len; 
      user_address = uio_p -> uio_iov[i].iov_base; 
 
      

      //DEBUGDMESG4("vector %d size %d address %lu",i,buffer_size,user_address);

      /*call linux read into the user buffer */ 
      if (open_file -> f_dentry == NULL)
	{
	  printk("f_dentry unallocated in write2!");
	}
      printk("calling linux write");
      error = open_file -> f_op -> write(open_file,
					      user_address,
					      buffer_size,
					      current_position); 
	
      if (error < 0)
	{
	  if (bytes_written == 0)
	    {
	      bytes_written = error;
	    }
	  goto error;
	}
      else
	{
	  uio_p -> uio_resid = uio_p -> uio_resid - error;
	  bytes_written += error;
	  sun_pollwakeup(instance_state,POLLIN);
	}

    } 
  //DEBUGDMESG2("new position is:%d",*current_position);
  //DEBUGDMESG2("new position is:%d",open_file -> f_pos); 
  //DEBUGDMESG2("bytes written:%d",bytes_written);
  //DEBUGDMESG2("left to write:%d",uio_p -> uio_resid);

  DEBUGDMESG("end sun_write()");
  return (0);
 error:
  DEBUGDMESG2("pits_char_write():Returning error: %d",-error);
  return (-error);

}
int pits_char_chpool(dev_t dev,short events,int anyyet,
			    short *reventsp,struct pollhead **phpp)
{
  return 0;
}

int pits_char_strategy(struct buf *bp)
{
  return nodev();
}

int pits_char_ioctl(dev_t dev,int cmd, intptr_t arg, 
			   int mode, cred_t *credp, int *rvalp)
{
    sun_linux_state *instance_state = NULL;
  int instance=0;
  int minor;
  //pid_t pid;
  //  int bitmask;
  int cmd_claimed = FALSE;
  int status = ENOTTY;
  linux_file *open_file;
  linux_inode *open_inode;

  DEBUGDMESG("Inside pits_char_ioctl");

  instance_state = ddi_get_soft_state(getSunStateP(),instance);
  if (instance_state == NULL)
    {
      return (ENXIO);
    }

  minor = getminor(dev);



  open_file = getOpenFile(instance_state -> linux_major,minor);
  open_inode = getOpenInode(instance_state -> linux_major,minor);
  /*grab creditantials and print out to screen.*/

  /*well,, try get pid;*/

  DEBUGDMESG2("pid is:%d",getpid());
  //DEBUGDMESG2("ppid is:%d",getppid());

  DEBUGDMESG2("cmd is:%d",cmd);

  switch(cmd)
    {
    case I_SETSIG:
      DEBUGDMESG("I_SET_SIG was called");
      if (open_file == NULL || open_file -> usage_count == 0) {return (ENXIO);}
      if (open_file -> f_op -> fasync == NULL)
	{
	  return (ENOTTY);
	}
      else
	{
	  //open_file = instance_state -> open_file;
	  DEBUGDMESG("Calling fasync");
	  open_file -> f_owner.pid = getpid();

	  status = open_file -> f_op -> fasync(1,open_file,arg);
	  DEBUGDMESG("done calling fasync");
	  cmd_claimed = TRUE;
	  return (status);
	}
      break;
      
    }
  if (cmd_claimed == FALSE && instance_state -> bdops != NULL)
    {
      printk("calling pits_block_ioctl");
      status = pits_block_ioctl(dev,cmd,arg,mode,credp,rvalp);
      if (status == DDI_SUCCESS)
	{
	  return status;
	} 
    }
  if (cmd_claimed == FALSE)
    {
      if (open_file != NULL && open_file -> f_op -> ioctl != NULL)
	{

	  DEBUGDMESG("Calling linux ioctl");
	  status = open_file -> f_op -> ioctl(open_inode,
					      open_file,
					      cmd,
					      arg);
	  DEBUGDMESG2("Done calling linux ioctl:%d",status);
	  *rvalp = status;
	  return (0);
	} 
      
    }
  DEBUGDMESG("Done pits_char_ioctl");
  return status;
}

/************************************************/
/*initDefaultFops() : initialize the default    */
/*file_opertations structure in the global_state*/
/*structure.                                    */
/************************************************/ 
void initDefaultFops(sun_linux_state *instance_state,
			    struct file_operations *fops) 
{ 
  instance_state -> fops = kmalloc(sizeof(struct file_operations),0);
  memcpy(instance_state -> fops,fops,sizeof(struct file_operations));

} 

void finiDefaultFops(sun_linux_state *instance_state)
{
  kfree(instance_state -> fops);
  instance_state -> fops = NULL;
  ;
}


static u_int sun_pollwakeup(sun_linux_state *instance,short event)
{
  sun_linux_state *instance_state = instance;

  

  if (event & POLLIN)
    {
      DEBUGDMESG("Calling Readable - pollin");
      pollwakeup(&instance_state -> phead,POLLIN);
      DEBUGDMESG("Calling Readable - pollrdnorm");
      pollwakeup(&instance_state -> phead,POLLRDNORM);
    }
  if (event & POLLOUT)
    {
      DEBUGDMESG("Calling Writeable");
      pollwakeup(&instance_state -> phead,POLLOUT);
    }
  else
    {
      DEBUGDMESG("Um, what happened? :");
    }

  DEBUGDMESG("Done wakeup call");
  return (DDI_INTR_CLAIMED);
}
