#include <asm/current.h>
#include <unistd.h>
#include <linux/kernel.h>
#include <misc/sched.h>


static struct task_struct *last_process=NULL;

struct task_struct_list
{
  kmutex_t mutex;
  struct list_head head;
};

static struct task_struct_list task_list;

int init_current()
{
  mutex_init(&(task_list.mutex),NULL,MUTEX_DRIVER,NULL);
  INIT_LIST_HEAD(&(task_list.head));
  return 0;
}
int fini_current()
{
  mutex_destroy(&(task_list.mutex));
  return 0;
}

void task_enter()
{
  struct task_struct *t = get_current();
  mutex_enter(&(t -> mutex));
  t -> pits_refcnt++;
  mutex_exit(&(t -> mutex));
}
void task_exit()
{
  struct task_struct *t = NULL;
  
  mutex_enter(&(task_list.mutex));
  t = task_find();
  
  if (t == NULL)
    {
      printk("task_find(): task is not found!\n");
      mutex_exit(&(task_list.mutex));
      return;
    }
  
  mutex_enter(&(t -> mutex));
  t -> pits_refcnt--;
  if (t -> pits_refcnt <= 0)
    {
      //mutex_enter(&(task_list.mutex));
      list_del(&(t -> list_entry));
      //mutex_exit(&(task_list.mutex));
      task_free(t);
    }
  mutex_exit(&(task_list.mutex));
}
void task_free(struct task_struct *t)
{
  mutex_destroy(&(t -> mutex));
  kmem_free(t,sizeof(struct task_struct));
}

struct task_struct *task_find()
{
  pid_t pid = getpid();
  struct list_head *ptr;
  int found;
  struct task_struct *ans;

  if (list_empty(&(task_list.head)))
    {return NULL;}
  
  ptr = &task_list.head;
  ptr = ptr -> next;
  found = FALSE;
  while (ptr != &task_list.head && found == FALSE)
    {
      ans = list_entry(ptr,struct task_struct,list_entry);
      if (ans -> pid == pid)
	{
	  found = TRUE;
	}
      else
	{
	  ptr = ptr -> next;
	}
    }
  if (found == TRUE)
    {
      return ans;
    }
  else
    {
      return NULL;
    }
  
}

struct task_struct *get_current(void)
{
  if (last_process == NULL)
    {
      printk("NULL process, create new one\n");
      last_process = kmalloc(sizeof(struct task_struct),0);
    }
  else if (last_process -> pid != getpid())
    {
      printk("pid is different return new process\n");
      last_process = kmalloc(sizeof(struct task_struct),0);
    }
  last_process -> uid = getuid();
  last_process -> pid = getpid();
  mutex_init(&(last_process -> mutex),NULL,MUTEX_DRIVER,NULL);
  last_process -> cv_wait = NULL;
  //  last_process -> proc_ref = proc_ref();
  //printk("current uid %d\n",last_process -> uid);
  return last_process;

}




int printk(const char *fmt,...)
{
  va_list args;
  va_start(args,fmt);
  vprintf(fmt,args);
  va_end(args);

  return(0);
}
