#include <misc/globals.h>
#include <misc/pits_dev_inst.h>



/*************************************************/
/*findlmajor() : given the linux major, return   */
/*the array index in the pdi_table               */
/*************************************************/
int findlmajor(major_t lmajor)
{
  int i;

  struct pits_dev_inst *table = getGlobal() -> pdi_table;

  for (i=0;i<table -> nlmajors;i++)
    {
      if (table -> lmajors[i] == lmajor)
	return i;
    }
  return -1;
}

/********************************************/
/*addlmajor() : adds a linux major number to*/
/*the pdi table if it doesn't already exist.*/
/*if it exists or the number of max majors  */
/*is exceeded, return -1.                   */
/********************************************/
int addlmajor(major_t lmajor)
{

  struct pits_dev_inst *table = NULL;
  if (findlmajor(lmajor) != -1)
    return -1;
  
  table = getGlobal() -> pdi_table;

  if (table -> nlmajors + 1 >= table -> maxmajors)
      return -1;
  
  table -> lmajors[table -> nlmajors] = lmajor;

  table -> nlmajors++;

  return 0;

} 

/**************************************/
/*register_lmajor() : if major is 0,  */
/*find a free linux major number,else */
/*attempt to register the number given*/
/**************************************/
int register_lmajor(major_t lmajor)
{
  major_t newmajor;
  if (lmajor == 0)
    {
      newmajor = 1;
      while(addlmajor(newmajor) != 0)
	{
	  newmajor++;
	}
      return newmajor;
    }
  else
    {
      if (addlmajor(lmajor) == 0)
	{
	  return lmajor;
	}
      else
	{
	  return -1;
	}
    }
}

int unregister_lmajor(major_t lmajor)
{
  int index = findlmajor(lmajor);
  struct pits_dev_inst *table = getGlobal() -> pdi_table;

  if (index != -1)
    {
      table -> lmajors[index] = -1;
    }
  return 0;
}

major_t getlmajor(dev_t sdev)
{
  minor_t sminor = getminor(sdev);
  major_t lmajor = -1;
  struct pits_dev_inst *table = getGlobal() -> pdi_table;
  int index = sminor/PITS_MINORS_PER_MAJOR;
  
  lmajor = table -> lmajors[index];
  
  return lmajor;
}
minor_t getlminor(dev_t sdev)
{
  minor_t sminor = getminor(sdev);
  minor_t lminor = sminor % PITS_MINORS_PER_MAJOR;
  
  return (lminor);
}

major_t getsmajor(dev_t ldev)
{
  printk("broken\n");
  return getmajor(ldev);
  //return (getGlobal() -> smajor);
}
major_t getsminor(dev_t ldev)
{
  major_t lmajor = getmajor(ldev);
  minor_t lminor = getminor(ldev);
  major_t sminor = -1;
  int index = findlmajor(lmajor);

  if (index != -1)
    {
      sminor = index * PITS_MINORS_PER_MAJOR + lminor;
    }
  
  return sminor;
  
}

dev_t getldev(dev_t sdev)
{
  minor_t lminor = getlminor(sdev);
  major_t lmajor = getlmajor(sdev);

  return makedevice(lmajor,lminor);
}

int getinstance_s(dev_t sdev)
{
  minor_t sminor = getminor(sdev);
  int instance = sminor/PITS_MINORS_PER_MAJOR;

  return (instance);
}

int getinstance_l(dev_t ldev)
{
  int instance = getsminor(ldev)/PITS_MINORS_PER_MAJOR;
  return instance;
}

int init_pdi_table()
{

  getGlobal() -> pdi_table = kmem_alloc(sizeof(struct pits_dev_inst),
					KM_NOSLEEP);
  if (!getGlobal() -> pdi_table)
    {
      printk("error allocating pdi_table\n");
      return DDI_FAILURE;
    }
  
  getGlobal() -> pdi_table -> lmajors = kmem_alloc(sizeof(major_t)*PITS_MAX_MAJORS,KM_NOSLEEP);
  if (!getGlobal() -> pdi_table -> lmajors)
    {
      printk("error allocating pdi_table -> lmajors\n");
      kmem_free(getGlobal() -> pdi_table,sizeof(struct pits_dev_inst));
      return DDI_FAILURE;
    }

  getGlobal() -> pdi_table -> maxmajors = PITS_MAX_MAJORS;
  getGlobal() -> pdi_table -> nlmajors = 0;

  return DDI_SUCCESS;
}
int fini_pdi_table()
{
  kmem_free(getGlobal() -> pdi_table -> lmajors,sizeof(major_t)*PITS_MAX_MAJORS);
  kmem_free(getGlobal() -> pdi_table,sizeof(struct pits_dev_inst));
  return DDI_SUCCESS;
}
