#include <linux/fs.h>
#include <linux/blkdev.h>

#include <misc/pits_block.h>

static int _read_ahead[MAX_BLKDEV]; 
int *get_read_ahead_ptr() 
{ 
  return (_read_ahead); 
}  

/* from fs/inode.c */
int invalidate_device(kdev_t dev, int do_sync) 
{
  printk("invalidate_device\n");
  return 0;
}


/*from fs/block_dev.c*/
/* 
 * This routine checks whether a removable media has been changed, 
 * and invalidates all buffer-cache-entries in that case. This 
 * is a relatively slow routine, so we have to try to minimize using 
 * it. Thus it is called only upon a 'mount' or 'open'. This 
 * is the best way of combining speed and utility, I think. 
 * People changing diskettes in the middle of an operation deserve 
 * to lose :-) 
 */ 

int check_disk_change(kdev_t dev)
{
  printk("inside check_disk_change\n");
  int i;
  const struct block_device_operations * bdops = NULL;

  i = MAJOR(dev);
  if (i < MAX_BLKDEV)
    {
      bdops = getBdops(dev);
    }
  /*
    bdops = blkdevs[i].bdops;
  */
	  
  if (bdops == NULL) 
    {
      devfs_handle_t de;
      
      de = devfs_find_handle (NULL, NULL, i, MINOR (dev),
			      DEVFS_SPECIAL_BLK, 0);
      if (de) 
	{
	  bdops = devfs_get_ops (de);
	  devfs_put_ops (de); /* We're running in owner module */
	}
    }
  if (bdops == NULL)
    return 0;
  if (bdops->check_media_change == NULL)
    return 0;
  if (!bdops->check_media_change(dev))
    return 0;
  /*
    printk(KERN_DEBUG "VFS: Disk change detected on device %s\n",
    bdevname(dev));
  */
  if (invalidate_device(dev, 0))
    printk("VFS: busy inodes on changed media.\n");
  
  if (bdops->revalidate)
    bdops->revalidate(dev);
  return 1;
}


int fsync_dev(kdev_t device)
{
  printk("inside fsync_dev,null definition\n");
  return 0;
} 

void refile_buffer(struct buffer_head * buf)
{
  printk("inside refile_buffer, null definition\n");
  ;
}
 
//extern ssize_t block_write(struct file *, const char *, size_t, loff_t *); 


struct buffer_head * linux_bread(kdev_t dev, int block , int size)
{

  printk("inside linux bread major %d minor %d block %d size %d",getmajor(dev),
	 getminor(dev),block,size);
  printk("linux_bread() : NULL fxn!\n");
  return NULL;
  /*
  struct buffer_head * bh;
  bh = getblk(dev, block, size);
  touch_buffer(bh);
  if (buffer_uptodate(bh))
    return bh;
  ll_rw_block(READ, 1, &bh);
  wait_on_buffer(bh);
  if (buffer_uptodate(bh))
    return bh;
  brelse(bh);
  return NULL;
  */
}

/*fs/devices.c*/
/*
 * Print device name (in decimal, hexadecimal or symbolic)
 * Note: returns pointer to static data!
 */
const char * kdevname(kdev_t dev)
{
	static char buffer[32];
	sprintf(buffer, "%02x:%02x", MAJOR(dev), MINOR(dev));
	return buffer;
}
