#include <linux/genhd.h>
#include <misc/globals.h>

/**
 * add_gendisk - add partitioning information to kernel list
 * @gp: per-device partitioning information
 *
 * This function registers the partitioning information in @gp
 * with the kernel.
 */
void add_gendisk(struct gendisk *gp)
{
	struct gendisk *sgp;

	//write_lock(&gendisk_lock);

	/*
 	 *	In 2.5 this will go away. Fix the drivers who rely on
 	 *	old behaviour.
 	 */

	for (sgp = getGlobal() -> _gendisk_head; sgp; sgp = sgp->next)
	{
		if (sgp == gp)
		{
//			printk(KERN_ERR "add_gendisk: device major %d is buggy and added a live gendisk!\n",
//				sgp->major)
			goto out;
		}
	}
	getGlobal() -> _gendisk_array[gp->major] = gp;
	gp->next = getGlobal() -> _gendisk_head;
	getGlobal() -> _gendisk_head = gp;
out:
	return;
	//write_unlock(&gendisk_lock);
}

/**
 * del_gendisk - remove partitioning information from kernel list
 * @gp: per-device partitioning information
 *
 * This function unregisters the partitioning information in @gp
 * with the kernel.
 */
void del_gendisk(struct gendisk *gp)
{
	struct gendisk **gpp;

	//write_lock(&gendisk_lock);
	getGlobal() -> _gendisk_array[gp->major] = NULL;
	for (gpp = &(getGlobal() -> _gendisk_head); *gpp; gpp = &((*gpp)->next))
		if (*gpp == gp)
			break;
	if (*gpp)
		*gpp = (*gpp)->next;
	//write_unlock(&gendisk_lock);
}

void grok_partitions(struct gendisk *dev, int drive, unsigned minors, long size)
{
  printk("inside grok_partitions\n");
  int i;
  int first_minor = drive << dev->minor_shift;
  int end_minor   = first_minor + dev->max_p;

  printk("first_minor %d end_minor %d minor_shift %d drive %d\n",
	 first_minor,
	 end_minor,
	 dev -> minor_shift,
	 drive);

  printk("minors %d size %ld\n",minors,size);

  if(!dev->sizes)
    blk_size[dev->major] = NULL;

  dev->part[first_minor].nr_sects = size;
  /* No such device or no minors to use for partitions */
  if ( !size && dev->flags && (dev->flags[drive] & GENHD_FL_REMOVABLE) )
    devfs_register_partitions (dev, first_minor, 0);
  if (!size || minors == 1)
    return;

  if (dev->sizes) {
    dev->sizes[first_minor] = size >> (BLOCK_SIZE_BITS - 9);
    for (i = first_minor + 1; i < end_minor; i++)
      dev->sizes[i] = 0;
  }
  blk_size[dev->major] = dev->sizes;

  //Need to check to see what this does, try to dumb it down to handle just Solaris partions (UFS?).
  //check_partition(dev, MKDEV(dev->major, first_minor), 1 + first_minor);

  /*
   * We need to set the sizes array before we will be able to access
   * any of the partitions on this device.
   */
  if (dev->sizes != NULL) {       /* optional safeguard in ll_rw_blk.c */
    for (i = first_minor; i < end_minor; i++)
      {
	dev->sizes[i] = dev->part[i].nr_sects >> (BLOCK_SIZE_BITS - 9);
      }
  }
  printk("done grok_partitions\n");
}

