#include <linux/slab.h>

#include <misc/pits_param.h>
//#include <misc/globals.h>
#include <misc/pits_state.h>

#include <sys/ddi.h>
#include <sys/sunddi.h>
//#include <string.h>
//#include <stdlib.h>
//#include <ctype.h>
#include <linux/ctype.h>

void PITS_set_parm(void *parm,char *type,void* pval)
{
  struct parameter_value *pvalue = (struct parameter_value *)pval;
  PITS_parameter *parameter = pvalue -> id;
  //  int min = parameter -> min;
  int n;
  char ptype = parameter -> type;
  char **sptr = NULL;
  char **dptr = NULL;
  int i;
  if (parameter -> max < pvalue -> n)
    {
      n = parameter -> max;
    }
  else
    {
      n = pvalue -> n;
    }


  switch(ptype)
    {
    case 'i':
      for (i=0;i<n;i++)
	{
	  printk("%d ",*(((int*)pvalue -> value) + i)); 
	  *((int*)parm + i) = *(((int*)pvalue -> value) + i);
	}
      break;
    case 'l':
      for (i=0;i<n;i++)
	{
	  printk("%ld ",*(((int*)pvalue -> value) + i));
	  *((int*)parm + i) = *(((int*)pvalue -> value) + i);
	}
      break;
    case 's':
      dptr = (char**)parm;
      sptr = (char**)pvalue -> value;
      for (i=0;i<n;i++)
	{
	  printk("%s",sptr[i]);
	  dptr[i] = sptr[i];
	}
      break;
    default:
      printk("Unrecognized parameter type %c\n",ptype);
    }
  
  printk("\n");




  //printk("done PITS_set_parm");
}


int set_linux_parameters(dev_info_t *dip)
{
  int index;
  
  int ret;
  PITS_parameter *parameters;
  int nparameters;
  sun_linux_state *instance_state;
  struct parameter_value pvalue;
  
  getInstanceFReturn(dip,instance_state);
  nparameters = instance_state -> npar;
  parameters  = instance_state -> par;

  //printk("Inside set_linux_parameters npar=%d",nparameters);
  
  for (index=0;index < nparameters;index++)
    {
      //get the parameter.
      //printk("checking for parameter %s of type %s",parameters[index].name,parameters[index].linux_type);
      pvalue.id = &parameters[index];

      ret = get_linux_parameter(dip,&pvalue);
      if (ret == DDI_SUCCESS)
	{
	  //okay parameter was found, call the setting function.
	  printk("setting parameter %s of type %s",parameters[index].name,parameters[index].linux_type);
	  parameters[index].set_parameter(&pvalue);
	  //now free the memory
	  /*
	  if (parameter != NULL)
	    {
	      kfree(parameter);
	    }
	  */
	}
    }
  return DDI_SUCCESS;
}

int todigit(char c)
{
  switch(c)
    {
    case '0':
      return 0;
    case '1':
      return 1;
    case '2':
      return 2;
    case '3':
      return 3;
    case '4':
      return 4;
    case '5':
      return 5;
    case '6':
      return 6;
    case '7':
      return 7;
    case '8':
      return 8;
    case '9':
      return 9;
    default:
      return 0;
    }
}



int atoi(char *buf)
{
  int ans = 0;
  int factor  = 1;
  int index;
  int i;
  int start = 0;
  int neg = FALSE;
  if (buf[0] == '-')
    {
      start = 1;
      neg = TRUE;
    }

  for (i=0;i<strlen(buf)-start;i++)
    {
      index = strlen(buf) - (i+1);

      ans += todigit(buf[index]) * factor;
      factor *= 10;
    }
  return(neg ? -ans:ans);

}

void extractRanges(char *type,int *min,int *max)
{
  char *buf;
  int i,j;
  buf = kmalloc(sizeof(char)*strlen(type),0);
  memset(buf,0,strlen(type));

  i=0;

  while (isdigit((int)type[i]))
    {
      buf[i] = type[i];
      i++;
    }
  *min = atoi(buf);

  memset(buf,0,strlen(type));

  j = i+1;
  i=0;

  while (isdigit((int)type[j]))
  {
    buf[i] = type[j];
    i++;
    j++;
  }

  *max = atoi(buf);
  
  printk("min:%d max:%d",*min,*max);

  kfree(buf);

}

/***********************************************
 *get_linux_parameter()                        *
 ***********************************************/

int get_linux_parameter(void *dip,struct parameter_value *pvalue)
{
  int ret = DDI_SUCCESS;
  //printk("inside get_linux_parameter");
  PITS_parameter *parameter = pvalue -> id;

  char *parname = parameter -> name;
  char type = parameter -> type;



  if (dip == NULL) {printk("dip is null, this shouldn't happen"); return DDI_FAILURE;}
  if (parname == NULL) {printk("parname is null, this shouldn't happen"); return DDI_FAILURE;}
  //if (type == NULL) {printk("type is null, this shouldn't happen"); return DDI_FAILURE;}
  //printk("does %s of strlen %d exist?",parname,strlen(parname));
  if (!ddi_prop_exists(DDI_DEV_T_ANY,                       //dev_t.
		      dip,                                  //dev_info_t structure.
		      DDI_PROP_DONTPASS | DDI_PROP_NOTPROM, //flags.
		       parname))                             //parameter name.
    {
      //printk("prop doesn't exist, return DDI_FAILURE");
      return DDI_FAILURE;
    }
  switch (type)
    {
    case 'i':
      //printk("readLinuxInteger");
      ret = readLinuxInteger(dip,pvalue);
      break;
    case 'l':
      //printk("readLinuxLong");
      ret = readLinuxLong(dip,pvalue);
      break;
    case 's':
      //printk("readLinuxString");
      ret = readLinuxString(dip,pvalue);
      break;
    default:
      printk("Unrecognized type");
      ret = DDI_FAILURE;
    }
  if (ret != DDI_SUCCESS)
    {
      printk("Error read parameter");
    }

  return ret;

}

int readLinuxInteger(void *dip,struct parameter_value *pvalue)
{
  
  PITS_parameter *parameter = pvalue -> id;
  int i;
  int min = parameter -> min;
  int max = parameter -> max;
  int *iarray;
  int nelements;
  int ret;
  int *iptr;
  //read in an int array.
  ret = ddi_prop_lookup_int_array(DDI_DEV_T_ANY,
				  dip,
				  DDI_PROP_DONTPASS | DDI_PROP_NOTPROM,
				  parameter -> name,
				  &iarray,
				  &nelements);

  if (ret == DDI_SUCCESS)
    {
      if (nelements <= max && nelements >= min)
	{
	  pvalue -> value = kmalloc(nelements * sizeof(int),GFP_KERNEL);
	  pvalue -> n = nelements;
	  iptr = (int*)pvalue -> value;
	  for (i=0;i<nelements;i++)
	    {
	      iptr[i] = iarray[i];
	    }
	  ret = DDI_SUCCESS;
	}
      else
	{
	  printk("Number of elements are out of range");
	  ret = DDI_FAILURE;
	}
      ddi_prop_free(iarray);
    }
  else
    {
      printk("There was an error reading the integer array prop %d",ret);
    }
				  
  return ret;
}

int readLinuxLong(void *dip,struct parameter_value *pvalue)
{
  return(readLinuxInteger(dip,pvalue));
}
int readLinuxString(void *dip,struct parameter_value *pvalue)
{

  PITS_parameter *parameter = pvalue -> id;
  int min = parameter -> min;
  int max = parameter -> max;
  char **sarray;
  int i;
  int nelements;
  int ret;
  char **sptr;
  //read in an int array.
  ret = ddi_prop_lookup_string_array(DDI_DEV_T_ANY,
				  dip,
				  DDI_PROP_DONTPASS | DDI_PROP_NOTPROM,
				  parameter -> name,
				  &sarray,
				  &nelements);

  if (ret == DDI_SUCCESS)
    {
      if (nelements <= max && nelements >= min)
	{
	  pvalue -> value = kmalloc(nelements * sizeof(char **),GFP_KERNEL);
	  pvalue -> n = nelements;

	  sptr = (char**)pvalue -> value;
	  for (i=0;i<nelements;i++)
	    {
	      sptr[i] = kmalloc(strlen(sarray[i])+1,GFP_KERNEL);
	      memcpy(sptr[i],sarray[i],strlen(sarray[i])+1);
	    }
	  ret = DDI_SUCCESS;
	}
      else
	{
	  printk("Number of elements are out of range");
	  ret = DDI_FAILURE;
	}
      ddi_prop_free(sarray);
    }
  else
    {
      printk("There was an error reading the string array %d",ret);
    }

  return ret;
}

