#include <linux/pci.h>
#include <misc/pits_dma.h>
#include <misc/pits_state.h>
#include <misc/slab.h>


typedef struct pits_dma_element
{
  ddi_dma_handle_t *handle;
  //ddi_dma_handle_t mem_handle;
  caddr_t          addr;
  uint_t size;
  uint_t real_length;
  ddi_dma_cookie_t cookie;
  uint_t cookie_count;
  int allocated;
  //int handle_allocated;
  //int use_mem_handle;
  //int mem_handle_allocated;
  //int handle_bound;
  int flags;
  struct pits_dma_element *next;
}pits_dma_element_t;

static ddi_dma_handle_t *shandle = NULL;
static pits_dma_element_t *pelement = NULL;
static pits_dma_element_t *bound_list = NULL;
static pits_dma_element_t *free_list = NULL;
//static struct pits_dma_handle_list free_dma_handles;
static struct pits_streaming_dma stream_dma;
static struct pits_dma_handle *consistent_list = NULL;



pits_dma_element_t *findDMAElementByDMA(pits_dma_element_t *head,dma_addr_t dma_addr)
{
  pits_dma_element_t *ptr = head;

  while (ptr != NULL && ptr -> cookie.dmac_address != dma_addr)
    {
      ptr = ptr -> next;
    }
  return ptr;

}
void removeDMAElement(pits_dma_element_t **head,pits_dma_element_t *ptr)
{
  pits_dma_element_t *prev=NULL;
  pits_dma_element_t *curr=NULL;

  curr = *head;

  //DEBUGCMD(printk("inside removeDMAElement()\n"));

  while(curr != NULL && curr != ptr)
    {
      prev = curr;
      curr = curr -> next;
    }
  if (curr == NULL)
    {
      printk("removeDMAElement():warning ptr not found!\n");
      return;
    }
  if (prev == NULL)
    {
      //we are at the beginning of the list.
      *head = NULL;
    }
  else
    {
      //we are in the middle or end.
      prev -> next = curr -> next;
    }


}
void addDMAElement(pits_dma_element_t **head,pits_dma_element_t *element)
{
  //DEBUGCMD(printk("inside add DMAElement\n"));
  if (*head == NULL)
    {
      *head = element;
    }
  else
    {
      element -> next = *head;
      *head = element;
    }
  //DEBUGCMD(printk("done add DMAElement\n"));
}



pits_dma_element_t *allocDMAElement()
{
  int ret;
  pits_dma_element_t *ptr = NULL;
  //ddi_dma_attr_t attr = DEFAULT_PCI_DMA_ATTR;
  //ptr = kmem_zalloc(sizeof(pits_dma_element_t),KM_NOSLEEP);
  ptr = pelement;
  //ptr = kmalloc(sizeof(pits_dma_element_t),0);

  ptr -> next = NULL;
  //bzero(ptr,sizeof(pits_dma_element_t));
  if (ptr == NULL)
    {
      printk("getDMAElement() : couldn't allocate DMA Element!\n");
      goto error;
    }
  //ret = init_dma_handle(&(ptr -> handle));

  //ptr -> handle = alloc_dma_handle(KM_SLEEP);
  ptr -> handle = shandle;

  if (ptr == NULL)
  //if (ret != DDI_SUCCESS)
    {
      printk("getDMAElement() : couldn't init DMA Handle!\n");
      goto error_kmem_alloc;
    }
			     
  return ptr;


 error_kmem_alloc:
  //kmem_free(ptr,sizeof(pits_dma_element_t));
  //kfree(ptr);
  ;
 error:
  return NULL;
  
}

void freeDMAElement(pits_dma_element_t *ptr)
{
  //DEBUGCMD(printk("inside freeDMAElement\n"));
  //ddi_dma_free_handle(&(ptr -> handle));
  //release_dma_handle(ptr -> handle);
  //kmem_free(ptr,sizeof(pits_dma_element_t));
  //kfree(ptr);
  //DEBUGCMD(printk("done freeDMAElement\n"));
}

/****************/
/*from asm/pci.h*/
/****************/
/* Allocate and map kernel buffer using consistent mode DMA for a device.
 * hwdev should be valid struct pci_dev pointer for PCI devices,
 * NULL for PCI-like buses (ISA, EISA).
 * Returns non-NULL cpu-view pointer to the buffer if successful and
 * sets *dma_addrp to the pci side dma address as well, else *dma_addrp
 * is undefined.
 */


/*build a dma list for the handles.*/
int init_pci_dma(dev_info_t *dip)
{
  /*
  struct pits_dma_handle *handle = NULL;
  ddi_dma_attr_t attr = DEFAULT_PCI_DMA_ATTR;
  printk("inside initialize_pci_dma()\n");

  bzero(&stream_dma,sizeof(struct pits_streaming_dma));

  handle = createDMAHandle(dip,&attr);
  if (handle == NULL)
    {
      printk("Error allocating streaming handle\n");
      return DDI_FAILURE;
    }

  stream_dma.dma_memory = createDMAMemory(handle,
					  DDI_DMA_STREAMING,
					  0xffff);

  if (stream_dma.dma_memory == NULL)
    {
      printk("Error allocating streaming dma buffer\n");
      deleteDMAHandle(&handle);
      return DDI_FAILURE;
    }
  stream_dma.dma_lock = SPIN_LOCK_UNLOCKED;
  */
  shandle = alloc_dma_handle(KM_SLEEP);
  if (shandle == NULL)
    {
      return DDI_FAILURE;
    }
  pelement = kmem_zalloc(sizeof(pits_dma_element_t),KM_NOSLEEP);
  consistent_list = NULL;
  getGlobal() -> pci_dma_initialized = TRUE;
  printk("done initialize_pci_dma()\n");
  return DDI_SUCCESS;
  
}

int fini_pci_dma(dev_info_t *dip)
{
  //delete stream dma memory.
  /*
  struct pits_dma_handle *handle = stream_dma.dma_memory -> phandle;
  deleteDMAMemory(&stream_dma.dma_memory);
  deleteDMAHandle(&handle);
  */
  getGlobal() -> pci_dma_initialized = FALSE;
  printk("done fini_pci_dma()\n");

  release_dma_handle(shandle);
  kmem_free(pelement,sizeof(pits_dma_element_t));
  return DDI_SUCCESS;
}

void *pci_alloc_consistent(struct pci_dev *hwdev, size_t size,
				  dma_addr_t *dma_handle)
{
  void *ret;
  int status;
  struct pits_dma_handle *handle;
  struct pits_dma_memory *memory;
  DEBUGDMESG("Inside pci_alloc_consistent\n");

  if (getGlobal() -> pci_dma_initialized == FALSE)
    {
      printk("pci_dma not initialized!\n");
      return NULL;
    }

  handle = createDMAHandle(getGlobal() -> current_dip,
			   &(getGlobal() -> pci_dma_attr));

  if (handle == NULL)
    {
      printk("error allocating handle\n");
      return NULL;
    }
  
  memory = createDMAMemory(handle,
			   DDI_DMA_CONSISTENT,
			   size);
  if (memory == NULL)
    {
      printk("error allocating memory\n");
      deleteDMAHandle(&handle);
      return NULL;
    }
  
  status = bindDMAMemory(memory,DDI_DMA_RDWR);
  if (status == DDI_DMA_MAPPED)
    {
      DEBUGCMD(printk("pci_alloc consistent success:\n");
	       printk("addr %lu bus %lu size %d count %d\n",
		      memory -> addr,
		      memory -> cookie.dmac_address,
		      memory -> size,
		      memory -> cookie_count););
      if (consistent_list == NULL)
	{
	  consistent_list = handle;
	}
      else
	{
	  handle -> next = consistent_list;
	  consistent_list = handle;
	}
      ret = (void*)memory -> addr;
      *dma_handle = memory -> cookie.dmac_address;
    }
  else
    {
      pits_dma_bind_handle_error(status);
      deleteDMAMemory(&memory);
      deleteDMAHandle(&handle);
      *dma_handle = 0;
      ret = NULL;
    }

  DEBUGDMESG("done pci_alloc_consistent\n");
  return ret;
}

/* Free and unmap a consistent DMA buffer.
 * cpu_addr is what was returned from pci_alloc_consistent,
 * size must be the same as what as passed into pci_alloc_consistent,
 * and likewise dma_addr must be the same as what *dma_addrp was set to.
 *
 * References to the memory and mappings associated with cpu_addr/dma_addr
 * past this call are illegal.
 */
void pci_free_consistent(struct pci_dev *hwdev, size_t size, 
			 void *vaddr, dma_addr_t dma_handle)
{
  DEBUGDMESG("Inside pci_free_consistent\n");
  struct pits_dma_handle *prev = NULL;
  struct pits_dma_handle *ptr = consistent_list;

  while (ptr != NULL && ptr -> memory -> addr != vaddr)
    {
      prev = ptr;
      ptr = ptr -> next;
    }
  if (ptr == NULL)
    {
      printk("warning, not found!\n");
      goto out;
    }
  
  if (prev == NULL)
    {
      //at front of list.
      consistent_list = consistent_list -> next;
    }
  else
    {
      prev -> next = ptr -> next;
    }
  
  unbindDMAMemory(ptr -> memory);
  deleteDMAMemory(&ptr -> memory);
  deleteDMAHandle(&ptr);

 out:
  DEBUGCMD(printk("Done pci_free_consistent\n"));
  
}
/************************/
/*from linux/pci.h      */
/************************/
int pci_set_dma_mask(struct pci_dev *dev,u64 mask)
{
  DEBUGDMESG("Inside pci_set_dma_mask\n");
  DEBUGDMESG("Done pci_set_dma_mask\n");
  return 0;
}

/******************************/
/*asm/pci.h                   */
/******************************/

static int pits_dma_bind_mem_cb(caddr_t arg)
{
  pits_dma_element_t *element = (pits_dma_element_t*)arg;
  int status;


  status = ddi_dma_addr_bind_handle(element -> handle,
				    NULL,
				    element -> addr,
				    element -> size,
				    element -> flags,
				    DDI_DMA_DONTWAIT,
				    NULL,
				    &(element -> cookie),
				    &(element -> cookie_count));
  switch(status)
    {
    case DDI_DMA_MAPPED:
      printk("DDI_DMA_MAPPED\n");
      return DDI_DMA_CALLBACK_DONE;
    case DDI_DMA_PARTIAL_MAP:
    case DDI_DMA_INUSE:
      printk("DDI_DMA_INUSE....\n");
      break;
    case DDI_DMA_NORESOURCES:
      printk("DDI_DMA_NORESOURCES....\n");
      break;
    case DDI_DMA_NOMAPPING:
      printk("DDI_DMA_NOMAPPING....\n");
      break;
    case DDI_DMA_TOOBIG:
      printk("DDI_DMA_TOOBIG....\n");
      break;
    default:
      printk("unknown error");
    }
  return DDI_DMA_CALLBACK_RUNOUT;
    
}

/* Map a single buffer of the indicated size for DMA in streaming mode.
 * The 32-bit bus address to use is returned.
 *
 * Once the device is given the dma address, the device owns this memory
 * until either pci_unmap_single or pci_dma_sync_single is performed.
 */

dma_addr_t pci_map_single(struct pci_dev *hwdev, void *ptr, 
			  size_t size,int direction)
{
  
  
  int status;
  pits_dma_element_t *element = NULL;
  kmemptr_t *mptr = NULL;

  //DEBUGCMD(printk("inside pci_map_single()\n"));
  
  if (ptr == NULL)
    {
      printk("pci_map_single() : error null ptr!\n");
      goto error;
    }

  element = allocDMAElement();
  //element = pelement;

  if (element == NULL)
    {
      printk("pci_map_single() : cannot allocate dma_element!\n");
      goto error;
    }

  element -> addr = ptr;
  element -> size = size;
  

  switch (direction)
    {
    case PCI_DMA_TODEVICE:
      element -> flags = DDI_DMA_WRITE;
      break;
    case PCI_DMA_FROMDEVICE:
      element -> flags = DDI_DMA_READ;
      break;
    case PCI_DMA_BIDIRECTIONAL:
      element -> flags = DDI_DMA_RDWR;
      break;
    case PCI_DMA_NONE:
    default:
      linux_panic("pci_map_single() : direction - unknown type");
      goto alloc_error;
    }
  
  element -> flags = element -> flags | DDI_DMA_STREAMING;

  //DEBUGCMD(printk("binding memory...\n"));
  status = 
    ddi_dma_addr_bind_handle(*(element -> handle),  /*handle*/
			     NULL,                 
			     element -> addr,
			     element -> size,
			     element -> flags,
			     /*
			       pits_dma_bind_mem_cb,
			       (caddr_t)element,
			     */
			     DDI_DMA_DONTWAIT,
			     NULL,
			     &(element -> cookie),
			     &(element -> cookie_count));



  //DEBUGCMD(printk("done binding memory...\n"));
  if (status != DDI_DMA_MAPPED)
    {
      pits_dma_bind_handle_error(status);
      panic("inside pci_map_single");
      goto alloc_error;
      return 0;
    }
  else if (element -> cookie_count != 1)
    {
      printk("cookie count %d > 1!\n",element -> cookie_count);
      goto bind_error;
    }
  //addDMAElement(&bound_list,element);
  return (element -> cookie.dmac_address);
  
 bind_error:
  ddi_dma_unbind_handle(*(element -> handle));
 alloc_error:
  //freeDMAElement(element);
 error:
  linux_panic("an error was detected, pits driver is unstable...\n");
  return 0;
}


/* Unmap a single streaming mode DMA translation.  The dma_addr and size
 * must match what was provided for in a previous pci_map_single call.  All
 * other usages are undefined.
 *
 * After this call, reads by the cpu to the buffer are guarenteed to see
 * whatever the device wrote there.
 */
void pci_unmap_single(struct pci_dev *hwdev, dma_addr_t dma_addr,
                                    size_t size, int direction)
{
  
  int status;
  pits_dma_element_t *element = NULL;

  //DEBUGCMD(printk("inside pci_unmap_single()\n"));

  
  //element = findDMAElementByDMA(bound_list,dma_addr);

  element = pelement;

  if (element == NULL)
    {
      printk("pci_unmap_single() : element not found!\n");
      goto error;
    }
  //unbind the element
  status = ddi_dma_unbind_handle(*(element -> handle));
  if (status != DDI_SUCCESS)
    {
      printk("pci_unmap_single() : unable to unbind handle!\n");
      goto error;
    }
  //zero out the addr.
  element -> addr = 0;
  
  //remove it from the list.
  //removeDMAElement(&bound_list,element);

  //free the element.
  //freeDMAElement(element);
  
  //DEBUGCMD(printk("pci_unmap_single():done\n"));

  return;
 error:
  linux_panic("pci_unmap_single() : error occured, driver unstable..\n");
  return;
}

/* Map a set of buffers described by scatterlist in streaming
 * mode for DMA.  This is the scather-gather version of the
 * above pci_map_single interface.  Here the scatter gather list
 * elements are each tagged with the appropriate dma address
 * and length.  They are obtained via sg_dma_{address,length}(SG).
 *
 * NOTE: An implementation may be able to use a smaller number of
 *       DMA address/length pairs than there are SG table elements.
 *       (for example via virtual mapping capabilities)
 *       The routine returns the number of addr/length pairs actually
 *       used, at most nents.
 *
 * Device ownership issues as mentioned above for pci_map_single are
 * the same here.
 */
int pci_map_sg(struct pci_dev *hwdev, struct scatterlist *sg,
                             int nents, int direction)
{
  int i;
  
  printk("inside pci_map_sg()\n");
  
  if (direction == PCI_DMA_NONE)
    BUG();

  /*
   * temporary 2.4 hack
   */
  for (i = 0; i < nents; i++ ) {
    if (sg[i].address && sg[i].page)
      BUG();
    else if (!sg[i].address && !sg[i].page)
      BUG();

    if (sg[i].address)
      sg[i].dma_address = virt_to_bus(sg[i].address);
    else
      sg[i].dma_address = page_to_bus(sg[i].page) + sg[i].offset;
  }

  flush_write_buffers();

  printk("done pci_map_sg() nents - %d\n",nents);

  return nents;
}

/* Unmap a set of streaming mode DMA translations.
 * Again, cpu read rules concerning calls here are the same as for
 * pci_unmap_single() above.
 */
void pci_unmap_sg(struct pci_dev *hwdev, struct scatterlist *sg,
                                int nents, int direction)
{

  printk("pci_unmap_sg()\n");

  if (direction == PCI_DMA_NONE)
    BUG();
  /* Nothing to do */
  printk("done pci_unmap_sg()\n");
}

struct pits_dma_handle *createDMAHandle(dev_info_t *dip,
                                        ddi_dma_attr_t *attr)
{
  int status;
  struct pits_dma_handle *phandle = kmem_alloc(sizeof(struct pits_dma_handle),0);

  status = ddi_dma_alloc_handle(dip,
				attr,
				DDI_DMA_SLEEP,
				NULL,
				&(phandle -> handle));
  if (status != DDI_SUCCESS)
    {
      pits_dma_alloc_handle_error(status);
      kmem_free(phandle,sizeof(struct pits_dma_handle));
      return NULL;
    }
  phandle -> dip = dip;
  bcopy(attr,&(phandle -> dma_attr),sizeof(ddi_dma_attr_t));
  phandle -> memory = NULL;
  phandle -> next   = NULL;

  return (phandle);
}

struct pits_dma_memory *createDMAMemory(struct pits_dma_handle *handle,
                                        int type,
                                        int size)
{
  int status;

  struct pits_dma_memory *pmemory = kmem_alloc(sizeof(struct pits_dma_memory),0);
  pmemory -> phandle = handle;
  handle -> memory = pmemory;
  pmemory -> flags   = type;
  status = ddi_dma_mem_alloc(pmemory -> phandle -> handle,
			     size,
			     &(getGlobal() -> sun_dev_acc_attr),
			     pmemory -> flags,
			     DDI_DMA_SLEEP,
			     NULL,
			     &(pmemory -> addr),
			     &(pmemory -> size),
			     &(pmemory -> mem_handle));

  if (status != DDI_SUCCESS)
    {
      pits_dma_alloc_mem_error(status);
      kmem_free(pmemory,sizeof(struct pits_dma_memory));
      return NULL;
    }

  printk("DMA Memory successfully allocated\n");
  return pmemory;
}
int bindDMAMemory(struct pits_dma_memory *memory,
                  int direction)
{
  int status;

  status = ddi_dma_addr_bind_handle(memory -> phandle -> handle,
				    NULL,
				    memory -> addr,
				    memory -> size,
				    direction | memory -> flags,
				    DDI_DMA_SLEEP,
				    NULL,
				    &(memory -> cookie),
				    &(memory -> cookie_count));

  if (status == DDI_DMA_MAPPED)
    {
      memory -> bound = TRUE;
    }
  else
    {
      pits_dma_bind_handle_error(status);
    }

  return status;
}
int unbindDMAMemory(struct pits_dma_memory *memory)
{
  int status;

  status = ddi_dma_unbind_handle(memory -> phandle -> handle);

  if (status == DDI_SUCCESS)
    {
      memory -> bound = FALSE;
    }

  return status;
}
int deleteDMAMemory(struct pits_dma_memory **memory)
{
  struct pits_dma_memory *tmemory = *memory;
  struct pits_dma_handle *phandle = tmemory -> phandle;
  ddi_dma_handle_t *mhandle = &(tmemory-> mem_handle);

  ddi_dma_mem_free(mhandle);

  (*memory) -> phandle = NULL;
  phandle -> memory = NULL;
  
  kmem_free(*memory,sizeof(struct pits_dma_memory));
  *memory = NULL;
  return DDI_SUCCESS;
}
int deleteDMAHandle(struct pits_dma_handle **handle)
{
  struct pits_dma_handle *thandle = *handle;
  ddi_dma_free_handle(&(thandle -> handle));
  
  kmem_free(*handle,sizeof(struct pits_dma_handle));

  return DDI_SUCCESS;
}


