#include <asm/semaphore.h>
#include <asm/atomic.h>
#include <sys/ksynch.h>


void linux_sema_init(struct semaphore *sem,int val)
{
  if (sem == NULL)
    {
      printk("linux_sema_init() : null pointer!\n");
      goto error;
    }
  if (sem -> initialized == TRUE)
    {
      printk("linux_sema_init() : already initialized!\n");
      goto error;
    }
  mutex_init(&(sem -> mutex), NULL, MUTEX_DRIVER, NULL);
  cv_init(&(sem -> cv_wait),NULL,CV_DRIVER,NULL);

  mutex_enter(&(sem -> mutex));
  atomic_set(&(sem -> count),val);

  sem -> initialized = TRUE;
  mutex_exit(&(sem -> mutex));

  return;

 error:
  return;



}

void down(struct semaphore *sem)
{
  //int ret=0;
  
  if (sem == NULL)
    {
      printk("down_interruptible(): NULL semaphore!\n");
      goto error;
    }
  if (sem -> initialized == FALSE)
    {
      linux_sema_init(sem,sem -> initialvalue);
    }
  mutex_enter(&(sem -> mutex));
 
  while (atomic_read(&(sem -> count)) <= 0)
    {
      cv_wait(&(sem -> cv_wait),&(sem ->mutex));
    }

  atomic_dec(&(sem -> count));
  mutex_exit(&(sem -> mutex));
  return;
 error:
  return;


}

int down_interruptible(struct semaphore *sem)
{
  int ret=0;
  int havemutex = FALSE;
  struct task_struct *current_task = current;
  if (sem == NULL)
    {
      printk("down_interruptible(): NULL semaphore!\n");
      goto error;
    }
  if (sem -> initialized == FALSE)
    {
      linux_sema_init(sem,sem -> initialvalue);
    }
  mutex_enter(&(sem -> mutex));
 
  while ((atomic_read(&(sem -> count)) <= 0) && (current_task -> sigpending == 0))
    {
      mutex_enter(&(current_task -> mutex));
      current_task -> cv_wait = &(sem -> cv_wait);
      current_task -> state = TASK_INTERRUPTIBLE; 
      mutex_exit(&(current_task -> mutex));
      printk("down_interruptible, waiting for signal...\n");
      ret = cv_wait_sig(&(sem ->cv_wait),&(sem -> mutex));
      mutex_enter(&(current_task -> mutex));
      current_task -> state = TASK_UNINTERRUPTIBLE;
      
      if (ret == 0) 
	{
	  //we were signalled and don't have the mutex.
	  havemutex = FALSE;
	  current_task -> sigpending = 1;
	}
      else if (current_task -> sigpending != 0)
	{
	  //we were signalled and have the mutex.
	  havemutex = TRUE;
	}
      mutex_exit(&(current_task -> mutex));
    }

  mutex_enter(&(current_task -> mutex));
  current_task -> cv_wait = NULL;
  current_task -> state = TASK_INTERRUPTIBLE;
  if (current_task -> sigpending != 0)
    {
      printk("down_interruptible() : signal detected!\n");
      mutex_exit(&(current_task -> mutex));
      if (havemutex)
	{
	  mutex_exit(&(sem -> mutex));
	}
      return current_task -> sigpending;
    }
  mutex_exit(&(current_task -> mutex));
  //okay, no signals, we have a go.
  atomic_dec(&(sem -> count));
  mutex_exit(&(sem -> mutex));
  return 0;
 error:
  return 1;


}

void up(struct semaphore *sem)
{
  if (sem == NULL)
    {
      printk("up():NULL semaphore!\n");
      goto error;
    }
  if (sem -> initialized == FALSE)
    {
      linux_sema_init(sem,sem -> initialvalue);
    }
  mutex_enter(&(sem -> mutex));
  atomic_inc(&(sem -> count));

  cv_broadcast(&(sem -> cv_wait));
  mutex_exit(&(sem -> mutex));
  return;
 error:
  return;

}
