/*
 *	linux/kernel/resource.c
 *
 * Copyright (C) 1999	Linus Torvalds
 * Copyright (C) 1999	Martin Mares <mj@ucw.cz>
 *
 * Arbitrary resource management.
 */

#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/spinlock.h>
#include <asm/io.h>

#include <linux/pci.h>
#include <misc/pits_dma.h>
#include <misc/globals.h>

//#include <string.h>

//static struct resource* ioport_resource_tree;
//static struct resource* iomem_resource_tree;

struct resource ioport_resource = { "PCI IO", 0x0000, IO_SPACE_LIMIT, IORESOURCE_IO };
struct resource iomem_resource = { "PCI mem", 0x00000000, 0xffffffff, IORESOURCE_MEM };

struct resource *get_port_by_addr(unsigned long addr)
{
  return(get_resource_by_addr(ioport_resource_tree,addr));
}
struct resource *get_mem_by_addr(unsigned long addr)
{
  return(get_resource_by_addr(iomem_resource_tree,addr));
}
struct resource *get_resource_by_addr(struct resource *list,unsigned long addr)
{
  struct resource *ptr = list;
  int found = FALSE;

  //DEBUGDMESG2("inside get_resource_by_addr %08lx\n",addr);

  while (ptr != NULL && !found)
    {
      //DEBUGDMESG4("looking at whether %08lx is within <%08lx-%08lx>\n",addr,ptr -> start,ptr -> end);
      if (addr >= ptr -> start && addr <= ptr -> end)
	{
	  //DEBUGDMESG("yes!, found is true\n");
	  found = TRUE;
	}
      else
	{
	  ptr = ptr -> next;
	}

    }
  if (!found || ptr == NULL) 
    {
      //printk("address %08lx not found return NULL\n",addr);
      return NULL;
    }
  //DEBUGDMESG("get_resource_by_addr : return ptr\n");
  return (ptr);

}
/*
  struct resource *get_resource_by_s_addr(unsigned long addr)
  {
  struct resource *ptr = io_resource_root;
  
  while (ptr != NULL)
  {
  //printk("ptr -> start %08lx, addr %08lx",ptr -> start,addr);
  if (addr >= ptr -> start && addr <= ptr -> end)
  {
  //printk("found resource,return ptr");
  return(ptr);
  }
  ptr = ptr -> child;
  }
  return (NULL);
}
*/
/*
struct resource *get_conflict(struct resource *res)
{
  struct resource *ptr;
  
  ptr = io_resource_root;

  while (ptr != NULL)
    {
      if (ptr -> start == res -> start)
	{
	  return (ptr);
	}
      ptr = ptr -> child;
    }

  return (NULL);

}
*/
void ddi_error(int error)
{
  switch (error)
    {
    case EAGAIN:
      printk("EAGAIN");
      break;
    case EFAULT:
      printk("EFAULT");
      break;
    case EINTR:
      printk("EINTR");
      break;
    case EINVAL:
      printk("EINVAL");
      break;
    case EIO:
      printk("EIO");
      break;
    case ENXIO:
      printk("ENXIO");
      break;
    case EPERM:
      printk("EPERM");
      break;
    case EROFS:
      printk("EROFS");
      break;
    default:
      printk("unknown error");
    }
}

struct resource * __request_region(int type,
				   unsigned long start, 
				   unsigned long n, 
				   const char *name)
{
  struct resource *res = NULL;

  printk("inside request_region ");
  switch (type)
    {
    case IORESOURCE_IO:
      res = get_resource_by_addr(ioport_resource_tree,start);
      printk("io ");
      break;
    case IORESOURCE_MEM:
      res = get_resource_by_addr(iomem_resource_tree,start);
      printk("mem ");
      break;
      
    default:
      printk("undefined type");
    }
  printk("<%08lx-%08lx>\n",start,(start+n));

  if (res == NULL) 
    {
      DEBUGDMESG("resource not found\n");
      return NULL;
    }

  if (res -> start != start && res -> length != n)
    {
      DEBUGDMESG("resource found, but out of alignment..\n");
      return NULL;
    }
  if (res -> flags & IORESOURCE_BUSY)
    {
      DEBUGDMESG("resource found, but is busy!");
      return NULL;
    }
  res -> flags = res -> flags | IORESOURCE_BUSY;
  res -> name = name;
  
  //printk("resource found, return structure pointer\n");

  return res;

}
int __check_ioregion(unsigned long start,unsigned long n)
{
  return (__check_region(IORESOURCE_IO,start,n));
}
int __check_region(int type, unsigned long start, unsigned long n)
{
	struct resource * res;
       
	//printk("Inside __check_region\n");

	res = __request_region(type,start, n, "check-region");

	if (res==NULL)
	  {
	    
	    return -EBUSY;
	  }
	__release_region(type,start,n);
	printk("Done __check_region\n");

	return 0;
}



void __release_region(int type,unsigned long start, unsigned long n)
{

  struct resource *res = NULL;
  switch (type)
    {
    case IORESOURCE_IO:
      res = get_resource_by_addr(ioport_resource_tree,start);
      break;
    case IORESOURCE_MEM:
      res = get_resource_by_addr(iomem_resource_tree,start);
      break;
    default:
      printk("unknown type\n");
    }
  if (res != NULL)
    {
      res -> flags = res -> flags & ~IORESOURCE_BUSY;
    }
  
  
}
/*
 * Called from init/main.c to reserve IO ports.
 */
#define MAXRESERVE 4
/*
static int __init reserve_setup(char *str)
{
  

  return 1;
}
*/
static void destroyResourceStructure(struct resource *res)
{
  int i;
  if (res)
    {
      for (i=0;i<res -> length;i++)
	{
	  if (res -> sun_handles8) ddi_regs_map_free(res -> sun_handles8[i]);
	}
      for (i=0;i<res -> length - 1;i++)
	{
	  if (res -> sun_handles16) ddi_regs_map_free(res -> sun_handles16[i]);
	}
      for (i=0;i<res -> length - 3;i++)
	{
	  if (res -> sun_handles32) ddi_regs_map_free(res -> sun_handles32[i]);
	}
      if (res -> sun_handles8) {kfree (res -> sun_handles8);}
      if (res -> mapped_addrs8) {kfree(res -> mapped_addrs8);}
      if (res -> sun_handles16) {kfree(res -> sun_handles16);}
      if (res -> mapped_addrs16) {kfree(res -> mapped_addrs16);}
      if (res -> sun_handles32) {kfree(res -> sun_handles32);}
      if (res -> mapped_addrs32) {kfree(res -> mapped_addrs32);}

      kfree(res);
    }
  
}
static struct resource* createResourceStructure(dev_info_t *dip,
						int regindex,
						int base,
						int size,
						int flags)
{

  struct resource *res = NULL;
  int i,status;
  //printk("allocating resource structure");
  int num8,num16,num32;
  

  num8 = size;
  num16 = size - 1;
  num32 = size - 3;

  


  res = kmalloc(sizeof(struct resource), GFP_KERNEL);
  if (res == NULL)
    {
      printk("Error allocating res structure!\n");
      goto error;
    }
  
  res -> sun_handles8 = kmalloc(sizeof(ddi_acc_handle_t)*num8,GFP_KERNEL);
  res -> mapped_addrs8 = kmalloc(sizeof(uint8_t*)*num8,GFP_KERNEL);
  
  if (!res -> sun_handles8 || !res -> mapped_addrs8)
    {
      printk("Error allocating sun_handles8 or mapped_addrs8\n");
      goto error;
    }


  if (num16 > 0)
    {
      res -> sun_handles16 = kmalloc(sizeof(ddi_acc_handle_t)*num16,GFP_KERNEL);
      res -> mapped_addrs16 = kmalloc(sizeof(uint16_t*)*num16,GFP_KERNEL);
      if (!res -> sun_handles16 || !res -> mapped_addrs16)
	{
	  printk("Error allocating sun_handles16 or mapped_addrs16\n");
	  goto error;
	}
    }

  if (num32 > 0)
    {
      res -> sun_handles32 = kmalloc(sizeof(ddi_acc_handle_t)*num32,GFP_KERNEL);
      res -> mapped_addrs32 = kmalloc(sizeof(uint32_t*)*num32,GFP_KERNEL);
      if (!res -> sun_handles32 || !res -> mapped_addrs32)
	{
	  printk("Error allocating sun_handles32 or mapped_addrs32\n");
	  goto error;
	}
    }

  res -> name = "";
  res -> start = base;
  res -> end   = base+size - 1;
  res -> length = size;
  //res -> child = NULL;
  //res -> parent = NULL;
  
  res -> next = NULL;
  res -> regindex = regindex;
  //printk("attempting to map <%08lx-%08lx>",res -> start, res -> end);
  
  //printk("node name:%s",ddi_node_name(pits_globals.current_dip));
  //printk("binding name:%s",ddi_binding_name(pits_globals.current_dip));
	  
  for (i = 0;i < num8;i++)
    {
      status = ddi_regs_map_setup(dip,
				  regindex,
				  (caddr_t *)&(res -> mapped_addrs8[i]),
				  (offset_t)i,
				  sizeof(uint8_t),
				  //&pits_globals.sun_dev_acc_attr,
				  &(getGlobal() -> sun_dev_acc_attr),
				  &(res -> sun_handles8[i]));
      if (status != DDI_SUCCESS)
	{
	  printk("mapping failed:base %08lx offset: %d size: 8\n",res -> start,i);
	  printk("error code:%d ",status);
	  ddi_error(status);
	  printk("\n");
	  //printk("errmsg:%s",strerror(status));
	  goto error;
	  //return NULL;
	}
    }/*for i= */
  for (i = 0;i < num16;i++)
    {
      status = ddi_regs_map_setup(dip,
				  regindex,
				  (caddr_t *)&(res -> mapped_addrs16[i]),
				  (offset_t)i,
				  sizeof(uint16_t),
				  //&pits_globals.sun_dev_acc_attr,
				  &(getGlobal() -> sun_dev_acc_attr),
				  &(res -> sun_handles16[i]));
      if (status != DDI_SUCCESS)
	{
	  printk("mapping failed:base %08lx offset: %d size: 16\n",res -> start,i);
	  printk("error code:%d",status);
	  ddi_error(status);
	  printk("\n");
	  //printk("errmsg:%s",strerror(status));
	  goto error;
	  //return NULL;
	}
    } /* for i= */

  for (i = 0;i < num32;i++)
    {
      status = ddi_regs_map_setup(dip,
				  regindex,
				  (caddr_t *)&(res -> mapped_addrs32[i]),
				  (offset_t)i,
				  sizeof(uint32_t),
				  //&pits_globals.sun_dev_acc_attr,
				  &(getGlobal() -> sun_dev_acc_attr),
				  &(res -> sun_handles32[i]));
      if (status != DDI_SUCCESS)
	{
	  printk("mapping failed:base %08lx offset: %d size: 32\n",res -> start,i);
	  printk("mapping failed:");
	  printk("error code:%d",status);
	  ddi_error(status);
	  //printk("errmsg:%s",strerror(status));
	  printk("\n");
	  goto error;
	  return NULL;
	}
    } /* for i= */


  res -> flags = flags;
  
  //printk("mapping was a success!");
  return res;
  
 error:
  
  if (res)
    {
      if (res -> sun_handles8) {kfree(res -> sun_handles8);}
      if (res -> mapped_addrs8) {kfree(res -> mapped_addrs8);}
      if (res -> sun_handles16) {kfree(res -> sun_handles16);}
      if (res -> mapped_addrs16) {kfree(res -> mapped_addrs16);}
      if (res -> sun_handles32) {kfree(res -> sun_handles32);}
      if (res -> mapped_addrs32) {kfree(res -> mapped_addrs32);}
      kfree(res);
    }
  return NULL; 
}

static void addResource(struct resource **list,struct resource *element)
{
  struct resource *ptr;
  if (*list == NULL)
    {
      *list = element;
    }
  else
    {
      ptr = *list;
      while (ptr -> next != NULL)
	{
	  ptr = ptr -> next;
	}
      ptr -> next = element;
    }
  //element -> next = NULL;


}


static int init_isa_resources(dev_info_t *dip)
{
  int *reg_array = NULL;
  char *parent = NULL;
  dev_info_t *pdip = ddi_get_parent(dip);
  int nreg;
  int nelements;
  int ret = 0;
  //int nport;
  int nmem;
  int i;
  int rtype,rbase,rsize;
  struct resource *temp_resource = NULL;

  //ioport_resource_tree = NULL;
  //iomem_resource_tree = NULL;

  //get the register property.
  if (!ddi_prop_exists(DDI_DEV_T_ANY,             //dev_t.
		       dip,                       //dev_info_t structure.
		       DDI_PROP_DONTPASS,         //flags.
                       "reg"))                    //parameter name.
    {
      DEBUGDMESG("prop reg doesn't exist, return DDI_FAILURE\n");
      goto error;
    }
  ret = ddi_prop_lookup_int_array(DDI_DEV_T_ANY,
				  dip,
				  DDI_PROP_DONTPASS,
				  "reg",
				  &reg_array,
				  &nelements);
  if (ret != DDI_SUCCESS) 
    {
      goto error;
    }

  //printk("Success, map the registers.");
  nreg = nelements / 3;
      
  for (i=0;i<nreg;i++)
    {
      //extract the tuple.
      rtype = reg_array[i*3];    
      rbase = reg_array[i*3 + 1];
      rsize = reg_array[i*3 + 2];

      switch (rtype)
	{
	case 0: //memory range
	  nmem++;
	  temp_resource = createResourceStructure(dip,i,rbase,rsize,IORESOURCE_MEM);
	  if (temp_resource)
	    addResource(&iomem_resource_tree,temp_resource);
	  break;
	case 1:
	  //add ioport resource to ioports list.
	  temp_resource = createResourceStructure(dip,i,rbase,rsize,IORESOURCE_IO);
	  if (temp_resource)
	    addResource(&ioport_resource_tree,temp_resource);
	  break;
	default:
	  printk("bad reg type %d\n",rtype);
	  ret = DDI_FAILURE;
	  goto error;
	} //switch(rtype)
    } //for (i=0

  ddi_prop_free(reg_array);

  return (DDI_SUCCESS);

 error:
  if (reg_array)
    {
      ddi_prop_free(reg_array);
    }
  return (ret);

}

int getIntArray(dev_info_t *dip,char *name,int **reg_array,int *nelements)
{

  int ret;
  //get the register property.
  if (!ddi_prop_exists(DDI_DEV_T_ANY,             //dev_t.
                       dip,                       //dev_info_t structure.
                       DDI_PROP_DONTPASS,         //flags.
                       name))                    //parameter name.
    {
      DEBUGDMESG2("prop %s doesn't exist, return DDI_FAILURE\n",name);
      return DDI_FAILURE;
    }
  ret = ddi_prop_lookup_int_array(DDI_DEV_T_ANY,
                                  dip,
                                  DDI_PROP_DONTPASS,
                                  name,
                                  reg_array,
                                  nelements);
  return ret;
}

int copyResourceStructure(struct resource *dest,struct resource *src)
{
  
  int num8,num16,num32;

  DEBUGDMESG("Inside copyResourceStructure()");

  if (!dest)
    {
      printk("copyResourceStructure: dest is null!\n");
      return DDI_FAILURE;
    }
  if (!src)
    {
      printk("copyResourceStructure: src is null!\n");
      return DDI_FAILURE;
    }
  
  num8 = src -> length;
  num16 = num8 - 1;
  num32 = num8 - 3;

  DEBUGDMESG("copying 8");
  if (!src -> sun_handles8)
    {
      printk("copyResourceStructure: sun_handles8 is null!");
      return DDI_FAILURE;
    }
  if (!src -> mapped_addrs8)
    {
      printk("copyResourceStructure: mapped_addrs8 is null!");
      return DDI_FAILURE;
    }
  printk("allocating sun_handles8 and mapped_addrs8\n");
  dest -> sun_handles8 = kmalloc(sizeof(ddi_acc_handle_t)*num8,GFP_KERNEL);
  dest -> mapped_addrs8 = kmalloc(sizeof(uint8_t*)*num8,GFP_KERNEL);
  
  printk("copying handles\n");
  bcopy(src -> sun_handles8,dest -> sun_handles8,sizeof(ddi_acc_handle_t)*num8);
  printk("copying addrs\n");
  bcopy(src -> mapped_addrs8,dest -> mapped_addrs8,sizeof(uint8_t*)*num8);

  printk("allocating sun_handles16 and mapped_addrs16\n");
  dest -> sun_handles16 = kmalloc(sizeof(ddi_acc_handle_t)*num16,GFP_KERNEL);
  dest -> mapped_addrs16 = kmalloc(sizeof(uint16_t*)*num16,GFP_KERNEL);
  printk("copying handles\n");
  bcopy(src -> sun_handles16,dest -> sun_handles16,sizeof(ddi_acc_handle_t)*num16);
  printk("copying addrs\n");
  bcopy(src -> mapped_addrs16,dest -> mapped_addrs16,sizeof(uint16_t*)*num16);

  printk("allocating sun_handles32 and mapped_addrs32\n");

  dest -> sun_handles32 = kmalloc(sizeof(ddi_acc_handle_t)*num32,GFP_KERNEL);
  dest -> mapped_addrs32 = kmalloc(sizeof(uint32_t*)*num32,GFP_KERNEL);
  printk("copying handles\n");
  bcopy(src -> sun_handles32,dest -> sun_handles32,sizeof(ddi_acc_handle_t)*num32);
  printk("copying addrs\n");
  bcopy(src -> mapped_addrs32,dest -> mapped_addrs32,sizeof(uint32_t*)*num32);

  DEBUGDMESG("copying name");
  dest -> name = src -> name;
  
  DEBUGDMESG("copying rest");
  dest -> start = src -> start;
  dest -> end = src -> end;
  dest -> length = src -> length;
  dest -> flags = src -> flags;
  
  
  DEBUGDMESG("Done copyResourceStructure()");

  return DDI_SUCCESS;
}

struct pci_dev *createPCIDevice(dev_info_t *dip)
{
  struct pci_dev *dev;
  uint16_t vendor;
  uint16_t device;
  uint16_t subdevice;
  uint16_t subvendor;
  int i;
  DEBUGDMESG("inside createPCIDevice");

  dev = kmalloc(sizeof(struct pci_dev),GFP_KERNEL);
  dev -> bus = kmalloc(sizeof(struct pci_bus),GFP_KERNEL);
  //printk("device count resource %d\n",DEVICE_COUNT_RESOURCE);
  dev -> resource = kmalloc(sizeof(struct resource)*DEVICE_COUNT_RESOURCE,GFP_KERNEL);
  
  pci_config_setup(dip,&(dev -> config_handle));
  //read configuration info to fill in structure.
  pci_read_config_word(dev,PCI_VENDOR_ID,&vendor);
  pci_read_config_word(dev,PCI_DEVICE_ID,&device);
  pci_read_config_word(dev,PCI_SUBSYSTEM_VENDOR_ID,&subvendor);
  pci_read_config_word(dev,PCI_SUBSYSTEM_ID,&subdevice);


  
  dev -> vendor = vendor;
  dev -> device = device;
  dev -> subsystem_vendor = subvendor;
  dev -> subsystem_device = subdevice;
  dev -> dma_mask = 0xffffffff;
  dev -> bus -> number = 0;
  DEBUGDMESG("done createPCIDEVICE");

  return dev;

}

int init_pci_resources(dev_info_t *dip)
{
  DEBUGDMESG("inside init_pci_resources\n");
  int *reg_array=NULL;
  int *addr_array = NULL;
  int i;
  int nelements;
  int nreg;
  sun_linux_state *instance_state = NULL;

  struct resource *temp_resource;
  int numb;
  int func;
  int dev;
  int bus;
  int id;
  ddi_acc_handle_t chandle;
  char *type;
  uint64_t raddr,rsize;
  int ret;
  uint32_t tag;
  uint32_t haddr,laddr;
  uint32_t hsize,lsize;
  struct pci_dev *pdev;

  getInstanceFReturn(dip,instance_state);


  pdev = createPCIDevice(dip);
  if (!pdev) 
    {
      printk("pdev doesn't exist!\n");
      return DDI_FAILURE;
    }
  instance_state -> linux_pci_device = pdev;
  ret = getIntArray(dip,"reg",&reg_array,&nelements);
  ret |= getIntArray(dip,"assigned-addresses",&addr_array,&nelements);
  ret |= ddi_dev_nregs(dip,&nreg);

  //the first register is the configuration space.
  if (ret == DDI_SUCCESS)
    {
      
      //nreg = nelements/5;
      for (i=0;i<nreg-1;i++)
	{
	  tag = addr_array[i*5];
	  haddr = addr_array[i*5+1];
	  laddr = addr_array[i*5+2];
	  hsize = addr_array[i*5+3];
	  lsize = addr_array[i*5+4];
	  //printk("tag 0x%08lx haddr 0x%08lx laddr 0x%08lx hsize 0x%08lx lsize 0x%08lx\n",tag,haddr,laddr,hsize,lsize); 
	  
	  id = (tag & 0x03000000) >> 24;

	  raddr = ((uint64_t)haddr << 32) + (uint64_t)laddr;
	  rsize = ((uint64_t)hsize << 32) + (uint64_t)lsize;
	  printk("i: %d id : %d raddr 0x%08lx rsize 0x%08lx\n",i,id,(uint32_t)raddr,(uint32_t)rsize);
	  
	  switch (id)
	    {
	    case 1: /* I/O Port*/
	      
	      temp_resource = createResourceStructure(dip,i+1,raddr,rsize,IORESOURCE_IO);
	      if (temp_resource)
		{
		  copyResourceStructure(&(pdev -> resource[i]),temp_resource);
		  addResource(&ioport_resource_tree,temp_resource);
		}
	      break;
	    case 2: /* I/O Memory */
	      temp_resource = createResourceStructure(dip,i+1,raddr,rsize,IORESOURCE_MEM);
	      if (temp_resource)
		{
		  copyResourceStructure(&(pdev -> resource[i]),temp_resource);
		  //add memory resource to iomems list.
		  addResource(&iomem_resource_tree,temp_resource);
		}
	      break;
	    default:
		  printk("unsupported io type %d",id);
	    }
	    
	} /*for (i=0;i<nreg-1;i++)*/
      
    }

  //now initialize dma.
  printk("initializing dma\n");
  init_pci_dma(dip);

  DEBUGDMESG("done init_pci_resources");
  return ret;
}

int init_resources(dev_info_t *dip)
{
  //right now we only support isa type cards, i.e. the parallell port is isa.
  //we can change this to init based upon bus later.
  int is_pci = FALSE;
  dev_info_t *pdip = ddi_get_parent(dip);
  char *parent= NULL;
  int ret;
  if (!ddi_prop_exists(DDI_DEV_T_ANY,             //dev_t.
                       pdip,                       //dev_info_t structure.
                       DDI_PROP_DONTPASS,         //flags.
                       "name"))                    //parameter name.
    {
      DEBUGDMESG("prop parent name doesn't exist, assume isa\n");
      //goto error;
    }
  else
    {
      ret = ddi_prop_lookup_string(DDI_DEV_T_ANY,
                                   pdip,
                                   DDI_PROP_DONTPASS,
                                   "name",
                                   &parent);
      if (ret == DDI_SUCCESS)
        {
          DEBUGDMESG2("prop parent is:%s\n",parent);
	  if (parent[0] == 'p' && parent[1] == 'c')
	    {
	      is_pci = TRUE;
	    }
	  ddi_prop_free(parent);
        }
  
    }

  if (is_pci)
    {
      ret = init_pci_resources(dip);
    }
  else
    {
      ret = init_isa_resources(dip);
    }

  if (ret == DDI_SUCCESS)
    {
      ret = set_initialized(dip,RESOURCE_INIT);
    }
  
  return(ret);
}

void fini_resource_list(struct resource *list)
{


  struct resource *ptr = list;
  struct resource *del = NULL;
  
  if (ptr != NULL)
    {
      while (ptr -> next != NULL)
	{
	  destroyResourceStructure(ptr);
	  del = ptr;
	  ptr = ptr -> next;
	  kfree(del);
	  del = NULL;
	}
    }
}

int fini_resources(dev_info_t *dip)
{
  printk("fini_resources\n");
  printk("releasing ioports\n");
  //fini_resource_list(ioport_resource_tree);
  printk("releasing iomemory\n");
  //fini_resource_list(iomem_resource_tree);

  return (DDI_SUCCESS);

}
