/**************************************/
/*sunlinux.c                          */
/*Barry Henrichs                      */
/*Sean McIlwain                       */
/*CS736 - Advanced Operating Systems  */
/*Spring '2003                        */
/*Final Project                       */
/**************************************/
/*sunlinuxdrv include*/
#include <misc/sunlinuxdrv.h>
#include <misc/instminor.h>
#include <misc/tasklets.h>
#include <misc/pits_init.h>
#include <misc/pits_param.h>
#include <misc/pits_block.h>
#include <misc/pits_char.h>
#include <misc/pits_dev_inst.h>
#include <misc/kernel.h>
#include <misc/slab.h>

/*sun include*/
#include <sys/types.h>
/*#include <sys/file.h>*/
#include <sys/errno.h>
#include <sys/open.h>
#include <sys/cred.h>
#include <sys/ddi.h>
#include <sys/sunddi.h>
//#include <stdarg.h>  /*for va_list */ 
#include <sys/signal.h>
#include <unistd.h>

#include <sys/stropts.h>

/*linux include*/
#include <linux/fs.h>
//#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/fcntl.h>
#include <linux/poll.h>


static void *sun_state=NULL; //look in sunlinux.c

void *getSunStateP()
{
  return (sun_state);
}


/*fs.h */
int fasync_helper(int fd, struct file *filp, 
		  int mode, struct fasync_struct **fapp)
{
  /*so if mode is 0 -> remove pid from fasync_struct*/
  /*if mode is non-zero, then find the pid in fa*/
  /*if not found, create a new structure and it to the list*/
  /*if found,change the mode to the new mode               */

  struct fasync_struct *fa, **fp;
  
  DEBUGDMESG("Inside fasync_helper\n");

  /* this is cool..... */
  /* http://www.cab.u-szeged.hu/linux/kernel/linux/drivers/char/tty_io.c.html */

  for (fp = fapp; (fa = *fp) != NULL; fp = &fa->fa_next) {
    if (fa->fa_owner.pid == filp -> f_owner.pid)
      break;
  }

  /*if the mode is zero, we want to remove the node.*/
  if (mode == 0)
    {
      /*node doesn't exist..., return 0 */
      if (fa == NULL) 
	{
	  return 0;
	}
      else
	{
	  DEBUGDMESG2("removing process %d from asynch\n",fa -> fa_owner.pid);
	  /*change pointer, free fa, and return 0*/
	  *fp = fa -> fa_next;
	  proc_unref(fa -> proc_ref);
	  kfree(fa);
	  return 0;
	}
      
    }
  else
    {
      //does fa already exist?
      //no,create a new one.
      if (fa == NULL)
	{
	  DEBUGDMESG("Creating new node\n");
	  fa = kmalloc(sizeof(struct fasync_struct),0);
	  fa -> fa_file = filp;
	  fa -> fa_owner.pid = filp -> f_owner.pid;
	  fa -> proc_ref = proc_ref();
	  fa -> fa_fd = fd;
	  fa -> bitmask = mode;

	  //add to the front.
	  fa -> fa_next = *fapp;
	  *fapp = fa;
	  return (0);
	}
      //yes,change bitmask
      else
	{
	  fa -> bitmask = mode;
	  return (0);
	}
      
    }

  

  DEBUGDMESG("Done fasync_helper\n");

  return 0;
}

void kill_fasync(struct fasync_struct **fa, int sig, int band)
{
  /*for every pid in the fa structure, send the signal, corresponding*/
  /*to what the user wants notification for, according to the bitmask.*/
  struct fasync_struct *fp = *fa;

  int pid = 0;
  int bitmask;
  DEBUGDMESG("Inside kill_fasync\n");
  
  while (fp != NULL)
    {
      pid = fp -> fa_owner.pid;
      bitmask = fp -> bitmask;

      if (band == POLL_IN)
	{
	  if ((bitmask & S_INPUT) || (bitmask & S_RDNORM))
	    {
	      //signal for POLL_IN;
	      DEBUGDMESG2("Sending SIGPOLL for POLL_IN to pid: %d\n",pid);
	      	      proc_signal(fp -> proc_ref,SIGPOLL);

	    }

	}
      if (band == POLL_OUT)
	{
	  if ((bitmask & S_OUTPUT) || (bitmask & S_WRNORM))
	    {
	      proc_signal(fp -> proc_ref,SIGPOLL);
	    }
	}


      fp = fp -> fa_next;
    }
  

  DEBUGDMESG("Done kill_fasync\n");
}

/**********************************************/
/*solaris calls                               */
/**********************************************/
int _init() 
{
  int error;
  //  const int max_instance = 10;
  
  
  DEBUGDMESG("_init()\n");

  //global_state.initialized = FALSE;
  
  //init_pits_printk(NULL);
  DEBUGCMD(printk("_init()\n"));
  //init_pits_slab(NULL);
  init_pits_globals();

  error = mod_install(&sun_linkage);

  if (error !=0)
    {
      printk("error init: %d\n",error);
      fini_pits_globals();
      //fini_pits_slab(NULL);
      //fini_pits_printk(NULL);
    }
  else
    {
      ddi_soft_state_init(&sun_state,sizeof(sun_linux_state),16);
    }

  DEBUGCMD(printk("end _init()\n"));

  return(error);
}

int _fini()
{
  int error;

  DEBUGDMESG("_fini()\n");

  error = mod_remove(&sun_linkage);

  if (error != 0)
    {
      printk("end do_fini() error\n");
      return(error);
    }
  ddi_soft_state_fini(&sun_state);
  fini_pits_globals();
  //fini_pits_slab(NULL);
  //fini_pits_printk(NULL);
  
  DEBUGDMESG("end do_fini()\n");
  
  return(0);

}

int _info(struct modinfo *modinfop)
{
  int error = 0;
  DEBUGDMESG("_info called\n");
  
  error = mod_info(&sun_linkage,modinfop);
  
  if (error == 0)
    {
      printk("_info():Error with mod_info\n");
    }

  return (error);
}


/*****************************************************/
/*sun_getinfo : follows the getinfo(9E) routine      */
/*****************************************************/
static int sun_getinfo(dev_info_t *dip,ddi_info_cmd_t infocmd, void *arg, void **result)
{
  sun_linux_state *instance_state = NULL;
  int error;
  //int instance = getInstance(minor);
  int instance = 0;
  //DEBUGDMESG("inside sun_getinfo()\n");

  if (instance < 0)
    {
      DEBUGDMESG("Instance doesn't exist!\n");
      DEBUGDMESG("not in the instminor structure\n");
      return (DDI_FAILURE);
    }

  switch (infocmd)
    {
      /*
       * return the device node if the driver has attached the driver
       * instance identified by the dev_t value which was passed  
       */
    case DDI_INFO_DEVT2DEVINFO:
      //DEBUGDMESG("DDI_INFO_DEVT2DEVINFO\n");
      instance_state = ddi_get_soft_state(sun_state,instance);
      if (instance_state == NULL)
	{
	  *result = NULL;
	  error = DDI_FAILURE;
	}
      else
	{
	  *result = instance_state -> dip;
	  error = DDI_SUCCESS;
	}
      break;
      /* 
       *the driver can always return the instance number given a dev_t
       *value, even if the instance is not attached. */

    case DDI_INFO_DEVT2INSTANCE:
      //DEBUGDMESG("DDI_INFO_DEV2INSTANCE\n");
      *result = (void *)instance;
      error = DDI_SUCCESS;
      break;
    default:
      *result = NULL;
      return DDI_FAILURE;
    }



  //DEBUGDMESG("done sun_getinfo()\n");


  return error;
} 
static int sun_probe(dev_info_t *dip)
{
  return 0;
}

/******************************************************
 *sun_attach()                                        *
 ******************************************************/
static int sun_attach(dev_info_t *dip, ddi_attach_cmd_t cmd)
{
  int instance = 0;
  int error = 0;
  sun_linux_state *instance_state = NULL;
  //  char *minorname = NULL;
  //  int minornumber = 0;
  
  //  int num_minor = 0;
  //  int *minor_array = NULL;
  //  int minor_index = 0;

  DEBUGDMESG("sun_attach\n");
  switch (cmd)
    {
    case DDI_ATTACH:
      
      DEBUGDMESG("DDI_ATTACH:\n");
      
      instance = ddi_get_instance(dip);  
      
      error = ddi_soft_state_zalloc(sun_state,instance);   
      if (error != 0)   
	{ 
	  printk("error allocating softstate %d\n",error); 
	  return DDI_FAILURE;   
	}   

      getInstanceFReturn(dip,instance_state);


      instance_state -> dip         = dip;  
      instance_state -> instance    = instance;  
      instance_state -> initialized = 0;
      instance_state -> pits_major  = ddi_driver_major(dip);
      printk("major : %d\n",instance_state -> pits_major);
      mutex_init(&(instance_state -> mutex),
		 NULL,
		 MUTEX_DRIVER,
		 NULL);
      getGlobal() -> current_dip = dip;
      //pits_globals.current_dip = dip;

      error = do_pits_init(dip);
      if (error != 0)
	{
	  printk("Error initializing linux driver %d\n",error);
	  do_pits_fini(dip);
	  ddi_soft_state_free(sun_state,instance);
	  return DDI_FAILURE;
	}
      if (getGlobal() -> panic == TRUE)
	{
	  printk("Linux panic detected. Driver unstable...\n");
	  ddi_soft_state_free(sun_state,instance);
	  return DDI_FAILURE;
	}
      

      instance_state -> num_open = 0;
      //instance_state -> fops = global_state.fops;
      ddi_report_dev(dip);
      //pits_globals.attach_count++;
      getGlobal() -> attach_count++;
      
      break;
    case DDI_RESUME:
      return (DDI_SUCCESS);
    default:
      return (DDI_FAILURE);
    }
  
  
  DEBUGDMESG("end sun_attach\n");
  
  return error;
} 

/**********************************************************
 *sun_detach()                                            *
 **********************************************************/
static int sun_detach(dev_info_t *dip,ddi_detach_cmd_t cmd)
{
  int instance; 
  //  int error = 0; 
 
  sun_linux_state *instance_state; 

  /*
  DEBUGDMESG("sun_detach");
  */
  switch (cmd)
    {
    case DDI_DETACH:

      /*
      DEBUGDMESG("DDI_DETACH");
      */
      
      instance = ddi_get_instance(dip);
      printk("instance is %d\n",instance);
      printk("getting instance_state\n");
      instance_state = ddi_get_soft_state(sun_state,0);

      //DEBUGDMESG("Removing from MinorInst");
      //remInstMinorByInstance(instance);

      //ddi_remove_minor_node(dip,NULL);

      getGlobal() -> attach_count --;
      //pits_globals.attach_count--;

      if (getGlobal() -> attach_count <= 0)
	{
	  DEBUGDMESG("No more attached nodes, uninitialize linux driver\n");
	  do_pits_fini(dip);
	}
      mutex_destroy(&(instance_state -> mutex)); 
      ddi_soft_state_free(sun_state,instance);
      break;
    case DDI_SUSPEND:
    default:
      return (DDI_FAILURE);
    }

  /*
  DEBUGDMESG("end sun_detach");
  */
  return (DDI_SUCCESS);
}



/************************************************************/
/*sun_open() : open the file, gets it ready for reading     */
/*calls the corresponding linux fops->open for the instance.*/
/************************************************************/
static int sun_open(dev_t *sdevp,int flag,int otyp,cred_t *credp)
{
  
  sun_linux_state *instance_state = NULL;
  int ret = 0;
  //printk("inside sun_open()\n");

  returnPanicValue("pits_open",ENXIO);
  
  instance_state = ddi_get_soft_state(getSunStateP(),0);
  if (instance_state == NULL) 
    {
      printk("pits_open(): NULL instance\n");
      return ENXIO;
    }
  
  mutex_enter(&(instance_state -> mutex));
  switch (otyp)
    {
    case OTYP_CHR:
      if (instance_state -> fops == NULL || instance_state -> fops -> open == NULL)
	{
	  if (instance_state -> bdops != NULL && instance_state -> bdops -> open != NULL)
	    {
	      ret = pits_block_open(sdevp,flag,OTYP_BLK,credp);
	    }
	  else
	    {
	      printk("open type: char, no method defined!\n");
	      printk("need a register_chrdev or register_blkdev!\n");
	      goto mutex_error;
	    }
	}
      else
	{
	  ret = pits_char_open(sdevp,flag,otyp,credp);
	}
      break;
    case OTYP_BLK:
    case OTYP_LYR:
      ret = pits_block_open(sdevp,flag,otyp,credp);
      break;
    default:
      goto mutex_error;
    }
  mutex_exit(&(instance_state -> mutex));
  return (ret);

 mutex_error:
  mutex_exit(&(instance_state -> mutex));
  printk("pits_open error: %d\n",ENXIO);
  return (ENXIO);
}

/**************************************************************/
/*sun_close()                                                 */
/**************************************************************/
static int sun_close(dev_t sdev,int flag,int otyp,cred_t *credp)
{
  sun_linux_state *instance_state = NULL;
  linux_file *open_file;
  linux_inode *open_inode;
  int ret;
  major_t smajor;
  minor_t sminor;
  int instance = 0;

  returnPanicValue("pits_close",ENXIO);

  sminor = getminor(sdev);

  

  instance_state = ddi_get_soft_state(getSunStateP(),instance);
  if (instance_state == NULL)
    {
      printk("Null instance?\n");
      return(ENXIO);
    }
  
  open_file  = getOpenFileS(smajor,sminor);
  open_inode = getOpenInodeS(smajor,sminor);

  if (open_file == NULL || open_file -> usage_count == 0)
    {
      printk("couldn't find file major %d minor %d\n",smajor,sminor);
      return(DDI_SUCCESS);
    }
  
  mutex_enter(&(instance_state -> mutex));
  switch (open_file -> sun_opentype)
    {
    case OTYP_CHR:
      printk("OTYP_CHR:%d\n",otyp);
      ret = pits_char_close(sdev,flag,otyp,credp);
      break;
    case OTYP_BLK:
    case OTYP_LYR:
      //printk("OTYP_BLK:%d\n",otyp);
      ret = pits_block_close(sdev,flag,otyp,credp);
      break;
    default:
      printk("Unknown:%d\n",otyp);
      ret = EINVAL;
    }
 
  mutex_exit(&(instance_state -> mutex));
  return ret;

}

/*****************************************************/
/*sun_read()                                         */
/*****************************************************/
static int sun_read(dev_t sdev,struct uio *uio_p,cred_t *credp)
{
  sun_linux_state *instance_state;

  linux_file *open_file;
  linux_inode *open_inode;
  major_t smajor;
  minor_t sminor;
  int instance;

  //DEBUGDMESG("inside sun_read()\n");

  returnPanicValue("pits_read",ENXIO);

  sminor = getminor(sdev);
  smajor = getmajor(sdev);
  instance = 0;
  if (instance <0 || instance > 16)
    {
      return (EINVAL);
    }
  instance_state = ddi_get_soft_state(sun_state,instance);
  if (instance_state == NULL)
    {
      printk("Null instance?\n");
      return(ENXIO);
    }
  open_file = getOpenFileS(smajor,sminor);
  open_inode = getOpenInodeS(smajor,sminor);
  if (open_file == NULL || open_file -> usage_count == 0)
    {
      printk("couldn't find file major %d minor %d\n",smajor,sminor);
      return(ENXIO);
    }
  
  switch (open_file -> sun_opentype)
    {
    case OTYP_CHR:
      return(pits_char_read(sdev,uio_p,credp));
    case OTYP_BLK:
    case OTYP_LYR:
      return(pits_block_read(sdev,uio_p,credp));
    default:
      printk("open type is hashed\n");
      return(ENXIO);
    }



}

/*************************************************************/
/*sun_write()                                                */
/*************************************************************/
static int sun_write(dev_t dev,struct uio *uio_p, cred_t *credp)
{
  sun_linux_state *instance_state;
  linux_file *open_file;
  int minornum,instance;
  
  //DEBUGDMESG("inside sun_write\n");
  
  returnPanicValue("pits_write",ENXIO);
  
  minornum = getminor(dev);
  instance = 0;
  if (instance < 0 || instance > 16)
    {
      return (EINVAL);
    }
  instance_state = ddi_get_soft_state(sun_state,instance);
  if (instance_state == NULL)
    {
      return (ENXIO);
    }
  open_file = getOpenFile(instance_state -> linux_major,minornum);
  if (open_file == NULL || open_file -> usage_count == 0)
    {
      return (ENXIO);
    }
  switch (open_file -> sun_opentype)
    {
    case OTYP_CHR:
      return (pits_char_write(dev,uio_p,credp));
    case OTYP_BLK:
    case OTYP_LYR:
      return (pits_block_write(dev,uio_p,credp));
    default:
      return (ENXIO);
    }

  
}

/******************************************/
/*sun_chpoll                              */
/******************************************/
static int sun_chpoll(dev_t dev,short events,int anyyet, 
                      short *reventsp,struct pollhead **phpp)
{
  uint8_t status;
  short revent;
  sun_linux_state *instance_state;
  int minornum;
  int instance;
  linux_file *open_file;

  poll_table *linux_poll_table=NULL;

  DEBUGDMESG("inside sun_chpoll\n");

  returnPanicValue("pits_chpoll",ENXIO);

  minornum = getminor(dev);
  //instance = getInstance(minornum);
  instance = 0;
  if (instance < 0)
    {
      return ENXIO;
    }

  instance_state = ddi_get_soft_state(sun_state,instance);

  if (instance_state == NULL)
    {
      return (ENXIO);
    }
  revent = 0;

  /* valid events are:
   * POLLIN | POLLOUT | POLLPRI | POLLHUP | POLLERR
   */

  open_file = getOpenFile(instance_state -> linux_major,minornum);
  if (open_file == NULL)
    {
      return (ENXIO);
    }

  status = open_file -> f_op -> poll(open_file,linux_poll_table);
  
  DEBUGDMESG3("events is:%d status is:%d\n",events,status);

  if (events & POLLPRI)
    {
      DEBUGDMESG("POLLPRI\n");
    }
  if (events & POLLHUP)
    {
      DEBUGDMESG("POLLHUP\n");
    }
  if (events & POLLERR)
    {
      DEBUGDMESG("POLLERR\n");
    }

  if ((events & POLLIN) || (events & POLLRDNORM))
    {
      DEBUGDMESG("Checking for readable\n");
      if (status & POLLIN)
	{
	  DEBUGDMESG("is readable\n");
	  revent |= POLLIN;
	  revent |= POLLRDNORM;
	}
      else
	{
	  DEBUGDMESG("is not readable\n");
	}
    }
    if ((events & POLLOUT) || (events & POLLWRNORM)) 
    {
      DEBUGDMESG("Checking for writable\n");
      
      if (status & POLLOUT)
	{
	  DEBUGDMESG("is writable\n");
	  revent |= POLLOUT;
	}
      else
	{
	  DEBUGDMESG("is not writable\n");
	}
    }
  /*if nothing has occured*/
  if (revent == 0)
    {
      if (!anyyet)
	{
	  *phpp = &instance_state -> phead;
	}
    }
  *reventsp = revent;

  DEBUGDMESG("Done chpoll\n");

  return(0);
     
}

/***************************************************
 *pits_strategy()
 ***************************************************/
static int pits_strategy(struct buf *bp)
{
  int instance = 0;
  sun_linux_state *instance_state = NULL;
  linux_file *open_file = NULL;
  minor_t minor = getminor(bp -> b_edev);

  returnPanicValue("pits_strategy",ENXIO);


  instance_state = ddi_get_soft_state(sun_state,instance);
  if (instance_state == NULL) {return ENXIO;}
  
  open_file = getOpenFile(instance_state -> linux_major,minor);
  if (open_file == NULL || open_file -> usage_count == 0)
    {
      return (ENXIO);
    }
  switch (open_file -> sun_opentype)
    {
    case OTYP_CHR:
      return (pits_char_strategy(bp));
    case OTYP_BLK:
    case OTYP_LYR:
      return (pits_block_strategy(bp));
    default:
      return (ENXIO);
    }
  
  
  
  
  
}

/***************************************************
 *sun_ioctl() : pg.158 of the manual.
 ***************************************************/
static int sun_ioctl(dev_t dev, int cmd, intptr_t arg, 
		     int mode,cred_t *credp, int *rvalp)
{
  sun_linux_state *instance_state = NULL;
  int instance=0;
  int minor;

  linux_file *open_file;

  printk("inside sun_ioctl\n");
  returnPanicValue("pits_ioctl",ENXIO);

  instance_state = ddi_get_soft_state(sun_state,instance);
  if (instance_state == NULL)
    {
      printk("sun_ioctl: NULL instance\n");
      return (ENXIO);
    }

  minor = getminor(dev);

  open_file = getOpenFile(instance_state -> linux_major,minor);
  //what type was the file opened on?
  switch (open_file -> sun_opentype)
    {
    case OTYP_CHR:
      return (pits_char_ioctl(dev,cmd,arg,mode,credp,rvalp));
    case OTYP_BLK:
    case OTYP_LYR:
      return (pits_block_ioctl(dev,cmd,arg,mode,credp,rvalp));
    default:
      return (ENXIO);
    }
} 
