#include <misc/sunlinuxfs.h>
#include <misc/pits_dev_inst.h>
#include <linux/slab.h>

static linux_file *file_list = NULL;
static linux_inode *inode_list = NULL;

/****************************************************/
/*createOpenFile() : creates an linux file structure*/
/*given the sun open flag and a pointer to the      */
/*instance_state                                    */
/****************************************************/
linux_file *createOpenFileS(linux_fops *fops,
			    linux_inode *d_inode,
			    int sun_flags,int otype)
{
  return (createOpenFile(fops,
			 d_inode,
			 sun_flags,
			 otype));
}
linux_file *createOpenFile(linux_fops *fops,
			   linux_inode *d_inode,
			   int sun_flags,int otype)
{
  linux_file *open_file = NULL;

  open_file  = kmalloc(sizeof(linux_file),0); 
  memset(open_file,0,sizeof(linux_file)); 
 
  open_file -> private_data = NULL; 
  open_file -> f_op = fops; 
  open_file -> sun_opentype = otype;


  printk("allocate f_dentry");
  open_file -> f_dentry = kmalloc(sizeof(struct dentry),0);
  
  if (open_file -> f_dentry == NULL)
    {
      printk("f_dentry not allocated!");
    }
  
  open_file -> f_dentry -> d_inode = d_inode;

  /* set file pointer to zero. */ 
  open_file -> f_pos = 0; 
 
  /*convert to linux f_flags and f_mode */
  DEBUGDMESG("creating open_file -> f_flags"); 

  setLinuxFileFlags(open_file,sun_flags);
 
  DEBUGDMESG("okay return open_file");
  return(open_file);

  
}

int checkLinuxFile(linux_file *file)
{
  if (file == NULL) {return FALSE;}
  if (file -> f_dentry == NULL) {return FALSE;}
  
  return TRUE;
}

/***************************************/
/*destroyOpenInode() : destroys the    */
/*linux inode.                         */
/***************************************/
void destroyOpenInode(linux_inode *open_inode)
{
  kfree(open_inode);
}

void destroyOpenFile(linux_file *open_file)
{
  kfree(open_file -> f_dentry);
  kfree(open_file);
}
 
/**************************************************
 *setLinuxFileFlags() Sets the linux file flags   *
 *in accordance between the sun's file mode flags *
 *and linux's mode and file flags                 *
 **************************************************/
void setLinuxFileFlags(linux_file *filep,int sunflags)
{
  if (sunflags & FWRITE) 
    { 
      if (sunflags & FREAD) 
        { 
          filep -> f_flags = O_RDWR; 
	  filep -> f_mode = FMODE_READ | FMODE_WRITE;
	} 
      else 
        { 
          filep -> f_flags = O_WRONLY;
	  filep -> f_mode = FMODE_WRITE;
        } 
    } 
  else 
    { 
      filep -> f_flags = O_RDONLY; 
      filep -> f_mode = FMODE_READ;
    } 
  
  if (sunflags & sun_O_NDELAY || sunflags & sun_O_NONBLOCK)
    {
      filep -> f_flags = filep -> f_flags | O_NONBLOCK;
    }
  
}

/***********************************************/
/*make_i_rdev(major,minor) : creates a i_rdev  */
/*field for the linux inode structure          */
/***********************************************/
/*
int make_i_rdev(int major,int minor)
{
  int i_rdev;

  i_rdev = major;
  i_rdev = i_rdev << MINORBITS;
  i_rdev += minor;

  DEBUGDMESG2("i_rdev is:%d",i_rdev);

  return(i_rdev);

} 

int get_minor_i_rdev(int i_rdev)
{
  
  printk("minor is %d",MINOR(i_rdev));
  return (MINOR(i_rdev));
}

int get_major_i_rdev(int i_rdev)
{
  int minor = get_minor_i_rdev(i_rdev);
  int major = (i_rdev - minor) >> MINORBITS;
  return major;
}
*/


/******************************************************/
/*createOpenInode() : creates an linux inode structure*/
/*based upon the minor number and major number        */
/******************************************************/
linux_inode *createOpenInodeS(major_t smajor,minor_t sminor)
{
  dev_t sdev = makedevice(smajor,sminor);
  return(createOpenInode(getlmajor(sdev),getlminor(sdev)));
}

linux_inode *createOpenInode(major_t lmajor,minor_t lminor)
{
  linux_inode *open_inode = NULL;
  open_inode = kmalloc(sizeof(linux_inode),0);
  if (open_inode == NULL)
    {
      printk("error allocating open_inode!\n");
      return NULL;
    }
  memset(open_inode,0,sizeof(linux_inode)); 
  
  DEBUGDMESG3("major is: %d minornum is:%d",lmajor,lminor);  

  //open_inode -> i_rdev = make_i_rdev(lmajor,lminor); 

  open_inode -> i_rdev = makedevice(lmajor,lminor);

  return (open_inode);
} 
/******************************************/
/*addOpenInode() : adds a linux_inode to  */
/*the static inode_list                   */
/******************************************/
int addOpenInode(linux_inode *inode)
{

  int major = getmajor(inode -> i_rdev);
  int minor = getminor(inode -> i_rdev);

  //  int major = get_major_i_rdev(inode -> i_rdev);
  //int minor = get_minor_i_rdev(inode -> i_rdev);


  if (getOpenInode(major,minor) == NULL)
    {
      if (inode_list == NULL)
	{
	  //make inode the root.
	  inode_list = inode;
	  return DDI_SUCCESS;
	}
      else
	{
	  //add to the head.
	  inode -> next_inode = inode_list;
	  inode_list = inode;
	  return DDI_SUCCESS;
	}
    }
  
  //inode already exists.
  return DDI_FAILURE;
}
int addOpenFile(linux_file *file)
{
  dev_t i_rdev = file -> f_dentry -> d_inode -> i_rdev;
  
  int major = getmajor(i_rdev);
  int minor = getminor(i_rdev);

  //int major = get_major_i_rdev(i_rdev);
  //int minor = get_minor_i_rdev(i_rdev);

  if (getOpenFile(major,minor) == NULL)
    {
      if (file_list == NULL)
	{
	  file_list = file;
	}
      else
	{
	  file -> next_file = file_list;
	  file_list = file;
	}
      return DDI_SUCCESS;
    }

  return DDI_FAILURE;
}
/*******************************************************/
/*getOpenInode() : returns the Open Inode if it exists */
/*otherwise, returns NULL.                             */
/*******************************************************/
linux_inode *getOpenInodeS(major_t smajor,minor_t sminor)
{
  dev_t sdev = makedevice(smajor,sminor);
  return getOpenInode(getlmajor(sdev),getlminor(sdev));
}
linux_inode *getOpenInode(major_t lmajor,minor_t lminor)
{
  linux_inode *ptr = inode_list;

  dev_t search = makedevice(lmajor,lminor);

  while (ptr != NULL && ptr -> i_rdev != search)
    {
      ptr = ptr -> next_inode;
    } 
  return ptr;
}

/**********************************************/
/*getOpenFile() : returns the open file if it */
/*exists, otherwise null                      */
/**********************************************/
linux_file *getOpenFileS(major_t smajor,minor_t sminor)
{
  dev_t sdev = makedevice(smajor,sminor);
  return (getOpenFile(getlmajor(sdev),getlminor(sdev)));
}
linux_file *getOpenFile(major_t lmajor,minor_t lminor)
{
  linux_file *ptr = file_list;

  //int search = make_i_rdev(majornum,minornum);
  
  int search = makedevice(lmajor,lminor);

  while (ptr != NULL && 
	 ptr -> f_dentry -> d_inode -> i_rdev != search)
    {
      ptr = ptr -> next_file;
    }

  return ptr;
}

void removeOpenInode(linux_inode *inode)
{

  linux_inode *curr = inode_list;
  linux_inode *prev = NULL;

  while (curr != NULL && curr -> i_rdev != inode -> i_rdev)
    {
      prev = curr;
      curr = curr -> next_inode;
    }

  if (curr != NULL)
    {
      if (prev == NULL)
	{
	  inode_list = inode_list -> next_inode;
	}
      else
	{
	  prev -> next_inode = curr -> next_inode;
	}
    }

  
}

void removeOpenFile(linux_file *file)
{
  linux_file *curr = file_list;
  linux_file *prev = NULL;

  while (curr != NULL && 
	 curr -> f_dentry -> d_inode -> i_rdev != 
	 file -> f_dentry -> d_inode -> i_rdev)
    {
      prev = curr;
      curr = curr -> next_file;
    }

  if (curr != NULL)
    {
      if (prev == NULL)
	{
	  file_list = file_list -> next_file;
	}
      else
	{
	  prev -> next_file = curr -> next_file;
	}
    }

}
