/**********************************************/
/*PITS project                                */
/*pits_kernel/pits_bh.c                       */
/*bottom half handler for pits drivers.       */
/**********************************************/
#include <misc/pits_init.h>
/*taken from /linux/softirq.c*/

#include <linux/config.h>
#include <linux/mm.h>
//#include <linux/kernel_stat.h>
#include <linux/interrupt.h>
#include <linux/smp_lock.h>
#include <linux/init.h>
#include <linux/tqueue.h>



void  init_bh_tq_immediate() ;
/* Old style BHs */

static void (*bh_base[32])(void);
static struct tasklet_struct bh_task_vec[32];

/* BHs are serialized by spinlock global_bh_lock.

   It is still possible to make synchronize_bh() as
   spin_unlock_wait(&global_bh_lock). This operation is not used
   by kernel now, so that this lock is not made private only
   due to wait_on_irq().

   It can be removed only after auditing all the BHs.
 */
spinlock_t global_bh_lock = SPIN_LOCK_UNLOCKED;
spinlock_t tqueue_lock = SPIN_LOCK_UNLOCKED;

static void bh_action(unsigned long nr)
{
  
  int cpu = smp_processor_id();

  if (!spin_trylock(&global_bh_lock))
    goto resched;
  
  if (!hardirq_trylock(cpu))
    goto resched_unlock;
  
  if (bh_base[nr])
    {
      bh_base[nr]();
    }
  else
    {
      printk("warning, NULL bh %d\n",(int)nr);
    }
  hardirq_endlock(cpu);
  spin_unlock(&global_bh_lock);
  return;

 resched_unlock:
  printk("bh() : unlock\n");
  spin_unlock(&global_bh_lock);
 resched:
  printk("bh() : rescheduling\n");
  mark_bh(nr);
  
}

void init_bh(int nr, void (*routine)(void))
{
  bh_base[nr] = routine;
  mb();
}

void remove_bh(int nr)
{
  //	tasklet_kill(bh_task_vec+nr);
	bh_base[nr] = NULL;
}


/**************************************************/
//void __init softirq_init()
/*init_bh_vec() : initializes the bh_vector       */
/*just calls tasklet_init a bunch of times..      */
/**************************************************/
int init_bh_vec(dev_info_t *dip)
{
	int i;

	for (i=0; i<32; i++)
	  {
	    tasklet_init(bh_task_vec+i, bh_action, i);

	    init_bh(i,NULL);
	  }

	init_bh_tq_immediate();
	/*
	open_softirq(TASKLET_SOFTIRQ, tasklet_action, NULL);
	open_softirq(HI_SOFTIRQ, tasklet_hi_action, NULL);
	*/

	set_initialized(dip,BH_INIT);

	return 0;
}

/***********************************/
/*mark_bh() : readies the bh for   */
/*execution.                       */
/***********************************/
void mark_bh(int nr)
{
  if (bh_base[nr] == NULL)
    {
      printk("mark_bh: warning null bh! %d\n",nr);
      return;
    }
  tasklet_hi_schedule(bh_task_vec+nr);
}


/***************************************/
/*fini_bh_vec() :                      */
/*uninitializes the bh_vectors.        */
/***************************************/
int fini_bh_vec(dev_info_t *dip)
{
  unset_initialized(dip,BH_INIT);
  return DDI_SUCCESS;
}
/******************************************/
/*bh_immediate() : runs everything in the */
/*task_queue tq_immediate                 */
/******************************************/
void bh_immediate(void)
{
  printk("Inside bh_immediate\n");
  run_task_queue(&tq_immediate);
  printk("Done bh_immediate\n");
}

/*******************************************/
/*init_bh_tq_immediate() : initializes the */
/*bh value with the function               */
/*******************************************/
void init_bh_tq_immediate()
{
  //printk("init_bh_tq_immediate %d\n",IMMEDIATE_BH);
  init_bh(IMMEDIATE_BH,bh_immediate);
  //printk("done init_bh_tq_immediate\n");
}

/**********************************************/
/*__run_task_queue() : runs the tasks within  */
/*the task queue.  Uses a global tqueue_lock  */
/**********************************************/
void __run_task_queue(task_queue *list)
{
	struct list_head head, *next;
	unsigned long flags;


	spin_lock_irqsave(&tqueue_lock, flags);
	list_add(&head, list);
	list_del_init(list);
	spin_unlock_irqrestore(&tqueue_lock, flags);

	next = head.next;
	while (next != &head) {
		void (*f) (void *);
		struct tq_struct *p;
		void *data;

		p = list_entry(next, struct tq_struct, list);
		next = next->next;
		f = p->routine;
		data = p->data;
		wmb();
		p->sync = 0;
		if (f)
			f(data);
	}
}
