#include <misc/globals.h>
#include <misc/kernel.h>

#include <stdarg.h>  /*for va_list */  
#include <linux/kernel.h>

#define NBUFFER 1000
static pits_printk_t pbuffer = {initialized : FALSE};

int init_pits_printk(dev_info_t *dip)
{
  //initialize structure.
  bzero(&(pbuffer),sizeof(pits_printk_t));
  mutex_init(&(pbuffer.mutex),NULL,MUTEX_DRIVER,NULL);
  pbuffer.pbuf = kmem_alloc(PBUFMAX+1,KM_NOSLEEP);
  if (pbuffer.pbuf == NULL)
    {
      cmn_err(CE_NOTE,"init_pits_printk() : error allocating buffer!\n");
      mutex_destroy(&(pbuffer.mutex));
      return DDI_FAILURE;
    }
  bzero(pbuffer.pbuf,PBUFMAX+1);
  pbuffer.plen = 0;
  pbuffer.initialized = TRUE;
  return (DDI_SUCCESS);
}


int fini_pits_printk(dev_info_t *dip)
{
  mutex_enter(&(pbuffer.mutex));
  pbuffer.initialized = FALSE;
  if (pbuffer.plen != 0)
    {
      cmn_err(CE_NOTE,"%s",pbuffer.pbuf);
    }
  kmem_free(&(pbuffer.pbuf),PBUFMAX+1);
  mutex_exit(&pbuffer.mutex);
  mutex_destroy(&(pbuffer.mutex));
  return DDI_SUCCESS;
}





int vprintk(const char *fmt,va_list ap)
{
  int len;
  int i;
  char buf[NBUFFER];

  len = vsnprintf(buf,NBUFFER,fmt,ap);

  if (len >= 500)
    {
      if (pbuffer.plen > 0)
	{
	  cmn_err(CE_NOTE,"%s",pbuffer.pbuf);
	  pbuffer.plen = 0;
	  pbuffer.pbuf[0] = '\0';
	}
      vcmn_err(CE_NOTE,fmt,ap);
    }
  else
    {
      i = 0;
      while (i < len)
	{
	  if (pbuffer.plen >= PBUFMAX)
	    {
	      cmn_err(CE_NOTE,"%s",pbuffer.pbuf);
	      pbuffer.plen = 0;
	      pbuffer.pbuf[0] = '\0';
	    }
	  if (buf[i] == '\n')
	    {
	      if (pbuffer.plen > 0)
		{
		  cmn_err(CE_NOTE,"%s",pbuffer.pbuf);
		  pbuffer.plen = 0;
		  pbuffer.pbuf[0] = '\0';
		}
	    }
	  else
	    {
	      pbuffer.pbuf[pbuffer.plen] = buf[i];
	      pbuffer.plen++;
	      pbuffer.pbuf[pbuffer.plen] = '\0';
	    }
	  i++;
	} /*while (i < len) */
    }
  return 0;

}

int printk(const char *fmt,...)
{
  /*
  int ret;
  va_list ap;

  if (pbuffer.initialized == FALSE)
    {
      va_start(ap,fmt);
      vcmn_err(CE_NOTE,fmt,ap);
      va_end (ap);
      return 0;
    }

  mutex_enter(&(pbuffer.mutex));

  va_start(ap,fmt);
  
  ret = vprintk(fmt,ap);

  va_end(ap);

  mutex_exit(&(pbuffer.mutex));
  */
  va_list args;
  va_start(args,fmt);
  vprintf(fmt,args);
  va_end(args);

  return(0);
}
void linux_panic(const char *fmt,...)
{
  va_list ap;


  printk("Linux panic:");
  va_start(ap,fmt);
  vprintf(fmt,ap);
  va_end(ap);
  printk("\n");

  getGlobal() -> panic = TRUE;
}


int pflush()
{
  mutex_enter(&(pbuffer.mutex));
  if (pbuffer.plen > 0)
    {
      cmn_err(CE_NOTE,"%s",pbuffer.pbuf);
      pbuffer.plen = 0;
      pbuffer.pbuf[0] = '\0';
    }
  mutex_exit(&(pbuffer.mutex));
  return 0;
}

