/*************************************/
/*Misc Includes                      */
/*************************************/
#include <misc/tasklets.h>
#include <misc/pits_init.h>

#include <asm/bitops.h>
#include <linux/slab.h>


static struct tasklet_head tasklet_vec[1];
static struct tasklet_head tasklet_hi_vec[1];
//struct tasklet_struct bh_task_vec[255];

struct tasklet_struct *dequeue_tasklet(struct tasklet_head *);

void enqueue_tasklet(struct tasklet_head *,struct tasklet_struct *);

/****************************************/
/*tasklet_softintr(caddr_t arg)         */
/****************************************/
static uint_t tasklet_softintr(caddr_t arg)
{
  
  /*get the tasklet_head to run*/
  struct tasklet_head *task_list = NULL;
  struct tasklet_struct *task = NULL;

  if (arg == 0) 
    {
      printk("tasklet_softintr() : null argument\n");
      return (DDI_INTR_UNCLAIMED);
    }
  task_list = (struct tasklet_head*)arg;
  
  mutex_enter(&(task_list -> mutex));
  printk("soft intr count: %d init %d queue empty :%s running %d\n",
	 task_list -> softint_count,
	 task_list -> initialized,
	 (task_list -> list == NULL) ? "yes" : "no",
	 task_list -> running);
	 
  if (task_list -> initialized == FALSE)
    {
      printk("task list is uninitialized\n");
      task_list -> fini_detected = TRUE;
      cv_broadcast(&(task_list -> fini_wait));
      mutex_exit(&(task_list -> mutex));
      return DDI_INTR_CLAIMED;
    }

  if (task_list -> softint_count == 0)
    {
      printk("softint_count is 0\n");
      mutex_exit(&(task_list -> mutex));
      return DDI_INTR_UNCLAIMED;
    }
  if (task_list -> softint_count > 1)
    {
      printk("soft interrupt already called..\n");
      task_list -> softint_count--;
      mutex_exit(&(task_list -> mutex));
      return DDI_INTR_CLAIMED;
    }
  if (task_list -> running == TRUE)
    {
      printk("soft interrupt already called...\n");
      mutex_exit(&(task_list -> mutex));
      return (DDI_INTR_UNCLAIMED);
    }
  if (task_list -> list == NULL)
    {
      //nothing on the queue.
      printk("nothing on the queue\n");
      task_list -> softint_count = 0;
      mutex_exit(&(task_list -> mutex));
      return (DDI_INTR_CLAIMED);
    }
        
  printk("there is something in the task list\n");
  task_list -> running = TRUE;
  while (task_list -> list != NULL)
    {
      task = dequeue_tasklet(task_list);
      mutex_exit(&(task_list -> mutex));
	  
      printk("executing task\n");

      //execute task here.
      if (task != NULL && task -> func != NULL)
	{
	  task -> func(task -> data);
	}
      printk("clearing bit\n");
      clear_bit(TASKLET_STATE_SCHED, &(task -> state));
      mutex_enter(&(task_list -> mutex));
	  
    }

  printk("now the task list is empty...\n");
  task_list -> softint_count = 0;
  task_list -> running = FALSE;
  mutex_exit(&(task_list -> mutex));

  //printk("softinterrupt done\n");

  return(DDI_INTR_CLAIMED);
}

void __tasklet_schedule(struct tasklet_struct *t)
{
  //printk("inside tasklet schedule\n");
  tasklet_reschedule(&tasklet_vec[0],t);
  //printk("done tasklet schedule\n");
}
void __tasklet_hi_schedule(struct tasklet_struct *t)
{
  //tasklet_reschedule(&tasklet_hi_vec[0],t);
  __tasklet_schedule(t);
}


struct tasklet_struct* dequeue_tasklet(struct tasklet_head *task_list)
{
  struct tasklet_struct *t;
  if (task_list -> list == NULL) {return NULL;}

  t = task_list -> list;

  task_list -> list = t -> next;

  t -> next = NULL;

  return(t);
  
}

void enqueue_tasklet(struct tasklet_head *task_list,struct tasklet_struct *t)
{
  struct tasklet_struct *ptr;
  t -> next = NULL;
  if (task_list -> list == NULL)
    {
      //printk("task_list is empty\n");
      task_list -> list = t;
    }
  else
    {
      ptr = task_list -> list;
      
      while (ptr -> next != NULL)
	{
	  ptr = ptr -> next;
	}
      //printk("found last task\n");
      ptr -> next = t;
    }
  //printk("done adding task\n");
}

void tasklet_reschedule(struct tasklet_head *task_list,struct tasklet_struct *t)
{
  int need_int;
  //printk("tasklet_reschedule\n");
  if (task_list -> initialized == FALSE)
    {
      printk("error tasklet-list not initialized!\n");
      return;
    }
  
  mutex_enter(&(task_list -> mutex));
  enqueue_tasklet(task_list,t);
  
  if (task_list -> running == TRUE)
    {
      need_int = FALSE;
    }
  else
    {
      need_int = TRUE;
    }
  if (need_int == TRUE)
    {
      task_list -> softint_count++;
      //printk("triggering interrupt\n");
      ddi_trigger_softintr(task_list -> id);
    }
  mutex_exit(&(task_list -> mutex));
  
}


int fini_tasklet_vectors(dev_info_t *dip)
{
  //  int status;
  clock_t wait_time;
  if (get_initialized(dip) & TASKLET_INIT)
    {

      //  printk("fini_tasklet_vectors\n");
      // printk("removing interrupt\n");
      ddi_remove_softintr(tasklet_vec[0].id);
      //ddi_remove_softintr(tasklet_hi_vec[0].id);
      
      mutex_enter(&(tasklet_vec[0].mutex));
      tasklet_vec[0].initialized = FALSE;
      tasklet_vec[0].fini_detected = FALSE;
      wait_time = ddi_get_lbolt() + drv_usectohz(5000000);
      cv_timedwait(&(tasklet_vec[0].fini_wait),
		   &(tasklet_vec[0].mutex),
		   wait_time);
      mutex_exit(&(tasklet_vec[0].mutex));
      
      // printk("destroying mutex");
      mutex_destroy(&(tasklet_vec[0].mutex));
      cv_destroy(&(tasklet_vec[0].fini_wait));
      //printk("done fini_tasklet_vectors\n");
      set_initialized(dip,~TASKLET_INIT);
    }
  return(DDI_SUCCESS);


}

int init_tasklet_vectors(dev_info_t *dip)
{
  //int i;
  int status;
  //sun_linux_state instance_state = NULL;

  //getInstanceFReturn(dip,instance_state,sun_state);
  bzero(&tasklet_vec[0],sizeof(struct tasklet_head));

  status = ddi_get_soft_iblock_cookie(dip,
				      DDI_SOFTINT_MED,
				      &tasklet_vec[0].iblock_cookie);

  if (status != DDI_SUCCESS)
    {
      printk("failure getting iblock cookie for tasklet_vec %d",status);
      return DDI_FAILURE;
    }

  mutex_init(&(tasklet_vec[0].mutex),
	     "tasklet_vec",
	     MUTEX_DRIVER,
	     (void*)tasklet_vec[0].iblock_cookie);
  cv_init(&(tasklet_vec[0].fini_wait),NULL,CV_DRIVER,NULL);
  tasklet_vec[0].initialized = TRUE;

  mutex_enter(&(tasklet_vec[0].mutex));

  status |= ddi_add_softintr(dip,
			     DDI_SOFTINT_MED,
			     &tasklet_vec[0].id,
			     NULL,
			     NULL,
			     tasklet_softintr,
			     (caddr_t)&tasklet_vec[0]);
  if (status != DDI_SUCCESS)
    {
      printk("Failure add softintr for tasklet_vec %d",status);
    }
  mutex_exit(&(tasklet_vec[0].mutex));
  if (status == DDI_SUCCESS)
    {
      printk("success tasklet_init\n");
      return(set_initialized(dip,TASKLET_INIT));
    }
  return (status);
}


void tasklet_init(struct tasklet_struct *t,
		  void (*func)(unsigned long), unsigned long data)
{
	t->next = NULL;
	t->state = 0;
	atomic_set(&t->count, 0);
	t->func = func;
	t->data = data;
}



