/***********************/
/*pits_irq.c           */
/***********************/
#include <misc/pits_irq.h>

#include <linux/interrupt.h>

static pits_irq_vec_t pits_irq_vec;

int init_pits_irq()
{
  int i;
  bzero(&(pits_irq_vec),sizeof(pits_irq_vec_t));

  mutex_init(&(pits_irq_vec.mutex),
	     "PITS IRQ VEC",
	     MUTEX_DRIVER,
	     NULL);

  pits_irq_vec.initialized = FALSE;
  for (i=0;i<PITS_NIRQ;i++)
    {
      bzero(&(pits_irq_vec.irqs[i]),sizeof(pits_irq_t));
      pits_irq_vec.irqs[i].initialized = FALSE;
      pits_irq_vec.irqs[i].inumber = i;
      tasklet_init(&(pits_irq_vec.irqs[i].irq_tasklet),
		   pits_irq_tasklet_func,
		   i);
    }
  pits_irq_vec.initialized = TRUE;
  return DDI_SUCCESS;
}
int fini_pits_irq()
{
  mutex_destroy(&(pits_irq_vec.mutex));
  pits_irq_vec.initialized = FALSE;
  return DDI_SUCCESS;
}

static void pits_isr_debug(void *arg)
{
  
  uint64_t total_intr;
  uint64_t claimed_intr;
  uint64_t unclaimed_intr;
  uint64_t claimed_sintr;
  uint64_t unclaimed_sintr;
  pits_irq_t *pits_irq = (pits_irq_t*)arg;

  if (pits_irq == NULL)
    {
      return;
    }

  mutex_enter(&(pits_irq -> smutex));
  mutex_enter(&(pits_irq -> hmutex));

  if (pits_irq -> initialized == TRUE)
    {

      total_intr = pits_irq -> total_intr;
      claimed_intr = pits_irq -> claimed_intr;
      unclaimed_intr = pits_irq -> unclaimed_intr;
      claimed_sintr = pits_irq -> claimed_sintr;
      unclaimed_sintr = pits_irq -> unclaimed_sintr;
      mutex_exit(&(pits_irq -> hmutex));
      mutex_exit(&(pits_irq -> smutex));
      printk("pits_irq debug:\n");
      printk("total interrupts:%lu\n",total_intr);
      printk("claimed interrupts:%lu\n",claimed_intr);
      printk("unclaimed interrupts:%lu\n",unclaimed_intr);
      printk("claimed soft intr:%lu\n",claimed_sintr);
      printk("unclaimed soft intr:%lu\n",unclaimed_sintr);
      mutex_enter(&(pits_irq -> smutex));
      mutex_enter(&(pits_irq -> hmutex));
      DEBUGCMD(pits_irq -> timer_id = timeout(pits_isr_debug,pits_irq,drv_usectohz(30*1000000)));
    }

  mutex_exit(&(pits_irq -> hmutex));
  mutex_exit(&(pits_irq -> smutex));




    
}

int finiPitsIrqStruct(unsigned int irq)
{
  pits_irq_t *pits_irq = NULL;
  //int ret;

  pits_irq = getPitsIrqStruct(irq);

  if (pits_irq == NULL)
    {
      printk("finiPitsIrqStruct() : null pits_irq %d\n",irq);
      return DDI_FAILURE;
    }

  if (pits_irq -> initialized == FALSE)
    {
      printk("finiPitsIrqStruct() : uninitialized irq %d\n",irq);
    }
  mutex_enter(&(pits_irq -> smutex));
  mutex_enter(&(pits_irq -> hmutex));
  ddi_remove_intr(pits_irq -> dip,0,pits_irq -> hblock_cookie);
  pits_irq -> initialized = FALSE;

  DEBUGCMD(untimeout(pits_irq -> timer_id));

  mutex_exit(&(pits_irq -> hmutex));
  mutex_exit(&(pits_irq -> smutex));

  mutex_destroy(&(pits_irq -> hmutex));
  mutex_destroy(&(pits_irq -> smutex));


  return DDI_SUCCESS;
}
int initPitsIrqStruct(unsigned int irq,
		      dev_info_t *dip,
		      linux_irq_t *linux_irq)
{
  pits_irq_t *pits_irq = NULL;
  int ret;

  pits_irq = getPitsIrqStruct(irq);

  if (pits_irq == NULL)
    {
      printk("initPitsIrqStruct() : null pits_irq %d\n",irq);
      return DDI_FAILURE;
    }
  

  pits_irq -> irq_list = linux_irq;
  /*get the iblock cookie for this interrupt*/
  ret = ddi_get_iblock_cookie(dip,
			      0,
			      &(pits_irq -> hblock_cookie));
  if (ret != DDI_SUCCESS)
    {
      printk("addLinuxIrq() : error getting hblock_cookie: %d\n",ret);
      goto error;
    }
  ret = ddi_get_soft_iblock_cookie(dip,DDI_SOFTINT_HIGH,
				   &(pits_irq -> sblock_cookie));
  if (ret != DDI_SUCCESS)
    {
      printk("addLinuxIrq() : error getting sblock_cookie: %d\n",ret);
      goto error;
    }

  /*initialize mutexes.*/
  mutex_init(&(pits_irq -> smutex),
	     NULL,
	     MUTEX_DRIVER,
	     (void*)pits_irq -> sblock_cookie);
  mutex_init(&(pits_irq -> hmutex),
	     NULL,
	     MUTEX_DRIVER,
	     (void*)pits_irq -> hblock_cookie);
  mutex_enter(&(pits_irq -> smutex));
  mutex_enter(&(pits_irq -> hmutex));

  ret = ddi_add_softintr(dip,
			  DDI_SOFTINT_HIGH,
			  &(pits_irq -> soft_id),
			  &(pits_irq -> sblock_cookie),
			  NULL,
			  pits_irq_soft_intr,
			  (caddr_t)pits_irq);
			  
  if (ret != DDI_SUCCESS)
    {
      printk("error adding soft interrupt: %d\n",ret);
      goto error_mutex;
    }



  ret = ddi_add_intr(dip,
		     0,
		     &(pits_irq -> hblock_cookie),
		     &(linux_irq -> idevice_cookie),
		     pits_isr,
		     (caddr_t)pits_irq);

  if (ret != DDI_SUCCESS)
    {
      goto error_soft;
      printk("addLinuxIrq() : cannot add interrupt handler irq - %d error - %d\n",irq,ret);
      mutex_exit(&(pits_irq -> hmutex));
      mutex_exit(&(pits_irq -> smutex));
      mutex_destroy(&(pits_irq -> hmutex));
      mutex_destroy(&(pits_irq -> smutex));
      return DDI_FAILURE;
    }
  pits_irq -> initialized = TRUE;
  pits_irq -> enable_count++;
  pits_irq -> dip = dip;

  pits_irq -> claimed_intr = 0;
  pits_irq -> unclaimed_intr = 0;
  pits_irq -> claimed_sintr = 0;
  pits_irq -> unclaimed_sintr = 0;


  DEBUGCMD(pits_irq -> timer_id = timeout(pits_isr_debug,pits_irq,drv_usectohz(30*1000000)));

  mutex_exit(&(pits_irq -> hmutex));
  mutex_exit(&(pits_irq -> smutex));
      
  return DDI_SUCCESS;

 error_soft:
  ddi_remove_softintr(pits_irq -> soft_id);
 error_mutex:
  mutex_exit(&(pits_irq -> hmutex));
  mutex_exit(&(pits_irq -> smutex));
  mutex_destroy(&(pits_irq -> hmutex));
  mutex_destroy(&(pits_irq -> smutex));
 error:
  return DDI_FAILURE;

}
/*******************************************************/
/*pits_irq_soft_intr():                                */
/*******************************************************/
uint_t pits_irq_soft_intr(caddr_t arg)
{
  pits_irq_t *pits_irq = (pits_irq_t*)arg;
  linux_irq_t *linux_irq = NULL;
  if (pits_irq == NULL || pits_irq -> initialized == FALSE)
    {
      printk("pits_irq_soft_intr(): not initialized!");
      return (DDI_INTR_UNCLAIMED);
    }
  mutex_enter(&(pits_irq -> smutex));
  mutex_enter(&(pits_irq -> hmutex));

  if (pits_irq -> enable_count <= 0)
    {
      goto unclaimed;
    }
  if (pits_irq -> running == TRUE || pits_irq -> scheduled == FALSE)
    {
      goto unclaimed;
    }

  pits_irq -> running = TRUE;
  pits_irq -> scheduled = FALSE;
  mutex_exit(&(pits_irq -> hmutex));
  mutex_exit(&(pits_irq -> smutex));

  linux_irq = pits_irq -> irq_list;
  while (linux_irq != NULL)
    {
      //printk("calling handler...\n");
      linux_irq -> handler(pits_irq -> inumber,
			   linux_irq -> dev_id,
			   &(linux_irq -> ipt_regs));
      //printk("done calling handler\n");
      linux_irq = linux_irq -> next;
    }
  mutex_enter(&(pits_irq -> smutex));
  mutex_enter(&(pits_irq -> hmutex));
  pits_irq -> running = FALSE;

  DEBUGCMD(pits_irq -> claimed_sintr++);
  mutex_exit(&(pits_irq -> hmutex));
  mutex_exit(&(pits_irq -> smutex));
  return DDI_INTR_CLAIMED;

 unclaimed:
  DEBUGCMD(pits_irq -> unclaimed_sintr++);
  mutex_exit(&(pits_irq -> hmutex));
  mutex_exit(&(pits_irq -> smutex));
  return DDI_INTR_UNCLAIMED;
}

/*******************************************************/
/*pits_irq_tasklet_func():                             */
/*******************************************************/
void pits_irq_tasklet_func(unsigned long inumber)
{
  pits_irq_t *pits_irq = NULL;
  linux_irq_t *linux_irq = NULL;

  mutex_enter(&(pits_irq_vec.mutex));
  pits_irq = getPitsIrqStruct(inumber);
  if (pits_irq == NULL || pits_irq -> initialized == FALSE)
    {
      printk("pits_irq_tasklet_func() : bad or uninitialized inumber\n");
      mutex_exit(&(pits_irq_vec.mutex));
      return;
    }
  mutex_exit(&(pits_irq_vec.mutex));

  mutex_enter(&(pits_irq -> hmutex));
  pits_irq -> running = TRUE;
  pits_irq -> scheduled = FALSE;
  if (pits_irq -> run_count <= 0)
    {
      pits_irq -> run_count = 0;
      pits_irq -> running = FALSE;
      cv_broadcast(&(pits_irq -> tasklet_wait));
      mutex_exit(&(pits_irq -> hmutex));
      return;
    }
  mutex_exit(&(pits_irq -> hmutex));
  linux_irq = pits_irq -> irq_list;
  while (linux_irq != NULL)
    {
      if (linux_irq -> flags & SA_INTERRUPT) 
      {mutex_enter(&(pits_irq -> hmutex));}
      linux_irq -> handler(pits_irq -> inumber,
			   linux_irq -> dev_id,
			   &(linux_irq -> ipt_regs));
      if (linux_irq -> flags & SA_INTERRUPT) 
      {mutex_exit(&(pits_irq -> hmutex));}
      
      linux_irq = linux_irq -> next;
    }
  mutex_enter(&(pits_irq -> hmutex));
  pits_irq -> run_count--;
  if (pits_irq -> run_count > 0)
    {
      //reschedule ourselves so we can handle more interrupts.
      printk("rescheduling irq...\n");
      tasklet_schedule(&(pits_irq -> irq_tasklet));
    }
  else
    {
      cv_broadcast(&(pits_irq -> tasklet_wait));
      pits_irq -> running = FALSE;
    }
  mutex_exit(&(pits_irq -> hmutex));
}



/****************************************************/
/*pits_isr() :       Here is the interrupt handler  */
/*for pits drivers.  It effectly calls the driver   */
/*Interrupt handler.                                */
/****************************************************/
uint_t pits_isr(caddr_t arg)
{
  pits_irq_t *pits_irq = (pits_irq_t *)arg;
  
  if ((pits_irq == NULL) || (pits_irq -> initialized == FALSE))
    {
      printk("pits_isr() : bad or uninitialized interrupt!");
      return DDI_INTR_UNCLAIMED;
    }
  mutex_enter(&(pits_irq -> hmutex));

  DEBUGCMD(pits_irq -> total_intr++);
  if (pits_irq -> enable_count <= 0)
    {
      goto unclaimed;
    }
  if (pits_irq -> running == TRUE || pits_irq -> scheduled == TRUE)
    {
      goto unclaimed;
    }

  pits_irq -> scheduled = TRUE;

  // claimed:

  DEBUGCMD(pits_irq -> claimed_intr++);
  mutex_exit(&(pits_irq -> hmutex));
  ddi_trigger_softintr(pits_irq -> soft_id);
  return DDI_INTR_UNCLAIMED;

 unclaimed:
  DEBUGCMD(pits_irq -> unclaimed_intr++);
  mutex_exit(&(pits_irq -> hmutex));
  return DDI_INTR_UNCLAIMED;
}


pits_irq_t *getPitsIrqStruct(int inumber)
{
  if (inumber < 0 || inumber > PITS_NIRQ-1)
    {
      return NULL;
    }
  return(&pits_irq_vec.irqs[inumber]);
}
linux_irq_t *getLinuxIrqStruct(int inumber)
{
  pits_irq_t *pits_irq = getPitsIrqStruct(inumber);

  if (pits_irq == NULL || pits_irq -> initialized == FALSE)
    {
      return NULL;
    }
  return pits_irq -> irq_list;
}


int addLinuxIrq(dev_info_t *dip,
		unsigned int irq, 
		void (*handler)(int, void *, struct pt_regs *),
		unsigned long irqflags, 
		const char * devname,
		void *dev_id)
{
  pits_irq_t *pits_irq;
  linux_irq_t *linux_irq;
  int ret;

  printk("addLinuxIrq() : irq - %d devname %s\n",irq,devname);
  if (irqflags & SA_INTERRUPT)
    {
      printk(" SA_INTERRUPT\n");
    }
  if (irqflags & SA_SHIRQ)
    {
      printk(" SA_SHIRQ\n");
    }


  //check and see if interrupt is allowed for this device.
  if (ddi_intr_hilevel(dip,irq) != 0)
    {
      printk("addLinuxIrq() : high-level interrupt for %s, is not supported!\n",devname,irq);
      return (DDI_FAILURE);
    }

  mutex_enter(&(pits_irq_vec.mutex));
  linux_irq = kmem_zalloc(sizeof(linux_irq_t),KM_NOSLEEP);
  
  if (linux_irq == NULL)
    {
      printk("addLinuxIrq() : error allocating linux_irq!\n");
      return DDI_FAILURE;
    }
  //fill in linux_irq structure.
  linux_irq -> dip = dip;
  linux_irq -> handler = handler;
  linux_irq -> dev_id = dev_id;
  bzero(&(linux_irq -> ipt_regs),sizeof(struct pt_regs));
  linux_irq -> flags = irqflags;


  pits_irq = getPitsIrqStruct(irq);
  if (pits_irq -> initialized == FALSE)
    {

      ret = initPitsIrqStruct(irq,dip,linux_irq);

    }
  else
    {
      printk("Already initialized.\n");
      printk("not supported yet...\n");
      kmem_free(linux_irq,sizeof(linux_irq_t));
      return DDI_FAILURE;
    }
  mutex_exit(&(pits_irq_vec.mutex));
  return ret;
}
void pits_disable_irq(unsigned int irq)
{
  pits_irq_t *pits_irq = getPitsIrqStruct(irq);

  printk("pits_disable_irq(%u)\n",irq);

  if (pits_irq != NULL && pits_irq -> initialized == TRUE)
    {
      mutex_enter(&(pits_irq -> hmutex));
      pits_irq -> enable_count--;
      while (pits_irq -> running == TRUE || pits_irq -> scheduled == TRUE)
	{
	  cv_wait(&(pits_irq -> tasklet_wait),&(pits_irq -> hmutex));
	}
      mutex_exit(&(pits_irq -> hmutex));
      
    }
  else
    {
      printk("pits_disable_irq() : unitialized irq!\n");
    }

}
void pits_disable_irq_nosync(unsigned int irq)
{
  pits_disable_irq(irq);
}
void pits_enable_irq(unsigned int irq)
{
  pits_irq_t *pits_irq = getPitsIrqStruct(irq);
  printk("pits_enable_irq()\n");
  if (pits_irq != NULL && pits_irq -> initialized == TRUE)
    {
      mutex_enter(&(pits_irq -> hmutex));
      pits_irq -> enable_count++;
      if (pits_irq -> enable_count > 1)
	{
	  pits_irq -> enable_count = 1;
	}
      mutex_exit(&(pits_irq -> hmutex));
    }
  else
    {
      printk("pits_enable_irq() : bad or uninitialized irq %d\n",irq);
    }
}

int removeLinuxIrq(unsigned int irq,void *dev_id)
{
  //int ret;
  pits_irq_t *pits_irq = getPitsIrqStruct(irq);
  linux_irq_t *curr = NULL;
  linux_irq_t *prev = NULL;

  pits_disable_irq(irq);
  

  if (pits_irq != NULL && pits_irq -> initialized == TRUE)
    {
      curr = pits_irq -> irq_list;
      prev = curr;
      //find irq.
      while (curr != NULL && curr -> dev_id != dev_id)
	{
	  prev = curr;
	  curr = curr -> next;
	}
      if (curr == NULL)
	{
	  printk("removeLinuxIrq() : linux_irq doesn't exist!\n");
	  pits_enable_irq(irq);
	  return DDI_FAILURE;
	}
      
      mutex_enter(&(pits_irq -> hmutex));
      if (prev == curr)
	{
	  pits_irq -> irq_list = curr -> next;
	}
      else
	{
	  prev -> next = curr -> next;
	}
      if (pits_irq -> irq_list == NULL)
	{
	  mutex_exit(&(pits_irq -> hmutex));
	  finiPitsIrqStruct(irq);
	  
	}
      else
	{
	  mutex_exit(&(pits_irq -> hmutex));
	  pits_enable_irq(irq);
	}
      kmem_free(curr,sizeof(linux_irq_t));
      return DDI_SUCCESS;
    }
  else
    {
      printk("removeLinuxIrq() : pits_irq not initialized!\n");
      return DDI_FAILURE;
    }
  
  
}
