#include "pits_kernel.h"
#include "pits_kernel_state.h"
#include <linux/kernel.h>
#include <linux/spinlock.h>
#include <linux/ioport.h>
//PITS includes
#include <misc/slab.h>
#include <misc/pits_blkdev.h>
#include <misc/tasklets.h>
#include <misc/pits_bh.h>


/*io_request_lock*/
spinlock_t io_request_lock;
struct resource *ioport_resource_tree;
struct resource *iomem_resource_tree;

pits_kernel_state_t *getPitsKernelState()
{
  return ((pits_kernel_state_t *)ddi_get_soft_state(statep,0));
}


/****************************/
/*_init(void) :             */
/****************************/
int _init(void)
{ 
  int error;
  ddi_soft_state_init(&statep,sizeof(struct pits_kernel_state),1); 
  /*10 is estimated number of instances*/
  printk("_init()\n");

  error = mod_install(&pits_kernel_modlinkage);
  if (error !=0)
    {
      printk("_init() - error: %d\n",error);
      //ddi_soft_state_init(&statep);
    }
  else
    {
      io_request_lock.locked = 0;
      io_request_lock.initialized = 0;
      spin_lock_init(&io_request_lock);
      
      ioport_resource_tree = NULL;
      iomem_resource_tree = NULL;
    }
  return (error);
}

/****************************/
/*_fini(void)               */
/****************************/
int _fini(void)
{
  int error;
  error = mod_remove(&pits_kernel_modlinkage);
  
  printk("_fini()\n");
  if (error == 0)
    {
      /*release per module resources if any were allocated*/
      ddi_soft_state_fini(&statep);
      pits_fini_spin_lock(&io_request_lock);
     
    }
  else
    {
      cmn_err(CE_CONT,"_fini-error:%d\n",error);
    }
  

  return (error);

}

/****************************/
/*_info                     */
/****************************/
int _info(struct modinfo *modinfop)
{

  int error = mod_info(&pits_kernel_modlinkage,modinfop);

  return(error);
}


/****************************/
/*pits_kernel_getinfo       */
/****************************/
static int pits_kernel_getinfo(dev_info_t *dip,ddi_info_cmd_t infocmd, void *arg, void **result)
{

  int error;

  struct pits_kernel_state *rs;

  int instance;
  printk("pits_kernel_getinfo()\n");
  instance = getminor((dev_t)arg);
  

  switch(infocmd)
    {
    case DDI_INFO_DEVT2DEVINFO:
      rs = ddi_get_soft_state(statep,instance);
      if (rs == NULL)
	{
	  *result = NULL;
	  error = DDI_FAILURE;
	}
      else
	{
	  *result = rs -> dip;
	  error = DDI_SUCCESS;
	}
      break;
    case DDI_INFO_DEVT2INSTANCE:
      *result = (void*)instance;
      error = DDI_SUCCESS;
    default:
      *result = NULL;
      error = DDI_FAILURE;
    }
  printk("done pits_kernel_getinfo()\n");

  return(error);
}

static int pits_kernel_probe(dev_info_t *dip)
{
  cmn_err(CE_CONT,"pits_kernel_probe");
  
  /*DONT care for right now*/
  return(DDI_PROBE_DONTCARE);
}
 
static int pits_kernel_attach(dev_info_t *dip, ddi_attach_cmd_t cmd) 
{
  int instance;
  struct pits_kernel_state *rs;
  ddi_dma_attr_t dma_attr = PITS_KERNEL_DMA_ATTR;
  ddi_device_acc_attr_t dev_acc_attr = PITS_KERNEL_DEV_ACC_ATTR;
  printk("pits_kernel_attach()\n");

  switch(cmd)
    {
    case DDI_ATTACH:
      instance = ddi_get_instance(dip);
     
      if (ddi_soft_state_zalloc(statep,instance) != DDI_SUCCESS)
	{
	  cmn_err(CE_NOTE,"couldn't allocate ");
	  return (DDI_FAILURE);
	  
	}
      if (init_tasklet_vectors(dip) != DDI_SUCCESS)
	{
	  printk("Error initializing tasklets\n");
	  return (DDI_FAILURE);
	}
      init_bh_vec(dip);
      init_pits_slab(dip);
      init_pits_blkdev();

      rs = ddi_get_soft_state(statep,instance);
      rs -> dip = dip;
      rs -> instance = instance;
      bcopy(&dma_attr,&(rs -> dma_attr),sizeof(ddi_dma_attr_t));
      bcopy(&dev_acc_attr,&(rs -> dev_acc_attr),sizeof(ddi_device_acc_attr_t));




      ddi_report_dev(dip);
      return(DDI_SUCCESS);
    case DDI_RESUME:
      return(DDI_SUCCESS);
    default:
      return(DDI_FAILURE);
    }

  printk("done pits_kernel_attach\n");
  /*return DDI_SUCESS*/
  return (DDI_SUCCESS);
}
static int pits_kernel_detach(dev_info_t *dip,ddi_detach_cmd_t cmd)
{
  int instance;
  struct pits_kernel_state *rs;

  cmn_err(CE_CONT,"pits_kernel_detach\n");

  switch(cmd)
    {
    case DDI_DETACH:
      instance = ddi_get_instance(dip);
      rs = ddi_get_soft_state(statep,instance);
      //fini_pits_slab(dip);
      fini_bh_vec(dip);
      fini_tasklet_vectors(dip);
      fini_pits_slab(dip);
      fini_pits_blkdev();

      ddi_soft_state_free(statep,instance);
      break;
    case DDI_SUSPEND:
    default:
      return(DDI_FAILURE);
    }


  return (DDI_SUCCESS);
} 
int pits_kernel_power(dev_info_t *dip,int component,int level)
{
  cmn_err(CE_CONT,"!pits_kernel_power");
  return 0;
} 


