#ifndef __PITS_KERNEL_STATE_
#define __PITS_KERNEL_STATE_

#include <sys/ddi.h>
#include <sys/sunddi.h>


#define BURST1         0x01
#define BURST2         0x02
#define BURST4         0x04
#define BURST8         0x08
#define BURST16        0x10
#define BURST32        0x20
#define BURST64        0x40
#define BURSTSIZE_MASK 0x7f
#define DEFAULT_BURSTSIZE BURST16|BURST8|BURST4|BURST2|BURST1



#define PITS_KERNEL_DMA_ATTR \
{ dma_attr_version    : DMA_ATTR_V0,                           \
  dma_attr_addr_lo    : (uint64_t)0,                           \
  dma_attr_addr_hi    : (uint64_t)0xffffffff,                  \
  dma_attr_count_max  : (uint64_t)0x00ffffff,                  \
  dma_attr_align      : (uint64_t)4,                           \
  dma_attr_burstsizes : 0x0f,                                  \
  dma_attr_minxfer    : 1,                                     \
  dma_attr_maxxfer    : (uint64_t)0x00ffffff,                  \
  dma_attr_seg        : 0x00ffffff,                            \
  dma_attr_sgllen     : 1,                                     \
  dma_attr_granular   : 1,                                     \
  dma_attr_flags      : 0                                      \
}

#define PITS_KERNEL_DEV_ACC_ATTR                               \
{ devacc_attr_version      : DDI_DEVICE_ATTR_V0,               \
  devacc_attr_endian_flags : DDI_NEVERSWAP_ACC,                \
  devacc_attr_dataorder    : DDI_STRICTORDER_ACC               \
}

typedef struct pits_kernel_state
{
  kmutex_t        mutex;          /* lock on driver */
  kcondvar_t      condvar;        /* waiting for I/O to complete */
  int             instance;
  dev_info_t      *dip;
  ddi_device_acc_attr_t dev_acc_attr;
  ddi_dma_attr_t dma_attr;
}pits_kernel_state_t;

pits_kernel_state_t *getPitsKernelState();


#endif

