#include <misc/pits_mutex.h>
#include <strings.h>



#ifndef PITS_DEBUG

void pits_mutex_init(pits_mutex_t *mp,
		     char *name, 
		     kmutex_type_t type,
		     void *arg)
{
  mutex_init(&(mp -> mutex),name,type,arg); 
}

void pits_mutex_destroy(pits_mutex_t *mp)
{
  mutex_destroy(&(mp -> mutex));
}

void pits_mutex_enter(pits_mutex_t *mp)
{
  mutex_enter(&(mp -> mutex));
}

void pits_mutex_exit(pits_mutex_t *mp)
{
  mutex_exit(&(mp -> mutex));
}

int pits_mutex_owned(pits_mutex_t *mp)
{
  return (mutex_owned(&(mp -> mutex)));
}

int pits_mutex_tryenter(pits_mutex_t *mp)
{
  return (mutex_tryenter(&(mp -> mutex)));
}

#else

#include <sys/types.h>
#include <sys/ddi.h>
#include <sys/sunddi.h>


#include <linux/kernel.h>

void pits_mutex_init(pits_mutex_t *mp,
		     char *name, 
		     kmutex_type_t type,
		     void *arg)
{

  bzero(mp,sizeof(pits_mutex_t));
  mutex_init(&(mp -> mutex),name,type,arg);

  mp -> name = name;
  mp -> type = type;
  mp -> arg =  arg;
  mp -> locked = FALSE;
  cv_init(&(mp -> cv_wait),NULL,CV_DRIVER,NULL);
 
}

void pits_mutex_destroy(pits_mutex_t *mp)
{
  mutex_destroy(&(mp -> mutex));
  cv_destroy(&(mp -> cv_wait));
  ;
}

void pits_mutex_enter(pits_mutex_t *mp)
{
  int ret = 0;
  mutex_enter(&(mp -> mutex));

  while (mp -> locked == TRUE)
    {
      ret = cv_timedwait(&(mp -> cv_wait),
			 &(mp -> mutex),
			 ddi_get_lbolt() + drv_usectohz(30 * 1000000));
      if (ret == -1)
	{
	  printk("Mutex timed out....");
	  mutex_enter(&(mp -> mutex));
	}
      
    }
  mp -> locked = TRUE;
  mutex_exit(&(mp -> mutex));
    

}

void pits_mutex_exit(pits_mutex_t *mp)
{
  mutex_enter(&(mp -> mutex));
  mp -> locked = FALSE;
  cv_signal(&(mp -> cv_wait));
  mutex_exit(&(mp -> mutex));
  
}

int pits_mutex_owned(pits_mutex_t *mp)
{
  return 0;
}


int pits_mutex_tryenter(pits_mutex_t *mp)
{
  int ret = 0;
  mutex_enter (&(mp -> mutex));

  if (mp -> locked == TRUE)
    {
      ret = 0;
    }
  else
    {
      ret = 1;
      mp -> locked = TRUE;
    }
  mutex_exit(&(mp -> mutex));

  return ret;
}

#endif

