#include <linux/spinlock.h>
#include <sys/types.h>
#include <sys/kmem.h>
#include <sys/ddi.h>
#include <sys/sunddi.h>
#include <linux/kernel.h>

//#include <misc/globals.h>
//#define spin_list getGlobal() -> spin_lock_list
#define TRUE 1
#define FALSE 0

void add_spin_list(spinlock_t **list,spinlock_t *element)
{
  element -> next = *list;
  *list = element;
}
/*************************************/
/*pits_add_spin_lock()               */
/*This function does two things      */
/*First it initializes the mutexes   */
/*and condition variables for locking*/
/*then it adds the lock to the list  */
/*for later removal via destroy_mutex*/
/*************************************/
void pits_add_spin_lock(spinlock_t *lock)
{
  //DEBUGDMESG("add_spin_lock()");
  lock -> normal_lock = (kmutex_t*)kmem_alloc(sizeof(kmutex_t),0);
  lock -> normal_cv = (kcondvar_t*)kmem_alloc(sizeof(kcondvar_t),0);
  mutex_init(lock -> normal_lock,NULL,MUTEX_DRIVER,NULL);
  cv_init(lock -> normal_cv,NULL,CV_DRIVER,NULL);
  lock -> initialized = TRUE;
  if (lock -> locked != TRUE && lock -> locked != FALSE)
    {
      lock -> locked = FALSE;
    }
  
  
  //add_spin_list(&(getGlobal() -> spin_lock_list),lock);
  //DEBUGDMESG("done add_spin_lock()");
}
void pits_spin_lock(spinlock_t *lock,int type)
{
  #ifdef PITS_SPIN_LOCK_ON
  //printk("pits_spin_lock() %d\n",type);
  //DEBUGCMD(printk("pits_spin_lock() %d\n",type));
  if (lock -> initialized != TRUE) {pits_add_spin_lock(lock);}
  switch(type)
    {
    case NORMAL_LOCK:
      pits_spin_lock_normal(lock);
      break;
    case IRQ_LOCK:
    case IRQSAVE_LOCK:
    case BH_LOCK:
      if (lock -> initialized_iblock == FALSE &&
	  pits_spin_lock_init_iblock(lock) == FALSE)
	{
	  printk("pits_spin_lock() : uninitialized iblock!\n");
	  pits_spin_lock_normal(lock);
	}
      else
	{
	  //printk("pits_spin_lock() : uninitialized iblock!\n");
	  pits_spin_lock_irq(lock);
	}
      break;
    default:
      printk("Warning unknown lock type\n");
      pits_spin_lock_normal(lock);
    }
  #endif
}

int pits_spin_lock_init_iblock(spinlock_t *lock)
{
  return FALSE;
}

/*****************************************/
/*pits_spin_lock_normal()                */
/*****************************************/
void pits_spin_lock_normal(spinlock_t *lock)
{
  mutex_enter(lock -> normal_lock);
  while (lock -> locked == TRUE)
    {
      printk("waiting for lock....\n");
      cv_timedwait(lock -> normal_cv,
		   lock -> normal_lock,
		   (ddi_get_lbolt() + drv_usectohz(5000000)));
    }
  lock -> locked = TRUE;
  mutex_exit(lock -> normal_lock);
}
void pits_spin_lock_irq(spinlock_t *lock)
{
  mutex_enter(lock -> normal_lock);
  while (lock -> locked == TRUE && 
	 mutex_tryenter(lock -> iblock_lock) == 0)
    {
      cv_wait(lock -> normal_cv,lock -> normal_lock);
    }
  
  lock -> locked = TRUE;
  mutex_exit(lock -> iblock_lock);
  mutex_exit(lock -> normal_lock);
}
void pits_spin_unlock_irq(spinlock_t *lock)
{
  mutex_enter(lock -> normal_lock);
  lock -> locked = FALSE;
  mutex_exit(lock -> iblock_lock);
  mutex_exit(lock -> normal_lock);
}

int pits_spin_trylock(spinlock_t *lock,int type)
{
  #ifdef PITS_SPIN_LOCK_ON
  int ret;
  if (lock -> initialized != TRUE) 
    {
      pits_add_spin_lock(lock);
    }
  //DEBUGCMD(printk("pits_spin_trylock() %d\n",type));
  mutex_enter(lock -> normal_lock);
  if (lock -> locked == TRUE)
    {
      //DEBUGCMD(printk("pits_spin_trylock(): already locked...\n"));
      ret = 0;
    }
  else
    {
      //DEBUGCMD(printk("pits_spin_trylock(): got lock....\n"));
      ret = 1;
      lock -> locked = TRUE;
    }
  mutex_exit(lock -> normal_lock);
  return ret;
  #else
  return 1;
  #endif
}
void pits_spin_unlock(spinlock_t *lock,int type)
{
  #ifdef PITS_SPIN_LOCK_ON
  switch(type)
    {
    case NORMAL_LOCK:
      pits_spin_unlock_normal(lock);
      break;
    case IRQ_LOCK:
    case IRQSAVE_LOCK:
    case BH_LOCK:
      if (lock -> initialized_iblock == FALSE)
	{
	  pits_spin_unlock_normal(lock);
	}
      else
	{
	  pits_spin_unlock_irq(lock);
	}
      break;
    default:
      pits_spin_unlock_normal(lock);
      
    }
  #endif
}
void pits_spin_unlock_normal(spinlock_t *lock)
{
#ifdef PITS_SPIN_LOCK_ON
  //printk("pits_spin_unlock() %d\n",type);
  mutex_enter(lock -> normal_lock);
  lock -> locked = FALSE;
  cv_broadcast(lock -> normal_cv);
  mutex_exit(lock -> normal_lock);
  //printk("done pits_spin_unlock %d\n",type);
  //DEBUGDMESG("done pits_spin_unlock()");
#endif
}

void pits_fini_spin_lock(spinlock_t *lock)
{
  printk("pits_fini_spin_lock()\n");
  cv_destroy(lock -> normal_cv);
  kmem_free(lock -> normal_cv,sizeof(kcondvar_t));
  mutex_destroy(lock -> normal_lock);
  kmem_free(lock -> normal_lock,sizeof(kmutex_t));
  lock -> initialized = FALSE;
  printk("done pits_fini_spin_lock()\n");
}

void pits_spin_lock_destroy()
{
  //spinlock_t *current_lock = getGlobal() -> spin_lock_list;
  /*
  while (current_lock != NULL)
    {
      cv_destroy(current_lock -> normal_cv);
      mutex_destroy(current_lock -> normal_lock);
      current_lock = current_lock -> next;
    }
  */
}

void add_rw_list(rwlock_t **list,rwlock_t *element)
{
  element -> next = *list;
  *list = element;
}

static rwlock_t *rw_list = NULL;


void pits_add_rw_lock(rwlock_t *lock)
{
  //printk("inside pits_add_rw_lock()\n");
  lock -> normal_lock = (kmutex_t*)kmem_alloc(sizeof(kmutex_t),0);
  lock -> normal_cv = (kcondvar_t*)kmem_alloc(sizeof(kcondvar_t),0);
  mutex_init(lock -> normal_lock,NULL,MUTEX_DRIVER,NULL);
  cv_init(lock -> normal_cv,NULL,CV_DRIVER,NULL);
  lock -> initialized = TRUE;
  if (lock -> write_lock != TRUE && lock -> write_lock != FALSE)
    {
      lock -> write_lock = FALSE;
    }
  
  lock -> readers = 0;
  add_rw_list(&rw_list,lock);


  //printk("done pits_add_rw_lock()\n");
}
void pits_read_lock(rwlock_t *lock,int type)
{
  //printk("inside pits_read_lock()\n");
  if (lock -> initialized != TRUE) {pits_add_rw_lock(lock);}
  mutex_enter(lock -> normal_lock);
  while (lock -> write_lock == TRUE)
    {
      cv_wait(lock -> normal_cv,lock->normal_lock);
    }
  lock -> readers++;
  mutex_exit(lock -> normal_lock);
  //printk("done pits_read_lock()\n");
}
void pits_write_lock(rwlock_t *lock,int type)
{
  //printk("inside pits_write_lock()\n");
  if (lock -> initialized != TRUE) {pits_add_rw_lock(lock);}
  mutex_enter(lock -> normal_lock);
  while (lock -> write_lock == TRUE &&
	 lock -> readers > 0)
    {
      cv_wait(lock -> normal_cv,lock->normal_lock);
    }
  lock -> write_lock = TRUE;
  mutex_exit(lock -> normal_lock);

  //printk("done pits_write_lock()\n");
}
void pits_read_unlock(rwlock_t *lock,int type)
{
  //printk("inside pits_read_unlock()\n");
  mutex_enter(lock -> normal_lock);
  lock -> readers--;
  if (lock -> readers < 0) 
    {
      lock -> readers = 0;
    }
  cv_broadcast(lock -> normal_cv);
  mutex_exit(lock -> normal_lock);
  //printk("done pits_read_unlock()\n");
}

void pits_write_unlock(rwlock_t *lock,int type)
{
  //printk("inside pits_write_unlock()\n");
  mutex_enter(lock -> normal_lock);
  lock -> write_lock = FALSE;
  cv_broadcast(lock -> normal_cv);
  mutex_exit(lock -> normal_lock);
  //printk("done pits_write_unlock()\n");
}
