/*************************************/
/*Misc Includes                      */
/*************************************/
#include <misc/tasklets.h>
//#include <misc/pits_init.h>
#include <asm/bitops.h>

/*************************************/
/*Linux Includes                     */
/*************************************/
#include <linux/slab.h>


/*************************************/
/*Static Global Variables            */
/*************************************/
static struct tasklet_head tasklet_vec[1];
//static struct tasklet_head tasklet_hi_vec[1];
//struct tasklet_struct bh_task_vec[255];

/************/
/*debug code*/
/************/
static void tasklet_debug(void *arg)
{

  uint64_t soft_intr;
  uint64_t soft_claimed;
  uint64_t soft_unclaimed;
  int state;
  int empty;
  struct tasklet_head *task_list = (struct tasklet_head*)arg;

  if (task_list == NULL)
    {
      printk("tasklet_debug, error null tasklet_list!\n");
      return;
    }
  mutex_enter(&(task_list -> mutex));
  if (task_list -> initialized == TRUE)
    {
      soft_intr = task_list -> soft_intr;
      soft_claimed = task_list -> soft_claimed;
      soft_unclaimed = task_list -> soft_unclaimed;
      state = task_list -> run_state;
      empty = list_empty(&(task_list -> list));
      mutex_exit(&(task_list -> mutex));
      printk("tasklet debug:\n");
      printk("Total interrupts:%lu\n",soft_intr);
      printk("Total claimed:%lu\n",soft_claimed);
      printk("Total unclaimed:%lu\n",soft_unclaimed);
      switch(state)
	{
	case TASKLET_STATE_DISABLED:
	  printk("state:DISABLED");
	  break;
	case TASKLET_STATE_SCHED:
	  printk("state:SCHED");
	  break;
	case TASKLET_STATE_RUN:
	  printk("state:RUN");
	  break;
	default:
	  printk("state:UNKNOWN");
	}
      if (empty)
	printk("list is empty\n");
      else
	printk("list is not empty\n");
     
      task_list -> timer_id = timeout(tasklet_debug,task_list,drv_usectohz(30*1000000));
      
    }
  else
    {
      mutex_exit(&(task_list -> mutex));
    }
  
  


}



/****************************************/
/*tasklet_softintr(caddr_t arg)         */
/****************************************/
static uint_t tasklet_softintr(caddr_t arg)
{
  
  /*get the tasklet_head to run*/
  struct tasklet_head *task_list = NULL;
  struct tasklet_struct *task = NULL;
  int need_reschedule = FALSE;
  

  if (arg == 0) 
    {
      printk("tasklet_softintr() : null argument\n");
      return (DDI_INTR_UNCLAIMED);
    }
  task_list = (struct tasklet_head*)arg;

  mutex_enter(&(task_list -> mutex));

  DEBUGCMD(task_list -> soft_intr++);

  if (task_list -> initialized == FALSE)
    {
      printk("task list is uninitialized\n");
      task_list -> fini_detected = TRUE;
      cv_broadcast(&(task_list -> fini_wait));
      goto claimed;
    }

  switch (task_list -> run_state)
    {
    case TASKLET_STATE_DISABLED:
      //DEBUGCMD(printk("task_list is not scheduled\n"));
      goto unclaimed;
    case TASKLET_STATE_RUN:
      //DEBUGCMD(printk("task_list is already running\n"));
      goto unclaimed;
    default:
      ;
    }
  //set the state to running.
  task_list -> run_state = TASKLET_STATE_RUN;
  //while (!list_empty(&(task_list -> list)))
    //{
      task = dequeue_tasklet(task_list);
      task -> state = TASKLET_STATE_RUN;
      mutex_exit(&(task_list -> mutex));
	  
      //execute state here.
      if (task != NULL && task -> func != NULL)
	{
	  task -> func(task -> data);
	}
      else
	{
	  printk("tasklet_softintr() : invalid tasklet....\n");
	}
      //printk("clearing bit\n");
      
      mutex_enter(&(task_list -> mutex));
      
      task -> run_count --;
      if (task -> run_count <= 0)
	{
	  task -> state = TASKLET_STATE_DISABLED;
	  task -> run_count = 0;
	}
      else
	{
	  //printk("needs rerun...\n");
	  task -> state = TASKLET_STATE_SCHED;
	  enqueue_tasklet(task_list,task);
	}
      // }
  if (!list_empty(&(task_list -> list)))
    {
      need_reschedule = TRUE;
      task_list -> run_state = TASKLET_STATE_SCHED;
    }
  else
    {
      task_list -> run_state = TASKLET_STATE_DISABLED;
    }
 claimed:
  DEBUGCMD(task_list -> soft_claimed++);
  mutex_exit(&(task_list -> mutex));
  if (need_reschedule == TRUE)
    {
      ddi_trigger_softintr(task_list -> id);
    }
  //DEBUGCMD(printk("tasklet_softintr() : task list is now empty.\n"));
  //DEBUGCMD(printk("tasklet_softintr() : softinterrupt done.\n"));
  return(DDI_INTR_CLAIMED);
  
 unclaimed:
  DEBUGCMD(task_list -> soft_unclaimed++);
  mutex_exit(&(task_list -> mutex));
  return(DDI_INTR_UNCLAIMED);

}
/************************************************/
/*__tasklet_schedule() : Schedules a tasklet    */
/************************************************/
void __tasklet_schedule(struct tasklet_struct *t)
{
  //printk("inside tasklet schedule\n");
  tasklet_reschedule(&tasklet_vec[0],t);
  //printk("done tasklet schedule\n");
}
/************************************************/
/*__tasklet_hi_schedule() : Schedules a tasklet */
/************************************************/
void __tasklet_hi_schedule(struct tasklet_struct *t)
{
  //tasklet_reschedule(&tasklet_hi_vec[0],t);
  __tasklet_schedule(t);
}
/***********************************************/
/*dequeue_tasklet() : removes a tasklet from   */
/*the tasklet queue.                           */
/***********************************************/
struct tasklet_struct* dequeue_tasklet(struct tasklet_head *task_list)
{
  struct tasklet_struct *t;
  struct list_head *entry = NULL;

  if (list_empty(&(task_list -> list)))
  {
    return NULL;
  }
  
  entry = (task_list -> list.next);

  t = list_entry(entry,
		 struct tasklet_struct,
		 head);
  
  list_del_init(entry);
  return(t);
  
}
/******************************************************************/
/*enqueue_tasklet() : queues a tasklet to the end of the task_list*/
/******************************************************************/
void enqueue_tasklet(struct tasklet_head *task_list,
		     struct tasklet_struct *t)
{
 
  struct list_head *theadp = &(t -> head);
  
  DEBUGCMD(printk("inside enqueue_tasklet\n"));
  INIT_LIST_HEAD(theadp);

  if (list_empty(&(task_list -> list)))
    {
      list_add(theadp,&(task_list -> list));
    }
  else
    {
      list_add_tail(theadp,&(task_list -> list));
    }
  DEBUGCMD(printk("done enqueue tasklet...\n"));
}
/**********************************************************/
/*tasklet_reschedule() : reschedules a tasklet            */
/**********************************************************/
void tasklet_reschedule(struct tasklet_head *task_list,
			struct tasklet_struct *t)
{
  int need_int;
  //printk("tasklet_reschedule\n");
  if (task_list -> initialized == FALSE)
    {
      printk("error tasklet-list not initialized!\n");
      return;
    }
  
  mutex_enter(&(task_list -> mutex));

  if (t -> state == TASKLET_STATE_SCHED ||
      t -> state == TASKLET_STATE_RUN)
    {
      DEBUGCMD(printk("already scheduled or running, increment count.\n"));
      t -> run_count++;
      DEBUGCMD(printk("count is now:%d\n",t -> run_count));
      mutex_exit(&(task_list -> mutex));
      return;
    }

  t -> run_count = 1;
  enqueue_tasklet(task_list,t);
  t -> state = TASKLET_STATE_SCHED;

  //check the state of the task_list to see if we need to run it.
  switch(task_list -> run_state)
    {
    case TASKLET_STATE_DISABLED:
      need_int = TRUE;
      task_list -> run_state = TASKLET_STATE_SCHED;
      break;
    case TASKLET_STATE_SCHED:
    case TASKLET_STATE_RUN:
    default:
      need_int = FALSE;
    }
  mutex_exit(&(task_list -> mutex));

  if (need_int == TRUE)
    {
      //DEBUGCMD(printk("tasklet_reschedule(): triggering tasklet interrupt\n"));
      ddi_trigger_softintr(task_list -> id);
    }
}


int fini_tasklet_vectors(dev_info_t *dip)
{
  //  int status;
  clock_t wait_time;
  //if (get_initialized(dip) & TASKLET_INIT)
    {

      //  printk("fini_tasklet_vectors\n");
      // printk("removing interrupt\n");
      ddi_remove_softintr(tasklet_vec[0].id);
      //ddi_remove_softintr(tasklet_hi_vec[0].id);
      
      

      mutex_enter(&(tasklet_vec[0].mutex));
      //untimeout(tasklet_vec[0].timer_id);
      

      tasklet_vec[0].initialized = FALSE;
      tasklet_vec[0].fini_detected = FALSE;
      
      wait_time = ddi_get_lbolt() + drv_usectohz(5000000);
      cv_timedwait(&(tasklet_vec[0].fini_wait),
		   &(tasklet_vec[0].mutex),
		   wait_time);
      mutex_exit(&(tasklet_vec[0].mutex));
      
      // printk("destroying mutex");
      mutex_destroy(&(tasklet_vec[0].mutex));
      cv_destroy(&(tasklet_vec[0].fini_wait));
      //printk("done fini_tasklet_vectors\n");
      //set_initialized(dip,~TASKLET_INIT);
    }
  return(DDI_SUCCESS);


}

int init_tasklet_vectors(dev_info_t *dip)
{
  //int i;
  int status;
  //sun_linux_state instance_state = NULL;

  //getInstanceFReturn(dip,instance_state,sun_state);
  bzero(&tasklet_vec[0],sizeof(struct tasklet_head));

  INIT_LIST_HEAD(&(tasklet_vec[0].list));

  status = ddi_get_soft_iblock_cookie(dip,
				      DDI_SOFTINT_MED,
				      &tasklet_vec[0].iblock_cookie);

  if (status != DDI_SUCCESS)
    {
      printk("failure getting iblock cookie for tasklet_vec %d",status);
      return DDI_FAILURE;
    }

  mutex_init(&(tasklet_vec[0].mutex),
	     "tasklet_vec",
	     MUTEX_DRIVER,
	     (void*)tasklet_vec[0].iblock_cookie);
  cv_init(&(tasklet_vec[0].fini_wait),NULL,CV_DRIVER,NULL);
  tasklet_vec[0].initialized = TRUE;

  mutex_enter(&(tasklet_vec[0].mutex));

  tasklet_vec[0].run_state = TASKLET_STATE_DISABLED;

  status |= ddi_add_softintr(dip,
			     DDI_SOFTINT_LOW,
			     &tasklet_vec[0].id,
			     NULL,
			     NULL,
			     tasklet_softintr,
			     (caddr_t)&tasklet_vec[0]);
  if (status != DDI_SUCCESS)
    {
      printk("Failure add softintr for tasklet_vec %d",status);
    }

  DEBUGCMD(tasklet_vec[0].timer_id = timeout(tasklet_debug,&(tasklet_vec[0]),drv_usectohz(30*1000000)));

  mutex_exit(&(tasklet_vec[0].mutex));

  if (status == DDI_SUCCESS)
    {
      printk("success tasklet_init\n");
      //return(set_initialized(dip,TASKLET_INIT));
    }
  return (status);
}


void tasklet_init(struct tasklet_struct *t,
		  void (*func)(unsigned long), unsigned long data)
{

  bzero(t,sizeof(struct tasklet_struct));
  INIT_LIST_HEAD(&(t -> head));
  //  t->next = NULL;
  t->state = TASKLET_STATE_DISABLED;
  atomic_set(&t->count, 0);
  t -> run_count = 0;
  t->func = func;
  t->data = data;
}



