/*********************************************/
/*pits_trghba.c                              */
/*********************************************/
#include <sys/scsi/scsi.h>
#include "pits_trghba.h"


char _depends_on="misc/scsi";



/**********************************************/
/*solaris calls                               */
/**********************************************/
int _init() 
{
  int error;
  //  const int max_instance = 10;
  
  
  DEBUGDMESG("_init()\n");
  DEBUGDMESG("mod_install\n");
  error = mod_install(&sun_linkage);
  if (error !=0)
    {
      DEBUGDMESG2("error : %d\n",error);  
    }
  else
    {
      DEBUGDMESG("ddi_soft_state_init\n");
      ddi_soft_state_init(&sun_state,sizeof(sun_linux_state),16);
    }
  DEBUGDMESG("end _init()\n");

  return(error);
}

int _fini()
{
  int error;

  DEBUGDMESG("_fini()\n");
  error = mod_remove(&sun_linkage);

  if (error != 0)
    {
      DEBUGDMESG("end do_fini() error\n");
      return(error);
    }
  ddi_soft_state_fini(&sun_state);
  
  DEBUGDMESG("end do_fini()\n");
  
  pflush();

  return(0);

}

int _info(struct modinfo *modinfop)
{
  int error = 0;
  DEBUGDMESG("_info called\n");
  
  error = mod_info(&sun_linkage,modinfop);
  
  if (error == 0)
    {
      printk("Error with mod_info\n");
    }

  return (error);
}


/*****************************************************/
/*sun_getinfo : follows the getinfo(9E) routine      */
/*****************************************************/
static int sun_getinfo(dev_info_t *dip,ddi_info_cmd_t infocmd, void *arg, void **result)
{
  sun_linux_state *instance_state = NULL;
  int error;
  //int minor = getminor((dev_t)arg);

  /*from instminor.c */
  //int instance = getInstance(minor);
  int instance = 0;
  DEBUGDMESG("inside sun_getinfo()\n");

  if (instance < 0)
    {
      DEBUGDMESG("Instance doesn't exist!\n");
      DEBUGDMESG("not in the instminor structure\n");
      return (DDI_FAILURE);
    }

  switch (infocmd)
    {
      /*
       * return the device node if the driver has attached the driver
       * instance identified by the dev_t value which was passed  
       */
    case DDI_INFO_DEVT2DEVINFO:
      DEBUGDMESG("DDI_INFO_DEVT2DEVINFO\n");
      instance_state = ddi_get_soft_state(sun_state,instance);
      if (instance_state == NULL)
	{
	  *result = NULL;
	  error = DDI_FAILURE;
	}
      else
	{
	  *result = instance_state -> dip;
	  error = DDI_SUCCESS;
	}
      break;
      /* 
       *the driver can always return the instance number given a dev_t
       *value, even if the instance is not attached. */

    case DDI_INFO_DEVT2INSTANCE:
      DEBUGDMESG("DDI_INFO_DEV2INSTANCE\n");
      *result = (void *)instance;
      error = DDI_SUCCESS;
      break;
    default:
      *result = NULL;
      return DDI_FAILURE;
    }



  DEBUGDMESG("done sun_getinfo()\n");


  return error;
} 
static int sun_probe(dev_info_t *dip)
{
  return 0;
}

/******************************************************
 *sun_attach()                                        *
 ******************************************************/
static int sun_attach(dev_info_t *dip, ddi_attach_cmd_t cmd)
{
  int instance = 0;
  int error = 0;
  sun_linux_state *instance_state = NULL;
  //  char *minorname = NULL;
  //  int minornumber = 0;
  
  //  int num_minor = 0;
  //  int *minor_array = NULL;
  //  int minor_index = 0;

  DEBUGDMESG("sun_attach\n");
  switch (cmd)
    {
    case DDI_ATTACH:
      
      DEBUGDMESG("DDI_ATTACH:\n");
      
      instance = ddi_get_instance(dip);  
      
      error = ddi_soft_state_zalloc(sun_state,instance);   
      if (error != 0)   
	{ 
	  printk("error allocating softstate %d\n",error); 
	  return DDI_FAILURE;   
	}   

      getInstanceFReturn(dip,instance_state);


      instance_state -> dip         = dip;  
      instance_state -> instance    = instance;  
      instance_state -> initialized = 0;

      getGlobal() -> current_dip = dip;
      //pits_globals.current_dip = dip;

      error = do_pits_init(dip);
      if (error != 0)
	{
	  printk("Error initializing linux driver %d\n",error);
	  do_pits_fini(dip);
	  ddi_soft_state_free(sun_state,instance);
	  return DDI_FAILURE;
	}

      instance_state -> num_open = 0;
      //instance_state -> fops = global_state.fops;
      ddi_report_dev(dip);
      //pits_globals.attach_count++;
      getGlobal() -> attach_count++;

      break;
    case DDI_RESUME:
      return (DDI_SUCCESS);
    default:
      return (DDI_FAILURE);
    }
  
  
  DEBUGDMESG("end sun_attach\n");
  
  return error;
} 

/**********************************************************
 *sun_detach()                                            *
 **********************************************************/
static int sun_detach(dev_info_t *dip,ddi_detach_cmd_t cmd)
{
  int instance; 
//  int error = 0; 
  
  sun_linux_state *instance_state; 
  /*
  DEBUGDMESG("sun_detach");
  */
  switch (cmd)
    {
    case DDI_DETACH:

      /*
      DEBUGDMESG("DDI_DETACH");
      */
      
      instance = ddi_get_instance(dip);
      printk("instance is %d\n",instance);
      printk("getting instance_state\n");
      instance_state = ddi_get_soft_state(sun_state,0);

      //DEBUGDMESG("Removing from MinorInst");
      //remInstMinorByInstance(instance);

      //ddi_remove_minor_node(dip,NULL);

      getGlobal() -> attach_count --;
      //pits_globals.attach_count--;

      if (getGlobal() -> attach_count <= 0)
	{
	  DEBUGDMESG("No more attached nodes, uninitialize linux driver\n");
	  do_pits_fini(dip);
	}
      ddi_soft_state_free(sun_state,instance);
      break;
    case DDI_SUSPEND:
    default:
      return (DDI_FAILURE);
    }

  /*
  DEBUGDMESG("end sun_detach");
  */
  return (DDI_SUCCESS);
}



/************************************************************/
/*sun_open() : open the file, gets it ready for reading     */
/*calls the corresponding linux fops->open for the instance.*/
/************************************************************/
static int sun_open(dev_t *devp,int flag,int otyp,cred_t *credp)
{
  
  sun_linux_state *instance_state = NULL;
  
  instance_state = ddi_get_soft_state(getSunStateP(),0);
  if (instance_state == NULL) {return ENXIO;}
  

  switch (otyp)
    {
    case OTYP_CHR:
      if (instance_state -> fops == NULL || instance_state -> fops -> open == NULL)
	{
	  if (instance_state -> bdops != NULL && instance_state -> bdops -> open != NULL)
	    {
	      return (pits_block_open(devp,flag,OTYP_BLK,credp));
	    }
	  return (ENXIO);
	}

      return (pits_char_open(devp,flag,otyp,credp));
    case OTYP_BLK:
    case OTYP_LYR:
      return (pits_block_open(devp,flag,otyp,credp));
    default:
      return (ENXIO);
    }
}

/**************************************************************/
/*sun_close()                                                 */
/**************************************************************/
static int sun_close(dev_t dev,int flag,int otyp,cred_t *credp)
{
  sun_linux_state *instance_state = NULL;

  linux_file *open_file;
  linux_inode *open_inode;

  minor_t minor;
  int instance = 0;

  printk("sun_close()\n");

  minor = getminor(dev);

  instance_state = ddi_get_soft_state(getSunStateP(),instance);
  if (instance_state == NULL)
    {
      printk("Null instance?\n");
      return(ENXIO);
    }
  
  open_file  = getOpenFile(instance_state -> linux_major,minor);
  open_inode = getOpenInode(instance_state -> linux_major,minor);

  if (open_file == NULL || open_file -> usage_count == 0)
    {
      printk("couldn't find file major %d minor %d\n",instance_state -> linux_major,minor);
      return(DDI_SUCCESS);
    }
  
  switch (open_file -> sun_opentype)
    {
    case OTYP_CHR:
      printk("OTYP_CHR:%d\n",otyp);
      return (pits_char_close(dev,flag,otyp,credp));
    case OTYP_BLK:
    case OTYP_LYR:
      printk("OTYP_BLK:%d\n",otyp);
      return (pits_block_close(dev,flag,otyp,credp));
    default:
      printk("Unknown:%d\n",otyp);
      return (EINVAL);
    }
 
}

/*****************************************************/
/*sun_read()                                         */
/*****************************************************/
static int sun_read(dev_t dev,struct uio *uio_p,cred_t *credp)
{
  sun_linux_state *instance_state;

  linux_file *open_file;
  linux_inode *open_inode;

  minor_t minor;
  int instance;

  DEBUGDMESG("inside sun_read()\n");
  minor = getminor(dev);
  instance = 0;
  if (instance <0 || instance > 16)
    {
      return (EINVAL);
    }
  instance_state = ddi_get_soft_state(sun_state,instance);
  if (instance_state == NULL)
    {
      printk("Null instance?\n");
      return(ENXIO);
    }
  open_file = getOpenFile(instance_state -> linux_major,minor);
  open_inode = getOpenInode(instance_state -> linux_major,minor);
  if (open_file == NULL || open_file -> usage_count == 0)
    {
      printk("couldn't find file major %d minor %d\n",instance_state -> linux_major,minor);
      return(ENXIO);
    }
  
  switch (open_file -> sun_opentype)
    {
    case OTYP_CHR:
      return(pits_char_read(dev,uio_p,credp));
    case OTYP_BLK:
    case OTYP_LYR:
      return(pits_block_read(dev,uio_p,credp));
    default:
      printk("open type is hashed\n");
      return(ENXIO);
    }



}

/*************************************************************/
/*sun_write()                                                */
/*************************************************************/
static int sun_write(dev_t dev,struct uio *uio_p, cred_t *credp)
{
  sun_linux_state *instance_state;
  linux_file *open_file;
  int minornum,instance;
  
  DEBUGDMESG("inside sun_write\n");
  minornum = getminor(dev);
  instance = 0;
  if (instance < 0 || instance > 16)
    {
      return (EINVAL);
    }
  instance_state = ddi_get_soft_state(sun_state,instance);
  if (instance_state == NULL)
    {
      return (ENXIO);
    }
  open_file = getOpenFile(instance_state -> linux_major,minornum);
  if (open_file == NULL || open_file -> usage_count == 0)
    {
      return (ENXIO);
    }
  switch (open_file -> sun_opentype)
    {
    case OTYP_CHR:
      return (pits_char_write(dev,uio_p,credp));
    case OTYP_BLK:
    case OTYP_LYR:
      return (pits_block_write(dev,uio_p,credp));
    default:
      return (ENXIO);
    }

  
}

/******************************************/
/*sun_chpoll                              */
/******************************************/
static int sun_chpoll(dev_t dev,short events,int anyyet, 
                      short *reventsp,struct pollhead **phpp)
{
  uint8_t status;
  short revent;
  sun_linux_state *instance_state;
  int minornum;
  int instance;
  linux_file *open_file;

  poll_table *linux_poll_table=NULL;

  DEBUGDMESG("inside sun_chpoll\n");

  minornum = getminor(dev);
  //instance = getInstance(minornum);
  instance = 0;
  if (instance < 0)
    {
      return ENXIO;
    }

  instance_state = ddi_get_soft_state(sun_state,instance);

  if (instance_state == NULL)
    {
      return (ENXIO);
    }
  revent = 0;

  /* valid events are:
   * POLLIN | POLLOUT | POLLPRI | POLLHUP | POLLERR
   */

  open_file = getOpenFile(instance_state -> linux_major,minornum);
  if (open_file == NULL)
    {
      return (ENXIO);
    }

  status = open_file -> f_op -> poll(open_file,linux_poll_table);
  
  DEBUGDMESG3("events is:%d status is:%d\n",events,status);

  if (events & POLLPRI)
    {
      DEBUGDMESG("POLLPRI\n");
    }
  if (events & POLLHUP)
    {
      DEBUGDMESG("POLLHUP\n");
    }
  if (events & POLLERR)
    {
      DEBUGDMESG("POLLERR\n");
    }

  if ((events & POLLIN) || (events & POLLRDNORM))
    {
      DEBUGDMESG("Checking for readable\n");
      if (status & POLLIN)
	{
	  DEBUGDMESG("is readable\n");
	  revent |= POLLIN;
	  revent |= POLLRDNORM;
	}
      else
	{
	  DEBUGDMESG("is not readable\n");
	}
    }
    if ((events & POLLOUT) || (events & POLLWRNORM)) 
    {
      DEBUGDMESG("Checking for writable\n");
      
      if (status & POLLOUT)
	{
	  DEBUGDMESG("is writable\n");
	  revent |= POLLOUT;
	}
      else
	{
	  DEBUGDMESG("is not writable\n");
	}
    }
  /*if nothing has occured*/
  if (revent == 0)
    {
      if (!anyyet)
	{
	  *phpp = &instance_state -> phead;
	}
    }
  *reventsp = revent;

  DEBUGDMESG("Done chpoll\n");

  return(0);
     
}

/***************************************************
 *pits_strategy()
 ***************************************************/
static int pits_strategy(struct buf *bp)
{
  int instance = 0;
  sun_linux_state *instance_state = NULL;
  linux_file *open_file = NULL;
  minor_t minor = getminor(bp -> b_edev);

  instance_state = ddi_get_soft_state(sun_state,instance);
  if (instance_state == NULL) {return ENXIO;}
  
  open_file = getOpenFile(instance_state -> linux_major,minor);
  if (open_file == NULL || open_file -> usage_count == 0)
    {
      return (ENXIO);
    }
  switch (open_file -> sun_opentype)
    {
    case OTYP_CHR:
      return (pits_char_strategy(bp));
    case OTYP_BLK:
    case OTYP_LYR:
      return (pits_block_strategy(bp));
    default:
      return (ENXIO);
    }
  
  
  
  
  
}

/***************************************************
 *sun_ioctl() : pg.158 of the manual.
 ***************************************************/
static int sun_ioctl(dev_t dev, int cmd, intptr_t arg, 
		     int mode,cred_t *credp, int *rvalp)
{
  sun_linux_state *instance_state = NULL;
  int instance=0;
  int minor;

  linux_file *open_file;
  DEBUGDMESG("Inside sun_ioctl\n");

  instance_state = ddi_get_soft_state(sun_state,instance);
  if (instance_state == NULL)
    {
      return (ENXIO);
    }

  minor = getminor(dev);

  open_file = getOpenFile(instance_state -> linux_major,minor);
  //what type was the file opened on?
  switch (open_file -> sun_opentype)
    {
    case OTYP_CHR:
      return (pits_char_ioctl(dev,cmd,arg,mode,credp,rvalp));
    case OTYP_BLK:
    case OTYP_LYR:
      return (pits_block_ioctl(dev,cmd,arg,mode,credp,rvalp));
    default:
      return (ENXIO);
    }
} 
