/* t1a.c
 *
 * ioctl tests
 */

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include "scull.h"
/*
 * S means "Set" through a ptr
 * T means "Tell" directly with the argument value
 * G means "Get": reply by setting through a pointer
 * Q means "Query": response is on the return value
 * X means "eXchange": G and S atomically
 * H means "sHift": T and Q atomically
 *
 * int quantum;
 * ioctl(fd,SCULL_IOCSQUANTUM, &quantum); //set
 * ioctl(fd,SCULL_IOCTQUANTUM, quantum); //tell
 * ioctl(fd,SCULL_IOCGQUANTUM, &quantum); //get
 * quantum = ioctl(fd,SCULL_IOCQQUANTUM); //query
 * ioctl(fd,SCULL_IOCXQUANTUM, &quantum); //exchage
 * quantum = ioctl(fd,SCULL_IOCHQUANTUM, quantum); //shift
 */

/*
 * Ioctl definitions
 */

extern int scull_quantum, scull_qset;
/* Use 'k' as magic number */
#define SCULL_IOC_MAGIC  'k'
#define SCULL_IOCRESET    _IO(SCULL_IOC_MAGIC, 0)
#define SCULL_IOCSQUANTUM _IOW(SCULL_IOC_MAGIC,  1, scull_quantum)
#define SCULL_IOCSQSET    _IOW(SCULL_IOC_MAGIC,  2, scull_qset)
#define SCULL_IOCTQUANTUM _IO(SCULL_IOC_MAGIC,   3)
#define SCULL_IOCTQSET    _IO(SCULL_IOC_MAGIC,   4)
#define SCULL_IOCGQUANTUM _IOR(SCULL_IOC_MAGIC,  5, scull_quantum)
#define SCULL_IOCGQSET    _IOR(SCULL_IOC_MAGIC,  6, scull_qset)
#define SCULL_IOCQQUANTUM _IO(SCULL_IOC_MAGIC,   7)
#define SCULL_IOCQQSET    _IO(SCULL_IOC_MAGIC,   8)
#define SCULL_IOCXQUANTUM _IOWR(SCULL_IOC_MAGIC, 9, scull_quantum)
#define SCULL_IOCXQSET    _IOWR(SCULL_IOC_MAGIC,10, scull_qset)
#define SCULL_IOCHQUANTUM _IO(SCULL_IOC_MAGIC,  11)
#define SCULL_IOCHQSET    _IO(SCULL_IOC_MAGIC,  12)



int main(int argc, char **argv)
{
    int written, offset, received, fd;
    char buf[1024];
    char* mes;
    int quantum, setto;
    
    memset(buf, 0, 1024);
    
    fd = open ("/dev/scull0", O_RDWR);
    printf("fd = %d\n", fd);

    ioctl(fd,SCULL_IOCGQUANTUM, &quantum); //get
    printf("get quantum returned: %d\n", quantum);
    quantum = -1;

    quantum = ioctl(fd,SCULL_IOCQQUANTUM); //query
    printf("query quantum returned: %d\n", quantum);
    quantum = -1;

    setto = 2000;
    ioctl(fd,SCULL_IOCSQUANTUM, &setto); //set
    printf("set quantum to: %d\n", setto); 
    setto = -1;

    ioctl(fd,SCULL_IOCGQUANTUM, &quantum); //get
    printf("get quantum returned: %d\n", quantum);
    quantum = -1;

    setto = 3000;
    ioctl(fd,SCULL_IOCTQUANTUM, setto); //tell
    printf("tell quantum to: %d\n", setto); 
    setto = -1;

    ioctl(fd,SCULL_IOCGQUANTUM, &quantum); //get
    printf("get quantum returned: %d\n", quantum);
    quantum = -1;

    quantum = setto = 1000;
    ioctl(fd,SCULL_IOCXQUANTUM, &quantum); //exchage
    printf("exchange quantum to: %d, from: %d\n", setto, quantum); 
    quantum = setto = -1;

    ioctl(fd,SCULL_IOCGQUANTUM, &quantum); //get
    printf("get quantum returned: %d\n", quantum);
    quantum = -1;

    setto = 2500;
    quantum = ioctl(fd,SCULL_IOCHQUANTUM, setto); //shift
    printf("shift quantum to: %d, from: %d\n", setto, quantum); 
    quantum = setto = -1;
    
    mes = "Hello, world!";
    written = write ( fd, mes, 13);

    printf("written = %d\n", written);
    
    offset = lseek (fd, 0, SEEK_SET);

    printf("offset = %d\n", offset);
    
    received = read ( fd, buf, 1024 );

    printf("received = %d\n", received);

    printf("message = %s\n", buf);

    close(fd);
}
