/* t2.c
 *
 * the t2 set tests the blocking read/write behavior 
 * of scullpipe
 * 
 * - t2 does simple hello world.  note that scullpipe's
 * read always starts at the beginning of the buffer
 * so there is no need to lseek
 *
 * - t2a is an interactive write tool.  use this with
 * t2b or t2c running in a different window.  to test
 * blocking write (blocks when buffer is full) you need
 * to fill the buffer, which is 4000 characters by
 * default.  (probably a good idea to make that smaller
 * for the sake of testing.)
 *
 * - t2b is a continuous read tool.  it block on read
 * until something shows up in the buffer and immediately
 * prints to screen
 *
 * - t2c is an interactive read tool.  it waits for 
 * the user to press enter before reading a chunk of upto
 * 50 chars from the buffer.  use this tool to test blocking
 * write (t2b won't allow you to fill the buffer because
 * it keeps reading from it.)
 *
 */

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

int main(int argc, char **argv)
{
    int written, offset, received, fd;
    char buf[1024];
    char* mes;
    
    memset(buf, 0, 1024);
    
    fd = open ("/dev/scullpipe", O_RDWR);

    printf("fd = %d\n", fd);
    
    mes = "Hello, world.";
    written = write ( fd, mes, 13);

    printf("written = %d\n", written);
    
    // note no need to lseek to front in scullpipe
    
    received = read ( fd, buf, 1024 );

    printf("received = %d\n", received);

    printf("message = %s\n", buf);

    close(fd);
}
