/* t3.c
 *
 * the t3 set tests the use of select on the 
 * scull pipe devices
 *
 * - t3 tests simple read/write on scullpipe0-3
 *
 * - t3a is an interactive writer -- user enters
 * target device number and message
 *
 * - t3b is a continuous reader that selects readability
 * on all 4 devices and prints to screen when 
 * something arrives
 *
 * - t3c is a continuous writer that first fills
 * all four pipe buffers and then selects writability 
 * on all 4 pipes, waiting for room to write more data
 * use with t3d.
 *
 * - t3d is an interactive reader that reads up to 50 
 * chars from the pipe number specified by the user
 */

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

int main(int argc, char **argv)
{
    int written, offset, received, fd[4], i;
    char buf[1024];
    char* mes;
    
    memset(buf, 0, 1024);
    
    fd[0] = open ("/dev/scullpipe0", O_RDWR);
    fd[1] = open ("/dev/scullpipe1", O_RDWR);
    fd[2] = open ("/dev/scullpipe2", O_RDWR);
    fd[3] = open ("/dev/scullpipe3", O_RDWR);

    printf("fd[4] = {%d, %d, %d, %d}\n", fd[0], fd[1], fd[2], fd[3]);
    
    mes = "Hello, world!";
    for ( i=0; i<4; i++ )
    {
        written = write ( fd[i], mes, 13);
        printf("written to scullpipe%d = %d\n", i, written);
    }

    for ( i=0; i<4; i++ )
    {
	printf("\nscull%d\n", i);
    
        received = read ( fd[i], buf, 1024 );

        printf("  received = %d\n", received);

        printf("  message = %s\n", buf);
    }
    for ( i=0; i<4; i++ )
        close(fd[i]);
}
